/**
 *  WS2811Driver.cpp - driver class implementation for the ws2811 rgb led controller IC.
 *  Supports MSP430G2 and MSP430F5529 Launchpads set t 16MHz.
 *
 *  Copyright (C) 2012  Rick Kimball rick@kimballsoftware.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Sections of code used from the AdaFruit NeoPixel library:
 *
 *  Written by Phil Burgess / Paint Your Dragon for Adafruit Industries,
 *  contributions by PJRC and other members of the open source community.

 *  Adafruit invests time and resources providing this open source code,
 *  please support Adafruit and open-source hardware by purchasing products
 *  from Adafruit!
 *
 * 09-02-2014
 */
 
#include <msp430.h>
#include <stdint.h>
#include <stdlib.h>
#include "driverlib.h"
#include "WS2811Driver.h"

volatile uint16_t _port_mask; // port mask
uint8_t pixels[LED_TOTAL * 3];   // pointer to array holding all color information
volatile uint8_t brightness;
volatile uint8_t rOffset;   // Index of red byte within each 3-byte pixel
volatile uint8_t gOffset;       // Index of green byte
volatile uint8_t bOffset;       // Index of blue byte

/** @brief Setup hardware for output.
 *
 *  @return Void.
 */
void ledBegin(void)
{
	int i;

	for (i = 0; i < (LED_TOTAL * 3); i++) {
		pixels[i] = 0;
	}
	brightness = 0;

	if (LED_ORDER == 1) { // GRB vs RGB; might add others if needed
		rOffset = 1;
		gOffset = 0;
		bOffset = 2;
	} else if (LED_ORDER == 2) {
		rOffset = 1;
		gOffset = 2;
		bOffset = 0;
	} else {
		rOffset = 0;
		gOffset = 1;
		bOffset = 2;
	}

	GPIO_setAsOutputPin(LED_PORT, LED_PIN);
	GPIO_setDriveStrength(LED_PORT, LED_PIN, GPIO_FULL_OUTPUT_DRIVE_STRENGTH);
	GPIO_setOutputHighOnPin(LED_PORT, LED_PIN);

	switch (LED_PORT)
	{
	default:
	case 0x01:
		_port_mask = 0x0202;
		break;
	case 0x02:
		_port_mask = 0x0203;
		break;
	case 0x03:
		_port_mask = 0x0222;	// __MSP430_BASEADDRESS_PORT3_R__ + OFS_P1OUT
		break;
	case 0x04:
		_port_mask = 0x0223;
		break;
	case 0x05:
		_port_mask = 0x0242;
		break;
	case 0x06:
		_port_mask = 0x0243;
		break;
	case 0x07:
		_port_mask = 0x0262;
		break;
	case 0x08:
		_port_mask = 0x0263;
		break;
	}
}

/** @brief End hardware output.
 *
 *  @return Void.
 */
void ledEnd(void)
{
	GPIO_setAsInputPin(LED_PORT, LED_PIN);
}

/** @brief Function to be used by application to update LED strip
 *
 *  @return Void.
 */
extern void write_ws2811_25(uint8_t *data, uint16_t length, uint8_t pinmask, uint16_t portmask);
void ledShow(void){
	__bic_SR_register(GIE);       	// disable interrupts
 	write_ws2811_25((uint8_t *)pixels, LED_TOTAL * 3, LED_PIN, _port_mask);
	__bis_SR_register(GIE);       	// enable interrupts
}

/** @brief Set pixel color from separate R,G,B components
 *
 *  @param n Number of LEDs modules in the strip
 *  @param r 8bit red value
 *  @param g 8bit green value
 *  @param b 8bit blue value
 *
 *  @return Void.
 */
void setPixelRGB(uint16_t n, uint8_t r, uint8_t g, uint8_t b) {
	uint8_t *p;

	if(n < LED_TOTAL) {
		if (brightness) { // See notes in setBrightness()
			r = (r * brightness) >> 8;
			g = (g * brightness) >> 8;
			b = (b * brightness) >> 8;
		}

		p = &pixels[n * 3];
		p[rOffset] = r;
		p[gOffset] = g;
		p[bOffset] = b;
	};
}

/** @brief Set pixel color from 'packed' 32-bit RGB color
 *
 *  @param n LED in strip to update (indexed at 0)
 *  @param c 32bit packed color value
 *
 *  @return Void.
 */
void setPixelColor(uint16_t n, uint32_t c) {
	setPixelRGB(n, (uint8_t)(c >> 16), (uint8_t)(c >>  8), (uint8_t)c);
}


/** @brief Convert separate R,G,B into packed 32-bit RGB color.
 *         Packed format is always RGB, regardless of LED strand color order.
 *
 *  @param r 8bit red value
 *  @param g 8bit green value
 *  @param b 8bit blue value
 *
 *  @return 32bit packed RGB value.
 */
uint32_t Color(uint8_t r, uint8_t g, uint8_t b) {
  return (((uint32_t)r << 16) | ((uint32_t)g <<  8) | b);
}

/** @brief Fetches the set numbers of LEDs in the strip
 *
 *  @return number of LEDs in strip
 */
uint16_t numPixels(void) {
  return (LED_TOTAL);
}

/** @brief Adjust output brightness; 0=darkest (off), 255=brightest.  This does
 *  NOT immediately affect what's currently displayed on the LEDs.  The
 *  next call to show() will refresh the LEDs at this level.  However,
 *  this process is potentially "lossy," especially when increasing
 *  brightness.  The tight timing in the WS2811/WS2812 code means there
 *  aren't enough free cycles to perform this scaling on the fly as data
 *  is issued.  So we make a pass through the existing color data in RAM
 *  and scale it (subsequent graphics commands also work at this
 *  brightness level).  If there's a significant step up in brightness,
 *  the limited number of steps (quantization) in the old data will be
 *  quite visible in the re-scaled version.  For a non-destructive
 *  change, you'll need to re-render the full strip data.  C'est la vie.
 *
 *  @param b Brightness value for strip
 *
 *  @return void
 */
void setBrightness(uint8_t b) {
	uint8_t  c;
	uint8_t *ptr;
	uint8_t newBrightness;
	uint8_t oldBrightness;
	uint16_t scale;
	uint16_t i;

	// Stored brightness value is different than what's passed.
	// This simplifies the actual scaling math later, allowing a fast
	// 8x8-bit multiply and taking the MSB.  'brightness' is a uint8_t,
	// adding 1 here may (intentionally) roll over...so 0 = max brightness
	// (color values are interpreted literally; no scaling), 1 = min
	// brightness (off), 255 = just below max brightness.
	newBrightness = b + 1;
	ptr = pixels;
	oldBrightness = brightness - 1; // De-wrap old brightness value

	if (newBrightness != brightness) { // Compare against prior value
		// Brightness has changed -- re-scale existing data in RAM
		if (oldBrightness == 0) {
			scale = 0; // Avoid /0
		}
		else if (b == 255) {
			scale = 65535 / oldBrightness;
		}
		else {
			scale = (((uint16_t)newBrightness << 8) - 1) / oldBrightness;
		}

		for (i=0; i < (LED_TOTAL * 3); i++) {
			c = *ptr;
			*ptr++ = (c * scale) >> 8;
		}
		brightness = newBrightness;
	}
}
