// To create a new lilght pattern:
// ---------------------------------------------------------------
// 1. define a LED_PATTERN_xxx value below
//
// 2. add a case statement for the new LED_PATTERN_xxx define,
//    you just created, in ProcessLights()
//
// 3. create a new method for your pattern and put it just above
//    the ProcessLights() method
//
// 4. your new method must have at least two parameters:
//    a) the wait time between internal patterns
//    b) the pattern it is for
//    for example: void ledPattern(uint32_t wait, uint16_t pattern)
//
// 5. in your new method, you need to check that the pattern
//    matches. this is done so if the pattern is used elsewhere,
//    it can skip 'over' the method when the lights are turned off
//    mid-pattern.
//    NOTE: look at the 'if' block at the top of ledColorWipe()
//
// 6. in your method you want the following lines of code:
//    a) call setPixelColor() to fill in the pixels as you wish
//    b) call ledShow() to send the colors to the actual LED strip
//    c) call ledDelay(wait) so the lights are lit for the length
//       of time you want; this will also make sure USB keeps running
//    d) check to see if the _ledPattern changed while ledDelay() was
//       waiting. if so, then exit your method immediately. this
//       handles the case of the lights being turned off while the
//       delay was happening
//
#include <msp430.h>
#include <stdint.h>
#include <stdlib.h>
#include "driverlib.h"
#include "led.h"
#include "WS2811Driver\\WS2811Driver.h"
#include "main.h"

// light patterns
#define LED_PATTERN_OFF				1		// turn off LEDs
#define LED_PATTERN_ALL				2		// cycle through all the patterns
#define LED_PATTERN_CHRIS			3		// 'breath' red, the ngreen, then blue
#define LED_PATTERN_RAINBOW			4
#define LED_PATTERN_RAINBOW_CHASE	5
#define LED_PATTERN_WIPE			6
#define LED_PATTERN_YELLOW			7
#define LED_PATTERN_GREEN			8
#define LED_PATTERN_BLUE_FLASH		9
#define LED_PATTERN_BLUE_SOLID		10
#define LED_PATTERN_RED_FLASH		11
#define LED_PATTERN_RED_SOLID		12
#define LED_PATTERN_WHITE_FLASH		13
#define LED_PATTERN_OPEN			14

// inidcators from the robot
#define ROBOT_AUTO		0x01
#define ROBOT_TELEOP	0x02
#define ROBOT_ALLIANCE	0x04 	// 0 = Blue
#define ROBOT_HDRIVE	0x08
#define ROBOT_ARCADE	0x10
#define ROBOT_PORCUPINE	0x20

// pattern specific variables
volatile uint16_t _ledPattern = (uint16_t)-1;

//int16_t controlsOpened = -1;
uint16_t LightPattern() {
	uint8_t	lights;
	uint32_t buttons;
	uint16_t indicators;

	// if not connected, then no lights
	if (!IsUsbConnected()) {
		return (LED_PATTERN_OFF);
	}

	buttons = GetButtons();

	// we always want to do opening
	if ((buttons & 0x400) != 0) {
		return (LED_PATTERN_OPEN);
	}

	// lights when Autonomous is running
	indicators = GetIndicators();
	if (indicators & ROBOT_AUTO)
	{
		if (indicators & ROBOT_PORCUPINE) {
			if (indicators & ROBOT_ALLIANCE) {
				return (LED_PATTERN_RED_FLASH);
			}
			else {
				return (LED_PATTERN_BLUE_FLASH);
			}
		}
		else if (indicators & ROBOT_ARCADE) {
			if (indicators & ROBOT_ALLIANCE) {
				return (LED_PATTERN_RED_SOLID);
			}
			else {
				return (LED_PATTERN_BLUE_SOLID);
			}
		}
		else {
			return (LED_PATTERN_WHITE_FLASH);
		}
	}

	// if here, we are disabled, so do that stuff
	lights = 0;
	if ((buttons & 0x100) != 0)
	{
		lights |= 0x01;
	}
	if ((buttons & 0x080) != 0)
	{
		lights |= 0x02;
	}
	if ((buttons & 0x040) != 0)
	{
		lights |= 0x04;
	}

	switch (lights)
	{
		default:
		case 0:
			return (LED_PATTERN_OFF);
		case 1:
			return (LED_PATTERN_CHRIS);
		case 2:
			return (LED_PATTERN_RAINBOW);
		case 3:
			return (LED_PATTERN_RAINBOW_CHASE);
		case 4:
			return (LED_PATTERN_WIPE);
		case 5:
			return (LED_PATTERN_YELLOW);
		case 6:
			return (LED_PATTERN_GREEN);
		case 7:
			return (LED_PATTERN_ALL);
	}
}

// turns all the leds Off
void ledTurnOff() {
	uint16_t i;

	// clear all pixels
	for (i = 0; i < numPixels(); i++) {
		setPixelColor(i, 0);
	}
	ledShow();			// turn off the lights
}

// delays for the given number of milliseconds
void ledDelay(int32_t wait) {
	// 1 MHz = 0.001ms
	// 25 MHz = 0.025ms
	// 1/0.25 = 40
	wait *= 10;			// TBD: random number at this point
	while (wait >= 0) {
		// if USB data was sent, then it took longer
		// to process, so subtract more time
		if (CheckUSB()) {
			wait -= 200;	// TBD: random number at this point
		}

		// if we jsut got the request to turn the lights off,
		// then turn them off now.
		// NOTE: setting _ledPattern will cause existing
		//       light pattern logic to kick out
		if (LightPattern() != _ledPattern) {
			_ledPattern = LightPattern();
			if (_ledPattern == LED_PATTERN_OFF) {
				ledTurnOff();
			}
		}

		wait--;
	}
}

// set all dots to given color
//
// c       - color to wipe
// wait    - delay between each wipe
// pattern - LED_PATTERN_xxx this is being used with
//
void ledSolid(uint32_t c, int32_t wait, uint16_t pattern) {
	uint16_t i;

	// clear all pixels
	for (i = 0; i < numPixels(); i++) {
		setPixelColor(i, c);
	}
	ledShow();			// turn off the lights
	ledDelay(wait);
}


uint32_t Wheel(uint8_t WheelPos) {
	if (WheelPos < 85) {
		return (Color(WheelPos * 3, 255 - WheelPos * 3, 0));
	} else if (WheelPos < 170) {
		WheelPos -= 85;
		return (Color(255 - WheelPos * 3, 0, WheelPos * 3));
	} else {
		WheelPos -= 170;
		return (Color(0, WheelPos * 3, 255 - WheelPos * 3));
	}
}

// Fill the dots one after the other with given color
//
// c       - color to wipe
// wait    - delay between each wipe
// pattern - LED_PATTERN_xxx this is being used with
//
void ledColorWipe(uint32_t c, uint32_t wait, uint16_t pattern) {
	uint16_t i;

	// if we are not supposed to do this pattern anymore, get out
	if (_ledPattern != pattern) {
		return;
	}

	// do each pixel one by one
	for (i = 0; i < numPixels(); i++) {
		setPixelColor(i, c);
		ledShow();
		ledDelay(wait);

		// if we are not supposed to do this pattern anymore, get out
		if (_ledPattern != pattern) {
			return;
		}
	}
}

// Theatre-style crawling lights.
//
// c       - color to chase
// wait    - delay between each chase
// pattern - LED_PATTERN_xxx this is being used with
//
void ledTheaterChase(uint32_t c, uint32_t wait, uint16_t pattern) {
	int i;
	int j;
	int k;
	int ii;

	// if we are not supposed to do this pattern anymore, get out
	if (_ledPattern != pattern) {
		return;
	}

	// do 10 chase cycles
	for (k = 0; k < 10; k++) {
		// a chase is 3 pixels
		for (j = 0; j < 3; j++) {
			// turn On every 3rd pixel
			for (i = 0; i < numPixels(); i += 3) {
				setPixelColor((i + j), c);
			}
			ledShow();
			ledDelay(wait);

			// if we are not supposed to do this pattern anymore, get out
			if (_ledPattern != pattern) {
				return;
			}

			// prep for next cycle by turning them back off
			for (ii = 0; ii < numPixels(); ii += 3) {
				setPixelColor((i + j), 0);
			}
		}
	}
}

// Fill the dots one after the other with given color
//
// wait    - delay between each wipe
// pattern - LED_PATTERN_xxx this is being used with
//
void ledRainbowCycle(uint32_t wait, uint16_t pattern) {
	int i;
	int j;

	// if we are not supposed to do this pattern anymore, get out
	if (_ledPattern != pattern) {
		return;
	}

	// cycle through all the colors
	for (j = 0; j < 256; j++) {
		// do each pixel one by one
		for (i = 0; i < numPixels(); i++) {
			setPixelColor(i, Wheel(((i * 256 / numPixels()) + j) & 255));
		}
		ledShow();
		ledDelay(wait);

		// if we are not supposed to do this pattern anymore, get out
		if (_ledPattern != pattern) {
			return;
		}
	}
}

// Theatre-style crawling lights with rainbow effect
//
// wait    - delay between each cycle
// pattern - LED_PATTERN_xxx this is being used with
//
void ledTheaterChaseRainbow(uint32_t wait, uint16_t pattern) {
	int i;
	int j;
	int k;

	// if we are not supposed to do this pattern anymore, get out
	if (_ledPattern != pattern) {
		return;
	}

	// cycle all 256 colors
	for (k = 0; k < 256; k++) {
		// a chase is 3 pixels
		for (j = 0; j < 3; j++) {
			// turn On every 3rd pixel
			for (i = 0; i < numPixels(); i += 3) {
				setPixelColor((i + j), Wheel((i + k) % 255));
			}
			ledShow();
			ledDelay(wait);

			// if we are not supposed to do this pattern anymore, get out
			if (_ledPattern != pattern) {
				return;
			}

			// prep for next cycle by turning them back off
			for (i = 0; i < numPixels(); i += 3) {
				setPixelColor((i + j), 0);
			}
		}
	}
}

// ???
//
// c       - color to chase
// wait    - delay between each cycle
// pattern - LED_PATTERN_xxx this is being used with
//
void ledDimToBright(uint32_t c, uint32_t wait, uint16_t pattern) {
	int i;
	int j;

	// if we are not supposed to do this pattern anymore, get out
	if (_ledPattern != pattern) {
		return;
	}

	// set the color
	for (j = 20; j < 255; j++) {
		for (i = 0; i < numPixels(); i++) {
			setPixelColor(i, c);
		}
		setBrightness(j);

		ledShow();
		ledDelay(wait);

		// if we are not supposed to do this pattern anymore, get out
		if (_ledPattern != pattern) {
			return;
		}
	}
}
void ledBrightToDim(uint32_t c, uint32_t wait, uint16_t pattern) {
	int i;
	int j;

	// if we are not supposed to do this pattern anymore, get out
	if (_ledPattern != pattern) {
		return;
	}

	// set the color
	for (j = 255; j >= 20; j--) {
		for (i = 0; i < numPixels(); i++) {
			setPixelColor(i, c);
		}
		setBrightness(j);

		ledShow();
		ledDelay(wait);

		// if we are not supposed to do this pattern anymore, get out
		if (_ledPattern != pattern) {
			return;
		}
	}
}

void ledFlash(uint32_t c, uint32_t wait, uint16_t pattern) {
	int i;

	// if we are not supposed to do this pattern anymore, get out
	if (_ledPattern != pattern) {
		return;
	}

	// set the color
	for (i = 0; i < numPixels(); i++) {
		setPixelColor(i, c);
	}
	ledShow();
	ledDelay(wait);

	ledTurnOff();
	ledDelay(wait);
}

void ledOpen(uint32_t c, uint32_t wait, uint16_t pattern) {
	int i;

	// if we are not supposed to do this pattern anymore, get out
	if (_ledPattern != pattern) {
		return;
	}

	ledTurnOff();

	// this function assumes 69 LEDs, so only run if we have that many
	if (numPixels() < 69) {
		for (i = 0; i < numPixels(); i += 3) {
			setPixelColor(i, Color(255, 0, 0));
		}
		for (i = 1; i < numPixels(); i += 3) {
			setPixelColor(i, Color(0, 255, 0));
		}
		for (i = 2; i < numPixels(); i += 3) {
			setPixelColor(i, Color(0, 0, 255));
		}
		ledShow();
		ledDelay(wait);
		return;
	}

	// i loop var:    34 33 32 31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
	// Front 1511:                                                                                   0  1  2  3  4  5  6  7  8
	// Lid 1511  :                                                                      17 18 19 20 21 22 23 24
	//                                                                                  16 15 14 13 12 11 10  9
	// Rolling Front: 46 45 44 43 42 41 40 39 38 37 36 35 34 33 32 31 30 29 28 27 26 25
	// Rolling Lid:   47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68
	// do each pixel one by one
	for (i = 0; i < 5; i++) {
		setPixelColor(8 - i, c);
		ledShow();
		ledDelay(wait);
	}

	for (i = 0; i < 8; i++) {
		if (i <= 3) {
			setPixelColor(3 - i, c);
		}
		setPixelColor(24 - i, c);
		setPixelColor(9 + i, c);
		ledShow();
		ledDelay(wait);
	}

	for (i = 0; i < 22; i++) {
		setPixelColor(25 + i, c);
		setPixelColor(68 - i, c);
		ledShow();
		ledDelay(wait);
	}

	ledDelay(1000);
}

// Handle Light patterns
void ProcessLights(bool checkUSB) {
	uint16_t	ledPatternPrev = _ledPattern;
	uint16_t	ledPattern;			// pattern for this process pass

	// get pattern for this pass nad update global variable.
	// the global variable is used to by ledDelay to kick out if
	// the light pattern changes while in the middle of this one
	_ledPattern = LightPattern();

	// save the pattern for this pass of ProcessLights. We need
	// our own variable so we can send in this pass' pattern; which
	// will allow each led function to kick out if the pattern
	// changed in the middle of this pass
	ledPattern = _ledPattern;

	// did the desired pattern
	switch (ledPattern) {
		case LED_PATTERN_OFF:
			// if lights are not off yet, then turn them off now
			if (ledPatternPrev != LED_PATTERN_OFF) {
				ledTurnOff();
			}
			ledDelay(100);	// give user a chance to flip the switch
			break;

		case LED_PATTERN_ALL:
			// NOTE: for each of following, pass in the actual
			//       define (and not _ledPattern) because
			//       ledDelay() may change _ledPattern and we
			//       use that get end the current pattern
			ledColorWipe(Color(255, 0, 0), 50, ledPattern); // Red
			ledColorWipe(Color(0, 255, 0), 50, ledPattern); // Green
			ledColorWipe(Color(0, 0, 255), 50, ledPattern); // Blue

			ledTheaterChase(Color(127, 127, 127), 50, ledPattern); // White
			ledTheaterChase(Color(127, 0, 0), 50, ledPattern); // Red
			ledTheaterChase(Color(0, 0, 127), 50, ledPattern); // Blue

			ledRainbowCycle(50, ledPattern);
			ledRainbowCycle(50, ledPattern);
			ledRainbowCycle(50, ledPattern);
			ledRainbowCycle(50, ledPattern);
			ledRainbowCycle(50, ledPattern);
			ledTheaterChaseRainbow(50, ledPattern);
			break;

		case LED_PATTERN_CHRIS:
			ledDimToBright(Color(255, 0, 0), 20, ledPattern);
			ledBrightToDim(Color(255, 0, 0), 20, ledPattern);
			ledDimToBright(Color(0, 255, 0), 20, ledPattern);
			ledBrightToDim(Color(0, 255, 0), 20, ledPattern);
			ledDimToBright(Color(0, 0, 255), 20, ledPattern);
			ledBrightToDim(Color(0, 0, 255), 20, ledPattern);
			break;

		case LED_PATTERN_RAINBOW:
			ledRainbowCycle(50, ledPattern);
			break;

		case LED_PATTERN_RAINBOW_CHASE:
			ledTheaterChaseRainbow(175, ledPattern);
			break;

		case LED_PATTERN_WIPE:
			ledColorWipe(Color(255, 0, 0), 50, ledPattern); // Red
			ledColorWipe(Color(0, 255, 0), 50, ledPattern); // Green
			ledColorWipe(Color(0, 0, 255), 50, ledPattern); // Blue
			break;

		case LED_PATTERN_YELLOW:
			ledColorWipe(Color(255, 192, 0), 50, ledPattern); // Yellow
			break;

		case LED_PATTERN_GREEN:
			ledFlash(Color(0, 255, 0), 250, ledPattern); // Green
			break;

		case LED_PATTERN_BLUE_FLASH:
			ledFlash(Color(0, 0, 255), 250, ledPattern);
			break;

		case LED_PATTERN_BLUE_SOLID:
			ledSolid(Color(0, 0, 255), 250, ledPattern);
			break;

		case LED_PATTERN_RED_FLASH:
			ledFlash(Color(255, 0, 0), 250, ledPattern);
			break;

		case LED_PATTERN_RED_SOLID:
			ledSolid(Color(255, 0, 0), 250, ledPattern);
			break;

		case LED_PATTERN_WHITE_FLASH:
			ledFlash(Color(127, 127, 127), 250, ledPattern);
			break;

		case LED_PATTERN_OPEN:
			ledOpen(Color(255, 192, 0), 50, ledPattern); // Yellow
			break;
	}
}
