/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
/*  
 * ======== main.c ========
 * Gamepad Demo:
 *
 * This example functions as a gamepad on the host. The gamepad has a HID report as described in
 * report_desc_HID0 variable in descriptors.c. Please note that if this report structure is
 * changed then the following lengths need to be updated -
 * 1. #define report_desc_size_HID0 in descriptors.h needs to be updated with descriptor size
 * 2. report_desc_size and report_len_input need t be uopdated in descriptors.c
 * As is this demo will enumerate with 18 bytes of input report and 2 bytes of output report
 * The input and output report structures for the gamepad as described in USB_gamepad.h
 * The input reports are used to report ADC values and status of buttons (GPIO)
 * The output report is used to set/reset indicators (GPIO)
 */
#include <string.h>
#include "driverlib.h"
#include "USB_config/descriptors.h"
#include "USB_API/USB_Common/device.h"
#include "USB_API/USB_Common/usb.h"                  // USB-specific functions
#include "USB_API/USB_HID_API/UsbHid.h"
#include "USB_app/USB_gamepad.h"
#include "WS2811Driver\\WS2811Driver.h"

/*
 * NOTE: Modify hal.h to select a specific device and board
 */
#include "hal.h"
#include "led.h"

/*********** Application specific globals **********************/
volatile uint8_t adcUpdateAvailable;
volatile USB_GamepadReportTx gamepadReportTx;
volatile USB_GamepadReportRx gamepadReportRx;
volatile uint8_t usbHidReportReceived;
volatile uint8_t usbHidReportSendComplete;

#define CLOCK_FREQ_MHZ		((uint32_t)25)
#define CONNECTION_DELAY	((CLOCK_FREQ_MHZ * 1000) * 25)
#define CONNECTION_SPAMTIME (CONNECTION_DELAY + (CLOCK_FREQ_MHZ * 32))

volatile uint32_t buttonState = 0;
volatile uint16_t indicatorState = 0;
volatile uint32_t connected = 0;	// Team 1511 - keeps track of "how long" we have been connected

bool IsUsbConnected() {
	return (connected >= CONNECTION_DELAY);
}

uint32_t GetButtons() {
	return (buttonState);
}

uint16_t GetIndicators() {
	return (indicatorState);
}

bool CheckUSB()
{
	bool update = false;
	uint32_t sendWait;

	switch (USB_connectionState())
	{
		// This case is executed while your device is enumerated on the
		// USB host
		case ST_ENUM_ACTIVE:
			// on our laptops we have a BIOS issue on boot up.
			// we were getting a "stuck key" error. the listed
			// key would change based on which joystick button
			// was 'pressed'. so we will delay doing any button
			// data sending until x seconds after cable is
			// plugged in.
			if (connected < CONNECTION_DELAY) {
				connected++;

				// if not the first time since we got connected,
				// then just continue to wait
				if (connected != 1) {
					break;
				}

				// this is our first time, so let's prime
				// the pump and send something to the PC
				// so it will be happy
				gamepadReportTx.buttons = 0;
				buttonState = 0;
				update = true;
			}
			else {
				buttonState = buttonsPoll();

				if (buttonState != gamepadReportTx.buttons)
				{
					gamepadReportTx.buttons = buttonState;
					update = true;
				}

				if (adcUpdateAvailable) {
					adcUpdateAvailable = false;
					update = true;
				}
			}

			if (update) {
				usbHidReportSendComplete = FALSE;
				USBHID_sendReport((const uint8_t *)&gamepadReportTx, HID0_INTFNUM);

				// when Windows was starting, the board would 'lock up'.
				// the original code was basically an infinite loop.
				// so we will wait a good amount of time, then give up.
				// if the send doesn't happen right away, it will never
				// happen, so just give up and try again later.
				sendWait = 0;
				while (!usbHidReportSendComplete) {
					sendWait++;
					if (sendWait > ((CLOCK_FREQ_MHZ * 1000) * 1)) {
						break;
					}
				}
				usbHidReportSendComplete = FALSE;
			}

			if (usbHidReportReceived) {
				usbHidReportReceived = FALSE;
				USBHID_receiveReport((uint8_t *)&gamepadReportRx, HID0_INTFNUM);
				setIndicators();

				if (connected >= CONNECTION_DELAY) {
					indicatorState = gamepadReportRx.indicators;
				}
			}
            break;

		// These cases are executed while your device is disconnected from
		// the host (meaning, not enumerated); enumerated but suspended
		// by the host, or connected to a powered hub without a USB host
		// present.
		case ST_PHYS_DISCONNECTED:			// ST_USB_DISCONNECTED
		case ST_ENUM_SUSPENDED:
		case ST_PHYS_CONNECTED_NOENUM_SUSP:	// ST_NOENUM_SUSPENDED
			// Team 1511: do NOT go into lower power mode. we want the lights to work
			//__bis_SR_register(LPM3_bits + GIE);
			_NOP();
			connected = 0;
			break;

		// The default is executed for the momentary state
		// ST_ENUM_IN_PROGRESS.  Usually, this state only last a few
		// seconds.  Be sure not to enter LPM3 in this state; USB
		// communication is taking place here, and therefore the mode must
		// be LPM0 or active-CPU.
		case ST_ENUM_IN_PROGRESS:
		case ST_USB_CONNECTED_NO_ENUM:
			connected = 0;
			break;

		default:
			connected = 0;
			break;
	}

	return (update);
}

/*  
 * ======== main ========
 */
void main (void)
{
	WDT_A_hold(WDT_A_BASE); // Stop watchdog timer

	// Minumum Vcore setting required for the USB API is PMM_CORE_LEVEL_2 .
#ifndef DRIVERLIB_LEGACY_MODE
	PMM_setVCore(PMM_CORE_LEVEL_2);
#else
	PMM_setVCore(PMM_BASE, PMM_CORE_LEVEL_2);
#endif

	// Init all GPIO ports as output to consume less current
	initPorts();
	// Init clocks for 8MHz operation
	initClocks(CLOCK_FREQ_MHZ * 1000000);   // Config clocks. MCLK=SMCLK=FLL=8MHz; ACLK=REFO=32kHz
	//Init Timer to trigger ADC every 10ms
	initTimer();
	//Init ADC to capture all analog channels
	initADC();
	//Init buttons as inputs with pull-ups
	initButtons();
	//Init indicators as output low
	initIndicators();

	//Init USB & events; if a host is present, connect
	USB_setup(TRUE, TRUE);

	// initialize LEDs
	ledBegin();

	//Enable global interrupts
	__enable_interrupt();

	// handle lights
	while (1)
	{
		CheckUSB();
		ProcessLights(true);
	}
} //main()


/*  
 * ======== UNMI_ISR ========
 */
#if defined(__TI_COMPILER_VERSION__) || (__IAR_SYSTEMS_ICC__)
#pragma vector = UNMI_VECTOR
__interrupt void UNMI_ISR (void)
#elif defined(__GNUC__) && (__MSP430__)
void __attribute__ ((interrupt(UNMI_VECTOR))) UNMI_ISR (void)
#else
#error Compiler not found!
#endif
{
	switch (__even_in_range(SYSUNIV, SYSUNIV_BUSIFG))
	{
		case SYSUNIV_NONE:
			__no_operation();
			break;

		case SYSUNIV_NMIIFG:
			__no_operation();
			break;

		case SYSUNIV_OFIFG:
#ifndef DRIVERLIB_LEGACY_MODE
			UCS_clearFaultFlag(UCS_XT2OFFG);
			UCS_clearFaultFlag(UCS_DCOFFG);
			SFR_clearInterrupt(SFR_OSCILLATOR_FAULT_INTERRUPT);
#else
			UCS_clearFaultFlag(UCS_BASE, UCS_XT2OFFG);
			UCS_clearFaultFlag(UCS_BASE, UCS_DCOFFG);
			SFR_clearInterrupt(SFR_BASE, SFR_OSCILLATOR_FAULT_INTERRUPT);
#endif
			break;

		case SYSUNIV_ACCVIFG:
			__no_operation();
			break;

		case SYSUNIV_BUSIFG:
			// If the CPU accesses USB memory while the USB module is
			// suspended, a "bus error" can occur.  This generates an NMI.  If
			// USB is automatically disconnecting in your software, set a
			// breakpoint here and see if execution hits it.  See the
			// Programmer's Guide for more information.
			SYSBERRIV = 0; //clear bus error flag
			USB_disable(); //Disable
			break;
	}
}

//Released_Version_4_10_02
