#include "WPILib.h"

inline double cube(double v) {
	return v*v*v;
}

inline double deadzone(double v, double r) {
	if (v > -r && v < r) {
		return 0;
	}
	return v;
}

class RobotDemo: public SimpleRobot {
	Joystick lstick;
	Joystick rstick;
	Victor left1;
	Victor left2;
	Victor left3;
	Victor right1;
	Victor right2;
	Victor right3;
	Relay compressor;
	DigitalInput overPressureSwitch;
	DoubleSolenoid shifter;
	DoubleSolenoid lift;
#if WORKING_ENCODERS
	Encoder leftEncoder;
	Encoder rightEncoder;
#else
  Counter encoder;
#endif

	double lastShiftTime;
	bool lastGear;
public:
	RobotDemo() :
		lstick(1), rstick(2),

		left1(1), left2(2), left3(3), right1(4), right2(5), right3(6),

		compressor(1, Relay::kForwardOnly),

		overPressureSwitch(1),

		shifter(1, 2), lift(3, 4),

#if WORKING_ENCODERS
		leftEncoder(8,9, false),
		rightEncoder(4,5, false)
#else
    encoder(5)
#endif
	{

#if WORKING_ENCODERS
		leftEncoder.SetDistancePerPulse(0.01);
		rightEncoder.SetDistancePerPulse(0.01);
		leftEncoder.Start();
		rightEncoder.Start();
#else
    encoder.Start();
#endif

		lastShiftTime = 0;
		lastGear = false;

		printf("Constructed\n");
	}

  double GetSpeed() {
#if WORKING_ENCODERS
#else
    double p = encoder.GetPeriod();
    if (p == 0.0) {
      return 1000000;
    } else {
      // ~20 is max free speed
      // 328 cycles / rotation
      // so, RPS = (1/R) / (S)
      return (1.0 / 328.0) / p;
    }
#endif
  }

	void RobotInit() {
		printf("ROBOT INIT\n");
		compressor.Set(Relay::kOff);
	}

	void Disabled() {
		printf("DISABLED\n");
	}

	void RunCompressor() {
		if (overPressureSwitch.Get()) {
			compressor.Set(Relay::kOff);
		} else {
			compressor.Set(Relay::kForward);
		}
	}

	void OscillateLift() {
		static double t = GetTime();
		if (GetTime() - t > 30) {
			t = GetTime();
		}

		if (GetTime() - t > 15) {
			lift.Set(DoubleSolenoid::kForward);
		} else {
			lift.Set(DoubleSolenoid::kReverse);
		}
	}

	void Autonomous() {
		printf("AUTO\n");

		double t = GetTime();
		lift.Set(DoubleSolenoid::kForward);
		while (IsEnabled() && IsAutonomous()) {
			printf("%f\n", GetTime() - t);
			OscillateLift();

			RunCompressor();
			Wait(0.05);
		}
		compressor.Set(Relay::kOff);
	}


	void RunTank(double leftPower, double rightPower, bool highGear) {
		if (highGear != lastGear) {
			lastShiftTime = GetTime();
			lastGear = highGear;
		}
		if (GetTime() - lastShiftTime < 0.150) {
			leftPower *= 0.30;
			rightPower *= 0.30;
		}

		if (highGear) {
			shifter.Set(DoubleSolenoid::kForward);
		} else {
			shifter.Set(DoubleSolenoid::kReverse);
		}

		left3.Set(leftPower);
		left2.Set(leftPower);
		left1.Set(leftPower);
		right1.Set(rightPower);
		right2.Set(rightPower);
		right3.Set(rightPower);
	}

	void DropCenter() {
		lift.Set(DoubleSolenoid::kReverse);
	}

	void LiftCenter() {
		lift.Set(DoubleSolenoid::kForward);
	}

	void OperatorControl() {
		printf("TELEOPERATED\n");
		while (IsEnabled() && IsOperatorControl()) {
			double leftPower = deadzone(cube(-lstick.GetY()), 0.05);
			double rightPower = deadzone(cube(-rstick.GetY()), 0.05);

			bool dropCenter = lstick.GetRawButton(2) || rstick.GetRawButton(2);
			if (dropCenter) {
				DropCenter();
			} else {
				LiftCenter();
			}			
			
			static bool autoShift = false;
			static bool last = false;
			if (lstick.GetRawButton(7) && !last) {
          autoShift = !autoShift;
			}
			last = lstick.GetRawButton(7);
			
			
			
			bool highGear;

      if (!autoShift) {
			highGear = (lstick.GetRawButton(1) || rstick.GetRawButton(1));
} else {
			
			double rps = GetSpeed();
			if (!lastGear && rps > 4) {
          highGear = true;
			} else if (lastGear && rps < 3) {
          highGear = false;
			} else {
          highGear = lastGear;
			}
						printf("%c %f -> %c\n", lastGear ? 'H' : 'L', rps, highGear ? 'H' : 'L');
			}
			

			RunTank(leftPower, rightPower, highGear);

			RunCompressor();
			Wait(0.05);
		}
	}
};

START_ROBOT_CLASS(RobotDemo)
;

