#pragma once

#include <stdint.h>
#include <vector>
#include <initializer_list>

/**
 * @brief Base class of autonomous framework
 *
 * A subclass of this class represents an action:
 * be it ThrowTheBall, DelayAction, PrintAction, or CloseFlangifier.
 *
 * Each action has four methods: #initialize(), #process(), #isFinished(), and #abort().
 * There are only two orders for the calling of the action's four methods:
 *
 * #1
 * - #initialize()
 * - #process()
 * - #isFinished() -> false
 * - ...
 * - #process()
 * - #isFinished() -> false
 * - #process()
 * - #isFinished() -> true
 *
 *
 * #2
 * - initialize()
 * - #process()
 * - #isFinished() -> false
 * - ...
 * - #process()
 * - #isFinished() -> false
 * - #abort()
 *
 * @see SequentialGroup
 * @see ParallelGroup
 */
class Action {
public:
	Action();
	virtual ~Action();

	/**
	 * @brief Called once when the action begins execution
	 *
	 * @see #process()
	 * @see Action
	 */
	virtual void initialize() = 0;
	/**
	 * @brief Called periodically after #initialize() until #abort() or #isFinished() == true
	 *
	 * @see Action
	 */
	virtual void process() = 0;
	/**
	 * @brief Called to abort the execution of an action
	 *
	 * @see #process()
	 * @see Action
	 */
	virtual void abort() = 0;
	/**
	 * @brief Called to check when the action is complete
	 *
	 * @see #process()
	 * @see Action
	 */
	virtual bool isFinished() = 0;

	/**
	 * @brief Called just before command runs.
	 *
	 * @warning Only base classes should override
	 */
	virtual void preHook();
	/**
	 * @brief Called just after command runs.
	 *
	 * @warning Only base classes should override
	 */
	virtual void postHook();
};

class ActionGroup : public Action {
public:
	virtual void add(Action* action)=0;
};

/**
 * @brief Action to perform other %Actions in sequence
 *
 * @warning All sub-Actions will be deleted when this
 *     instance is
 *
 * @see ParallelGroup
 */
class SequentialGroup: public ActionGroup {
public:
	/**
	 * @brief Constructor with no actions
	 */
	SequentialGroup();
	virtual ~SequentialGroup();
	/**
	 * @brief Brace-initialization constructor with several actions
	 *
	 * Equivalent to
	 * @code
	 * auto group1 = new SequentialGroup{new DoAwesomeStuff(), new DoNotAwesomeStuff()}
	 * auto group2 = new SequentialGroup()
	 * group2->add(new DoAwesomeStuff())
	 * group2->add(new DoNotAwesomeStuff())
	 * @endcode
	 *
	 * @see #add()
	 */
	explicit SequentialGroup(std::initializer_list<Action*>);
	/**
	 * @brief Add an action after all previously added actions
	 *
	 * @warning Ownership (and responsibility to delete) of the added Action
	 *   is transferred to the SequentialGroup
	 *
	 * @param action The Action to add
	 */
	void add(Action* action) override;

	void initialize() override;
	void process() override;
	void abort() override;
	bool isFinished() override;
private:
	std::vector<Action*> actions;
	uint32_t index;
};

/**
 * @brief Action to perform other %Actions in parallel
 *
 * @warning All sub-Actions will be deleted when this
 *     instance is
 *
 * @see SequentialGroup
 */
class ParallelGroup: public ActionGroup {
public:
	/**
	 * @brief Constructor with no actions
	 */
	ParallelGroup();
	virtual ~ParallelGroup();
	/**
	 * @brief Brace-initialization constructor with several actions
	 *
	 * Equivalent to
	 * @code
	 * auto group1 = new ParallelGroup{new DoAwesomeStuff(), new DoNotAwesomeStuff()}
	 * auto group2 = new ParallelGroup()
	 * group2->add(new DoAwesomeStuff())
	 * group2->add(new DoNotAwesomeStuff())
	 * @endcode
	 *
	 * @see #add()
	 */
	explicit ParallelGroup(std::initializer_list<Action*>);
	/**
	 * @brief Add an action performed in addition to all previously added actions
	 *
	 * @warning Ownership (and responsibility to delete) of the added Action
	 *   is transferred to the ParallelGroup
	 *
	 * @param action The Action to add
	 */
	void add(Action* action) override;

	void initialize() override;
	void process() override;
	void abort() override;
	bool isFinished() override;
private:
	typedef struct {
		Action* action;
		bool done;
	} Combo;
	std::vector<Combo> actions;
};
