#pragma once

#include <actions/action.h>
#include <string>

/**
 * @brief Action to wait N seconds
 *
 * This is a utility for your convenience.
 *
 * This is best used to delay an immediately following action,
 * or create a minimum duration for a set of actions.
 *
 * Example:
 * @code
 *     // DoImportantStuff after 5 seconds
 *     SequentialGroup* g = new SequentialGroup();
 *     g->add(new DelayAction(5.0));
 *     g->add(new DoImportantStuff());
 * @endcode
 * Example:
 * @code
 *     // DoQuickButNotDeterministicStuff, and complete no earlier than 2 seconds
 *     ParallelGroup* g = new ParallelGroup();
 *     g->add(new DelayAction(2.0));
 *     g->add(new DoQuickButNotDeterministicStuff());
 * @endcode
 */
class DelayAction : public Action {
public:
	/**
	 * @brief Constructor
	 *
	 * @param seconds Number of seconds to delay for
	 */
	DelayAction(double seconds);
private:
	void initialize() override;
	void process() override;
	void abort() override;
	bool isFinished() override;
	double delay;
	double targetTime;
};

/**
 * @brief Action to print a message to standard output
 *
 * Action takes one cycle to complete.
 *
 * This is a utility for your convenience.
 */
class PrintAction : public Action {
public:
	/**
	 * @brief Constructor
	 *
	 * @param message The message to print to standard output
	 */
	PrintAction(std::string message);
private:
	void initialize() override;
	void process() override;
	void abort() override;
	bool isFinished() override;
	std::string text;
};
