#include "auto.h"
#include "actions/basic.h"
#include "autonomous/autobase.h"
#include "autonomous/driveStraight.h"
#include "autonomous/driveTurn.h"
#include "autonomous/openGrippers.h"
#include "autonomous/closeGrippers.h"
#include "autonomous/moveLifter.h"
#include "autonomous/strafe.h"
#include "autonomous/openGrippersTime.h"
#include "autonomous/moveRC.h"
#include "autonomous/closeStabilizers.h"
#include "autonomous/openStabilizers.h"

AutoRunner::AutonConfig::AutonConfig(Drive* d, Gripper* g, Lifter* l,
		TinCanPuller* t) {
	AutoBaseAction::setSystems(d, g, l, t);
}

AutoRunner::AutoRunner(Drive& d, Gripper& g, Lifter& l, TinCanPuller& t) :
				SystemBase("Autonomous"),
				config(&d, &g, &l, &t) {
	initYet = false;
	createModes();
	choice = getNullAutoModeID();
}

AutoRunner::~AutoRunner() {
	for (NamedMode& m : modes) {
		delete m.routine;
	}
}

void AutoRunner::initialize() {
	AutoBaseAction::beginOrderDebug();
	initYet = false;

}

void AutoRunner::process() {
	if (!initYet) {
		modes[choice].routine->initialize();
		initYet = true;
	}
	modes[choice].routine->process();
}

void AutoRunner::debug(Logger& logger) {
}

int AutoRunner::getNumAutoModes() {
	return modes.size();
}
int AutoRunner::getDefaultAutoModeID() {
	return defaultMode;
}
int AutoRunner::getNullAutoModeID() {
	return nullMode;
}
std::string AutoRunner::getAutoModeDescription(int id) {
	if (id < 0 || id >= (int) modes.size()) {
		return "Nonexistent auto mode. Will do nothing.";
	} else {
		return modes[id].label;
	}
}
void AutoRunner::setAutoMode(int id) {
	if (id < 0 || id >= (int) modes.size()) {
		choice = getNullAutoModeID();
	} else {
		choice = id;
	}
}

void AutoRunner::createModes() {

	{
		SequentialGroup* doNothing = new SequentialGroup();
		modes.push_back( { doNothing, "Do nothing for the duration of auto." });
		nullMode = modes.size() - 1;

		ParallelGroup* printAtTimes = new ParallelGroup();
		printAtTimes->add(new PrintAction("Wait for it..."));
		printAtTimes->add(new SequentialGroup {
				new DelayAction(3.0),
				new ParallelGroup {
						new PrintAction("This is"),
						new SequentialGroup {
								new DelayAction(1.2),
								new PrintAction("kinda") } },
				new PrintAction("boring.") }); //IKR, Totes ma'goats
		doNothing->add(printAtTimes);
		doNothing->add(new DelayAction(5.3));
		doNothing->add(new PrintAction("Yes it is.")); //literally this^
	}
	{
		SequentialGroup* driveRight = new SequentialGroup();
		std::string driveRightDesc = "Strafe right in to the Auto zone";
		modes.push_back( { driveRight, driveRightDesc });
		driveRight->add(new Strafe(3, 0.5));
	}
	{
		SequentialGroup* OneRCAboveRamp = new SequentialGroup();
		std::string OneRCAboveRampDesc =
				"Grabs RC and Moves to Position over Ramp";
		modes.push_back( { OneRCAboveRamp, OneRCAboveRampDesc });
		OneRCAboveRamp->add(new DelayAction(.5));
		OneRCAboveRamp->add(new MoveLifter(Lifter::LifterPosition::kPickUpBottomTote));
		OneRCAboveRamp->add(new OpenGrippers());
		OneRCAboveRamp->add(new DriveStraight(40));
		OneRCAboveRamp->add(new CloseGrippers());
		OneRCAboveRamp->add(new DelayAction(.1));
		OneRCAboveRamp->add(new MoveLifter(Lifter::LifterPosition::kTopOfChannel));
		//OneRCAboveRamp->add(new DelayAction(.1));
		//OneRCAboveRamp->add(new DriveStraight(-20));
		//OneRCAboveRamp->add(new Strafe(1,-0.5));
		//OneRCAboveRamp->add(new DriveTurn(-30,0));
	}
	{
		SequentialGroup* oneTote = new SequentialGroup();
		std::string oneToteDesc =
				"Grab a tote and pull it back into the auto zone.";
		modes.push_back( { oneTote, oneToteDesc });
		// Test code here!
		oneTote->add(new DelayAction(.1));
		oneTote->add(new MoveLifter(Lifter::LifterPosition::kPickUpBottomTote));
		oneTote->add(new CloseGrippers());
		oneTote->add(new DelayAction(.1));
		oneTote->add(new MoveLifter(Lifter::LifterPosition::kPickUpSecondToteInStackOrDefaultRCGrabPosition));
		oneTote->add(new DelayAction(.1));
		oneTote->add(new DriveTurn(-85, 0));
		oneTote->add(new DriveStraight(288));
	}
	{
		SequentialGroup* oneRC = new SequentialGroup();
		std::string oneRCDesc =
				"Grab a Recycling Container and pull it back into the auto zone.";
		modes.push_back( { oneRC, oneRCDesc });
		oneRC->add(new DelayAction(.5));
		oneRC->add(new MoveLifter(Lifter::LifterPosition::kPickUpSecondToteInStackOrDefaultRCGrabPosition));
		oneRC->add(new OpenGrippers());
		oneRC->add(new DriveStraight(50));
		oneRC->add(new CloseGrippers());
		oneRC->add(new MoveLifter(Lifter::LifterPosition::kClearFeederSlot));
		oneRC->add(new DriveStraight(-150));
	}
	{
		SequentialGroup* stepRCs = new SequentialGroup();
		std::string stepRCsDesc =
				"<SLOW> RC Puller";
		modes.push_back( { stepRCs, stepRCsDesc });
		stepRCs->add(new DriveStraight(998));
		stepRCs->add(new MoveRC(TinCanPuller::Direction::kOut, 0.96)); //old time for full reel out 1.32, .96....0.05 for drop
		stepRCs->add(new DelayAction(0.5));
		stepRCs->add(new DriveTurn(2, 0));
		stepRCs->add(new DriveTurn(-2, 0));
		stepRCs->add(new DriveTurn(-2, 0));
		stepRCs->add(new DriveTurn(2, 0));
		//stepRCs->add(new DelayAction(.5));
		//stepRCs->add(new DriveTurn(4,0));
		//stepRCs->add(new DriveTurn(-4,0));
		stepRCs->add(new DelayAction(.5));
		stepRCs->add(new DriveStraight(999));
	}
	{
		SequentialGroup* stepRCsQuick = new SequentialGroup();
		std::string stepRCsQuickDesc =
				"<QUICK> RC Puller";
		modes.push_back( { stepRCsQuick, stepRCsQuickDesc });
		stepRCsQuick->add(new DriveStraight(998));
		stepRCsQuick->add(new MoveRC(TinCanPuller::Direction::kOut, 0.05)); //old time for full reel out 1.32, .96....0.05 for drop
		stepRCsQuick->add(new DelayAction(0.5));
		stepRCsQuick->add(new DriveTurn(2, 0));
		stepRCsQuick->add(new DriveTurn(-2, 0));
		stepRCsQuick->add(new DriveTurn(-2, 0));
		stepRCsQuick->add(new DriveTurn(2, 0));
		//stepRCsQuick->add(new DelayAction(.5));
		//stepRCsQuick->add(new DriveTurn(4,0));
		//stepRCsQuick->add(new DriveTurn(-4,0));
		stepRCsQuick->add(new DelayAction(.5));
		stepRCsQuick->add(new DriveStraight(999));
	}
	{
		SequentialGroup* TwoTotesRC = new SequentialGroup();
		std::string TwoTotesRCDesc =
				"Double Stack with RC on top";
		modes.push_back( { TwoTotesRC, TwoTotesRCDesc });
		TwoTotesRC->add(new OpenGrippers);
		TwoTotesRC->add(new MoveLifter(Lifter::LifterPosition::kPickUpSecondToteInStackOrDefaultRCGrabPosition));
		TwoTotesRC->add(new CloseGrippers);
		TwoTotesRC->add(new MoveLifter(Lifter::LifterPosition::kClearToteWithRC));
		TwoTotesRC->add(new DriveStraight(40));
		//TwoTotesRC->add(new MoveLifter(Lifter::LifterPosition::kClearToteWithRC));
		TwoTotesRC->add(new OpenGrippers);
		TwoTotesRC->add(new MoveLifter(Lifter::LifterPosition::kPickUpBottomTote));
		TwoTotesRC->add(new CloseGrippers);
		TwoTotesRC->add(new MoveLifter(Lifter::LifterPosition::kPlaceToteWithRCOnToteToCreateTwoStackWithRCForAutoModeNine));
		TwoTotesRC->add(new CloseStabilizers());
		TwoTotesRC->add(new DriveTurn(160, 0));
		TwoTotesRC->add(new DriveStraight(80));
		TwoTotesRC->add(new OpenStabilizers());
		TwoTotesRC->add(new OpenGrippers);
		TwoTotesRC->add(new MoveLifter(Lifter::LifterPosition::kPickUpBottomTote));
	}
}


