#pragma once

#include "systembase.h"
#include <string>
#include <vector>

class TinCanPuller;
class Drive;
class Gripper;
class Lifter;
class Action;

/**
 * @brief %Controls other systems during the %Autonomous period
 *
 * @see Drive
 * @see Gripper
 * @see Lifter
 */
class AutoRunner: public SystemBase {
public:
	/**
	 * @brief Initialize the Autonomous instance.
	 */
	AutoRunner(Drive&, Gripper&, Lifter&, TinCanPuller&);
	~AutoRunner();

	void initialize() override;
	void process() override;
	void debug(Logger&) override;

	/**
	 * @brief The number of available autonomous modes
	 *
	 * @return N such that the individual autonomous modes are numbered 0 to N-1.
	 */
	int getNumAutoModes();
	/**
	 * @brief The ID for the default autonomous mode
	 *
	 * This mode is the most preferred/used option.
	 *
	 * @return The ID
	 */
	int getDefaultAutoModeID();
	/**
	 * @brief The ID for no autonomous mode
	 *
	 * In this mode, the robot should do nothing.
	 *
	 * @return The ID
	 */
	int getNullAutoModeID();
	/**
	 * @brief A description for a given autonomous mode ID
	 *
	 * @param id The ID for the autonomous mode whose description is wanted
	 *
	 * @return The description string
	 */
	std::string getAutoModeDescription(int id);
	/**
	 * @brief Set the auto mode to be run during the Autonomous period
	 *
	 * The accepted mode ID's include:
	 * - the one returned by #getDefaultAutoModeID()
	 * - the one returned by #getNullAutoModeID()
	 * - any integer in the range 0 to N-1 inclusive, where N comes from
	 *     #getNumAutoModes()
	 *
	 * @param id The ID for the selected mode.
	 */
	void setAutoMode(int id);
private:
	typedef struct {
		Action* routine;
		std::string label;
	} NamedMode;

	void createModes();

	class AutonConfig {
	public:
		AutonConfig(Drive*, Gripper*, Lifter*, TinCanPuller*);
	} config;

	std::vector<NamedMode> modes;
	int choice;
	int defaultMode;
	int nullMode;
	bool initYet;
};
