#include "profiles.h"
#include <cmath>

// WARNING: should time this, because
// a 667 MHz A9 is pretty slow at floating point
// (esp with no compile options, softfp, and pipeline stalls like main memory access)

real binarySolveStep3(real dist, real v2, real a, real maxT, real J) {
	real step = maxT * 0.5;
	real t = step;
	for (int i = 0; i < 10; i++) {
		step *= 0.5;
		real d_trial = t * v2 + a * t * t / 2 - J * t * t * t / 6;
		if (d_trial < dist) {
			t += step;
		} else {
			t -= step;
		}
	}
	return t;
}

CubicProfile::CubicProfile(real travel, real V, real mA, real mJ) {
	A = mA;
	J = mJ;
	totalDist = travel;
	real dist = totalDist / 2;

	real riseTime = V / A - A / J;
	maxV = V;
	if (riseTime >= 0) {
		fullRampDist = V * (V / A + A / J) / 2;
		threestage = true;
	} else {
		fullRampDist = std::sqrt(V / J) * V;
		threestage = false;
	}
	if (dist < fullRampDist) {
		real maxV1 = A * (-A / J + std::sqrt((A * A) / (J * J) + 8 * dist / A))
				/ 2;
		real maxV2 = std::cbrt(dist * dist * J);
		fullRampDist = dist;
		if (maxV1 > maxV2) {
			threestage = true;
			maxV = maxV1;
		} else {
			threestage = false;
			maxV = maxV2;
		}
		threestage = maxV1 > maxV2;
	}

	iJx6 = 6.0 / J;
	if (threestage) {
		v1 = A * A / J / 2.0;
		v2 = maxV - v1;
		d1 = A * A * A / J / J / 6.0;
		d2 = d1 + maxV * riseTime / 2.0;
		iA = 1 / A;
		AdJ = A / J;
	} else {
		tmid = std::cbrt(fullRampDist / J);
		d1 = J * tmid * tmid * tmid / 6;
		v1 = J * tmid * tmid / 2;
		a1 = J * tmid;
	}

}

CubicProfile::State CubicProfile::getState(real distLeft) {
	real median = totalDist / 2;
	if (distLeft > median) {
		return getSubState(totalDist - distLeft);
	} else {
		State q = getSubState(distLeft);
		q.acc = -q.acc;
		return q;
	}
}

CubicProfile::State CubicProfile::getSubState(real x) {
	if (x >= fullRampDist) {
		return {maxV, 0, 0};
	}

	if (threestage) {
		if (x < d1) {
			real t = std::cbrt(x * iJx6);
			return {J * t * t / 2, J * t, J};
		} else if (x < d2) {
			real dz = x - d1;
			real tz = (-v1 + std::sqrt(v1 * v1 + 2 * A * dz)) * iA;
			return {v1 + A * tz, A, 0};
		} else {
			real dz = x - d2;
			real tz = binarySolveStep3(dz, v2, A, AdJ, J);
			real tq = AdJ - tz;
			return {maxV - J * tq * tq / 2, J * tq, -J};
		}
	} else {
		if (x < d1) {
			real t = std::cbrt(x * iJx6);
			real a = J * t;
			real v = J * t * t / 2;
			return {v, a, J};
		} else {
			real dz = x - d1;
			real tz = binarySolveStep3(dz, v1, a1, tmid, J);
			real v = v1 + a1 * tz - J * tz * tz / 2;
			real a = a1 - J * tz;
			return {v, a, -J};
		}
	}
}
