#include "controls.h"
#include "SmartDashboard/SmartDashboard.h"
#include "HAL/HAL.hpp"
#include "DriverStation.h"

TuneableConstant(BACK_UP_DISTANCE, -5.0);

const int JOYSTICK_DRIVER_LEFT = 0;
const int JOYSTICK_DRIVER_RIGHT = 1;
const int JOYSTICK_AUXILIARY = 2;
const int LAUNCHPAD = 3;

// Joysticks 1 and 2 buttons
const int SHIFT_GEARS = 1;
const int RAISE_CENTER_WHEEL = 2;
const int SIGNAL = 3;
const int PONG_MODE_ACTIVATE = 10;
const int DRIVER_OVERRIDE_OPEN = 11;

// Joystick 3 buttons
const int CLOSE_LOWER_GRIPS = 1;
const int OPEN_LOWER_GRIPS = 2;
const int LEVEL_2_PRESET = 3;

const int FEEDER_LOW = 4;
const int FEEDER_HIGH = 5;
const int TOGGLE_STABILIZERS = 6;
const int AUTO_STACK = 7;

const int DRIVE_BACK_A_SMIDGE = 10;

// Joystick 4/Launchpad buttons
const int MAIN_AUTO_MODE = 1;
const int CHOOSE_AUTO_MODE = 2;
const int LIFT_POSITION_SENSORS_BROKEN = 3;
const int GRIPPER_POSITION_SENSORS_BROKEN = 4;
const int ALT_ENABLED = 8;
const int DEBUG_DISABLED = 10;
const int ENABLE_ARCADE_DRIVE = 5;

const double DRIVE_JOYSTICK_DEADBAND = 0.05;
const double LIFTER_JOYSTICK_DEADBAND = 0.1;

const double STABILIZER_HEIGHT_CUTOUT = 0.60;

const char* AUTO_CHOICE = "auto_choice";
const char* AUTO_SWITCH = "auto_switch";
const char* AUTO_DESCRIPTION = "auto_description";
const char* LIFTER_POSITION = "lifter_position";
const char* GRIPPER_POSITION = "gripper_position";

Controls::Controls(Drive& d, Gripper& g, Lifter& l, Logging &u, Vision& v,
		AutoRunner& a, TinCanPuller& p) :
				SystemBase("Controls"),
				drive(d),
				gripper(g),
				lifter(l),
				logging(u),
				vision(v),
				autonomous(a),
				puller(p),
				joystickDriverLeft(JOYSTICK_DRIVER_LEFT),
				joystickDriverRight(JOYSTICK_DRIVER_RIGHT),
				joystickAuxiliary(JOYSTICK_AUXILIARY),
				launchpad(LAUNCHPAD) {
	pongModeOn = false;
	pongButtonPressed = false;
	autoLoadState = kAutoInactive;
}

Controls::~Controls() {
}

void Controls::debug(Logger& logger) {
	Joystick& j = launchpad;
	logger.write(this, "switches %c %c %c %c %c %c %c %c",
			Yn(j.GetRawButton(3)), Yn(j.GetRawButton(4)), Yn(j.GetRawButton(5)),
			Yn(j.GetRawButton(6)), Yn(j.GetRawButton(7)), Yn(j.GetRawButton(8)),
			Yn(j.GetRawButton(9)), Yn(j.GetRawButton(10)));
}

void Controls::initialize() {
	drive.shiftGears(Drive::GearPosition::kHighGear);

	prevGripperPosition = gripper.getPosition();
	prevLifterPosition = lifter.getPosition();
	prevAutoChoice = getAutoMode();
	prevAutoSwitchPosition = getAutoSwitchPosition();
	lastDropWheel = joystickDriverLeft.GetRawButton(RAISE_CENTER_WHEEL)
			|| joystickDriverRight.GetRawButton(RAISE_CENTER_WHEEL);
	dropDown = false;
	lightState = false;
	lastButtonState = false;
	lastStabPos = false;
	isClosed = false;
	stabClosed = false;
	lastIntakePos = false;
	intakeClosed = false;
	backupPressed = false;

	SmartDashboard::PutNumber(GRIPPER_POSITION, prevGripperPosition);
	SmartDashboard::PutNumber(LIFTER_POSITION, prevLifterPosition);
	SmartDashboard::PutNumber(AUTO_SWITCH,
			static_cast<double>(prevAutoSwitchPosition));

}

double deadbandF(double v, double radius) {
	if (v <= radius && v >= -radius) {
		return 0.0;
	} else {
		return v;
	}
}

double stick_power(double v, double exp, double dband) {
	const double x = deadbandF(v, dband);
	double out = sign(x) * pow(fabs(x), exp);
	return out;
//	return sign(x) * x * x;
}

double trad_stick_power(double v) {
	double x = deadbandF(v, LIFTER_JOYSTICK_DEADBAND);
	return x;
}

void Controls::process() {

	DriveMode currentMode = processDrive();

	if( autoLoadState == kAutoInactive)
	{
		processToteManipulation();
	}
	processAutoLoader();

	bool reelInPressed = joystickDriverRight.GetRawButton(5);
	bool reelOutPressed = joystickDriverRight.GetRawButton(4);
	if (reelInPressed == true && reelOutPressed == false) {
		puller.reel(TinCanPuller::Direction::kIn);
	} else if (reelInPressed == false && reelOutPressed == true) {
		puller.reel(TinCanPuller::Direction::kOut);
	} else {
		puller.reel(TinCanPuller::Direction::kStop);
	}

	processBrokenSwitches();
	updateSmartDashboard();
	updateLaunchpadOutputs(currentMode);

}
void Controls::processBrokenSwitches() {
	bool LiftPositionSensorsBroken = launchpad.GetRawButton(
			LIFT_POSITION_SENSORS_BROKEN);
	bool GripperPositionSensorsBroken = launchpad.GetRawButton(
			GRIPPER_POSITION_SENSORS_BROKEN);
	bool UsingAlternateCamera = launchpad.GetRawButton(ALT_ENABLED);
	lifter.setPositionSensorsBroken(LiftPositionSensorsBroken);
	gripper.setPositionSensorsBroken(GripperPositionSensorsBroken);
	logging.setWritingToDashboard(!launchpad.GetRawButton(DEBUG_DISABLED));
	vision.setAlternateMode(UsingAlternateCamera);
}

Controls::DriveMode Controls::processDrive() {
	DriveMode currentMode;
	double exponent, deadband;

	bool driveBackPressed = joystickAuxiliary.GetRawButton(DRIVE_BACK_A_SMIDGE);
	if (driveBackPressed != backupPressed) {
		if (driveBackPressed) {
			printf("trigger: drive back an amount\n");
			drive.autonDriveStraight(BACK_UP_DISTANCE.get(), 0.5);
		} else {
			drive.backToManualControl();
		}
		backupPressed = driveBackPressed;
	}

	if (launchpad.GetRawButton(ENABLE_ARCADE_DRIVE)) {
		currentMode = kArcade;
	} else {
		currentMode = kPorcupine;
	}
	if (backupPressed) {
		printf("backing up....\n");
		currentMode = kAutomaticRoutine;
	}

	bool CenterWheelRaised = joystickDriverLeft.GetRawButton(RAISE_CENTER_WHEEL)
			|| joystickDriverRight.GetRawButton(RAISE_CENTER_WHEEL);

	double joystickDriverLeftPowerY = stick_power(-joystickDriverLeft.GetY(),
			exponent = 2, deadband = DRIVE_JOYSTICK_DEADBAND);
	double joystickDriverRightPowerY = stick_power(-joystickDriverRight.GetY(),
			exponent = 2, deadband = DRIVE_JOYSTICK_DEADBAND);
	double joystickDriverRightPowerX = stick_power(joystickDriverRight.GetX(),
			exponent = 2, deadband = DRIVE_JOYSTICK_DEADBAND);
	double joystickDriverLeftPowerX = stick_power(joystickDriverLeft.GetX(),
			exponent = 2, deadband = DRIVE_JOYSTICK_DEADBAND);

	if (CenterWheelRaised && !lastDropWheel) {
		dropDown = !dropDown;
	}
	lastDropWheel = CenterWheelRaised;
	Drive::MiddleWheelPosition wheelPosition;
	if (dropDown) {
		wheelPosition = Drive::MiddleWheelPosition::kWheelDown;
	} else {
		wheelPosition = Drive::MiddleWheelPosition::kWheelUp;
	}

	drive.setWeightComp(joystickDriverRight.GetZ());
	if (currentMode == kPorcupine) { // porcupine mode
		if (joystickDriverLeft.GetRawButton(3)) {
			drive.slideDrive(joystickDriverLeftPowerY, joystickDriverLeftPowerY,
					0, wheelPosition);
		} else {
			drive.slideDrive(joystickDriverLeftPowerY,
					joystickDriverRightPowerY,
					(0.5 * joystickDriverRightPowerX)
							+ (0.5 * joystickDriverLeftPowerX), wheelPosition);
		}
	} else if (currentMode == kArcade) { // arcade mode
		bool pivotOverLeft = joystickDriverRight.GetRawButton(4);
		bool pivotOverRight = joystickDriverRight.GetRawButton(5);
		if (pivotOverLeft == pivotOverRight) {
			// arcade drive proper
			double leftSidePower = bound(
					joystickDriverLeftPowerY + joystickDriverRightPowerX, -1,
					1);
			double rightSidePower = bound(
					joystickDriverLeftPowerY - joystickDriverRightPowerX, -1,
					1);
			drive.slideDrive(leftSidePower, rightSidePower,
					joystickDriverLeftPowerX, wheelPosition);
		} else {
			if (pivotOverLeft) {
				drive.slideDrive(0, joystickDriverLeftPowerY,
						joystickDriverLeftPowerX, wheelPosition);
			} else {
				drive.slideDrive(joystickDriverLeftPowerY, 0,
						joystickDriverLeftPowerX, wheelPosition);
			}
		}
	} else if (currentMode == kAutomaticRoutine) {
		// do nothing, because the other automatic drive code handles it
	} else {
		printf("Error: Controls drive mode not found %d\n", currentMode);
	}

	// While the shift button is pressed, the robot will be in high gear.
	// If the button isn't pressed, then the robot will be in low gear.
	if (joystickDriverLeft.GetRawButton(SHIFT_GEARS)
			|| joystickDriverRight.GetRawButton(SHIFT_GEARS)) {
		drive.shiftGears(Drive::GearPosition::kHighGear);
	} else {
		drive.shiftGears(Drive::GearPosition::kLowGear);
	}

	return currentMode;
}

void Controls::processToteManipulation() {
	// todo convert to stick_power with exp=1, deadband
	double joystickAuxiliaryLifterSpeed = trad_stick_power(
			-joystickAuxiliary.GetY());
	// If the level 3 preset button is pressed then the Y-Axis is ignored and
	// the lifter will go to level 3. When the button is released, the control
	// is immediately given back to the auxiliary driver.
	bool lifterMoving;
	bool lifterTooHigh = lifter.getPosition() > STABILIZER_HEIGHT_CUTOUT;
	if (joystickAuxiliary.GetRawButton(FEEDER_HIGH)) {
		lifter.goToPosition(Lifter::LifterPosition::kClearFeederSlot);
		lifterMoving = lifter.isAtPosition();
	} else if (joystickAuxiliary.GetRawButton(LEVEL_2_PRESET)) {
		lifter.goToPosition(Lifter::LifterPosition::kPickUpSecondToteInStackOrDefaultRCGrabPosition);
		lifterMoving = lifter.isAtPosition();
	} else if (joystickAuxiliary.GetRawButton(FEEDER_LOW)) {
		lifter.goToPosition(Lifter::LifterPosition::kGrabFromFeeder);
		lifterMoving = lifter.isAtPosition();
	} else {
		lifter.move(-1 * joystickAuxiliaryLifterSpeed);
		lifterMoving = (joystickAuxiliaryLifterSpeed != 0);
	}

	// Controls direction for intake wheels
#if HAS_INTAKE
	bool intakeInwardsPressed = joystickAuxiliary.GetRawButton(INTAKE_INWARDS);
	bool intakeOutwardsPressed = joystickAuxiliary.GetRawButton(
			INTAKE_OUTWARDS);
	if (intakeInwardsPressed == true && intakeOutwardsPressed == true) {
		gripper.spinWheels(Gripper::WheelDirection::kStop);
	} else if (intakeInwardsPressed == true) {
		gripper.spinWheels(Gripper::WheelDirection::kInward);
	} else if (intakeOutwardsPressed == true) {
		gripper.spinWheels(Gripper::WheelDirection::kOutward);
	} else {
		gripper.spinWheels(Gripper::WheelDirection::kStop);
	}
#endif

	/////////////////////////////////////////
	// Gripper, Intake, Stabilizer control //
	/////////////////////////////////////////

	bool forcingAllOpen = joystickDriverRight.GetRawButton(
			DRIVER_OVERRIDE_OPEN);

	bool OpeningGrippers = joystickAuxiliary.GetRawButton(OPEN_LOWER_GRIPS);
	bool ClosingGrippers = joystickAuxiliary.GetRawButton(CLOSE_LOWER_GRIPS);
	// While the trigger is pressed, close the grips. While button 2 is held,
	// open the grips. If both buttons are held, then nothing happens.
	if (forcingAllOpen) {
		gripper.move(Gripper::Direction::kOpen);
	} else if (OpeningGrippers && ClosingGrippers) {
		gripper.move(Gripper::Direction::kStop);
	} else if (ClosingGrippers) { // While closing and NOT opening
		gripper.move(Gripper::Direction::kClose);
	} else if (OpeningGrippers) { // While opening and NOT closing
		gripper.move(Gripper::Direction::kOpen);
	} else {
		gripper.move(Gripper::Direction::kStop);
	}

	// Button 6 toggles the stabilizers
	bool stabPressed = joystickAuxiliary.GetRawButton(TOGGLE_STABILIZERS);
	bool stabRising;
	if (stabPressed == true && lastStabPos == false) {
		stabRising = true;
	} else {
		stabRising = false;
	}
	lastStabPos = stabPressed;

	if (lifterTooHigh || lifterMoving || forcingAllOpen) {
		isClosed = false;
	} else if (stabRising == true) {
		isClosed = !isClosed;
	}
	gripper.closeStabilizer(isClosed);

	if (forcingAllOpen) {
		intakeClosed = false;
	}
	gripper.closeIntake(intakeClosed);
}

void Controls::updateSmartDashboard() {
	double gripperPosition = gripper.getPosition();
	double lifterPosition = lifter.getPosition();
	int autoChoice = getAutoMode();
	AutoSwitchMode autoSwitchPosition = getAutoSwitchPosition();

	if (lifterPosition != prevLifterPosition) {
		prevLifterPosition = lifterPosition;
		SmartDashboard::PutNumber(LIFTER_POSITION, lifterPosition);
	}

	if (gripperPosition != prevGripperPosition) {
		prevGripperPosition = gripperPosition;
		SmartDashboard::PutNumber(GRIPPER_POSITION, gripperPosition);
	}

	if (autoChoice != prevAutoChoice) {
		prevAutoChoice = autoChoice;
//		SmartDashboard::PutNumber(AUTO_CHOICE, static_cast<double>(autoChoice));
		SmartDashboard::PutString(AUTO_DESCRIPTION,
				autonomous.getAutoModeDescription(autoChoice));
	}

	if (autoSwitchPosition != prevAutoSwitchPosition) {
		prevAutoSwitchPosition = autoSwitchPosition;
		SmartDashboard::PutNumber(AUTO_SWITCH,
				static_cast<double>(autoSwitchPosition));
		SmartDashboard::PutString(AUTO_DESCRIPTION,
				autonomous.getAutoModeDescription(autoChoice));
	}

	double matchTime = DriverStation::GetInstance()->GetMatchTime();
	SmartDashboard::PutNumber("match_time", matchTime);
	SmartDashboard::PutNumber("match_yikes", matchTime < 20 ? 1.0 : -1.0);
}

void Controls::updateLaunchpadOutputs(DriveMode mode) {
	int32_t status;
	HALAllianceStationID allianceStationID;
	HALGetAllianceStation(&allianceStationID);
	uint32_t outputs = 0;
	switch (allianceStationID) {
	case HALAllianceStationID::kHALAllianceStationID_blue1:
		break;
	case HALAllianceStationID::kHALAllianceStationID_blue2:
		outputs |= 1 << 11;
		break;
	case HALAllianceStationID::kHALAllianceStationID_blue3:
		outputs |= 1 << 10;
		outputs |= 1 << 11;
		break;
	case HALAllianceStationID::kHALAllianceStationID_red1:
		outputs |= 1 << 2;
		break;
	case HALAllianceStationID::kHALAllianceStationID_red2:
		outputs |= 1 << 2;
		break;
	case HALAllianceStationID::kHALAllianceStationID_red3:
		outputs |= 1 << 2;
		outputs |= 1 << 11;
		break;
	default:
		outputs |= 1 << 8;
		outputs |= 1 << 10;
		outputs |= 1 << 11;
		break;
	}

	outputs |= RobotState.IsAutonomous() << 0;
	outputs |= (RobotState.IsEnabled() && RobotState.IsOperatorControl()) << 1;
	outputs |= false << 3;
	outputs |= mode == kArcade << 4;
	outputs |= mode == kPorcupine << 5;
	outputs |= RobotState.IsFMSConnected() << 6;
	outputs |= HALGetBrownedOut(&status) << 7;
	outputs |= RobotState.IsEnabled() << 9;
	launchpad.SetOutputs(outputs);
}

double SafeGetAutoNumber() {
	try {
		return SmartDashboard::GetNumber(AUTO_CHOICE);
	} catch (const std::exception& e) {
		return 0.0;
	}
}

int Controls::getAutoMode() {
	bool MainAutoMode = launchpad.GetRawButton(MAIN_AUTO_MODE);
	bool ChooseAutoMode = launchpad.GetRawButton(CHOOSE_AUTO_MODE);

	if (MainAutoMode) {
		return autonomous.getDefaultAutoModeID();
	} else if (ChooseAutoMode) {
		return static_cast<int>(SafeGetAutoNumber());
	} else {
		return autonomous.getNullAutoModeID();
	}
}

Controls::AutoSwitchMode Controls::getAutoSwitchPosition() {
	bool MainAutoMode = launchpad.GetRawButton(MAIN_AUTO_MODE);
	bool ChooseAutoMode = launchpad.GetRawButton(CHOOSE_AUTO_MODE);

	if (MainAutoMode) {
		return kUp;
	} else if (ChooseAutoMode) {
		return kDown;
	} else {
		return kNeutral;
	}
}

bool Controls::PongMode() {
	if (RobotState.IsEnabled()) {
		pongModeOn = false;
		return false;
	}

	bool pong = joystickDriverRight.GetRawButton(PONG_MODE_ACTIVATE);

	if (pong != pongButtonPressed) {
		pongButtonPressed = pong;
		if (pong == true) {
			pongModeOn = !pongModeOn;
		}
	}
	return pongModeOn;
}

bool Controls::shouldSignalWithLights() {
	bool buttonPressed = joystickDriverRight.GetRawButton(SIGNAL);
	if (buttonPressed != lastButtonState) {
		lightState = !lightState;
	}
	lastButtonState = buttonPressed;
	return lightState;
}

bool Controls::rightPlayerForward() {
	double rightPlayerDirection = joystickAuxiliary.GetY();
	if (rightPlayerDirection > 0.3) {
		return true;
	}
	return false;
}
bool Controls::rightPlayerBackwards() {
	double rightPlayerDirection = joystickAuxiliary.GetY();
	if (rightPlayerDirection < -0.3) {
		return true;
	}
	return false;
}
bool Controls::leftPlayerForward() {
	double rightPlayerDirection = joystickDriverRight.GetY();
	if (rightPlayerDirection > 0.3) {
		return true;
	}
	return false;
}
bool Controls::leftPlayerBackwards() {
	double rightPlayerDirection = joystickDriverRight.GetY();
	if (rightPlayerDirection < -0.3) {
		return true;
	}
	return false;
}

void Controls::processAutoLoader() {
	bool autostackbutton = joystickAuxiliary.GetRawButton(AUTO_STACK);
	bool LiftPositionSensorsBroken = launchpad.GetRawButton(
			LIFT_POSITION_SENSORS_BROKEN);
	bool GripperPositionSensorsBroken = launchpad.GetRawButton(
			GRIPPER_POSITION_SENSORS_BROKEN);

	switch (autoLoadState) {
	case (kAutoInactive): //Inactive as long as Broken are set || Gripper/Lifter is moving || Button is not pressed
		if (autostackbutton == true) {
			if (!LiftPositionSensorsBroken && !GripperPositionSensorsBroken) {
				if (gripper.isStopped() && lifter.getVelocity() == 0) {
					gripper.move(Gripper::Direction::kOpen);
					autoLoadState = kWaitForGripperOpen;
				}
			}
		}
		break;
	case (kWaitForGripperOpen):	//Waits for Grippers to open to a clearance value
		if (checkForAutoLoadAbort()) {
			gripper.move(Gripper::Direction::kStop);
			autoLoadState = kWaitForActionComplete;
		}
		else if (gripper.isStopped()) {
			lifter.goToPosition(Lifter::LifterPosition::kAutoStackBottom);
			autoLoadState = kWaitForLifterDown;
		}
		break;
	case (kWaitForLifterDown):	//Waits for lifter to be at bottom
		if (checkForAutoLoadAbort()) {
			lifter.goToPosition(Lifter::LifterPosition::kStop);
			autoLoadState = kWaitForActionComplete;

		}
		else if (lifter.isAtPosition()) {
			gripper.move(Gripper::Direction::kClose);
			autoLoadState = kWaitForGripperClosed;
		}
		break;
	case (kWaitForGripperClosed)://Waits for Grippers to close on totes at bottom
		if (checkForAutoLoadAbort()) {
			gripper.move(Gripper::Direction::kStop);
			autoLoadState = kWaitForActionComplete;
		}
		else if (gripper.isStopped()) {
			lifter.goToPosition(Lifter::LifterPosition::kAutoStackTop);
			autoLoadState = kWaitForLifterUp;
		}
		break;
	case (kWaitForLifterUp):	//Waits for lifter to get to clearance value
		if (checkForAutoLoadAbort()) {
			lifter.goToPosition(Lifter::LifterPosition::kStop);
			autoLoadState = kWaitForActionComplete;
		}
		else if (lifter.isAtPosition()) {
			autoLoadState = kWaitForActionComplete;
		}
		break;
	case (kWaitForActionComplete)://Waits to see if all other actions are complete
		if (!autostackbutton) {
			autoLoadState = kAutoInactive;
		}
		break;
	default:	//if nothing is true goes to default state
		break;
	}
}
bool Controls::checkForAutoLoadAbort() {
	bool autostackbutton = joystickAuxiliary.GetRawButton(AUTO_STACK);
	bool LiftPositionSensorsBroken = launchpad.GetRawButton(
			LIFT_POSITION_SENSORS_BROKEN);
	bool GripperPositionSensorsBroken = launchpad.GetRawButton(
			GRIPPER_POSITION_SENSORS_BROKEN);
	double joystickAuxiliaryLifterSpeed = trad_stick_power(
			-joystickAuxiliary.GetY());

	if (!autostackbutton)
	{
		return true;
	}
	if (LiftPositionSensorsBroken || GripperPositionSensorsBroken)
	{
		return true;
	}
	if (joystickAuxiliaryLifterSpeed != 0)
	{
		return true;
	}
	if (joystickAuxiliary.GetRawButton(CLOSE_LOWER_GRIPS)
			|| joystickAuxiliary.GetRawButton(OPEN_LOWER_GRIPS))
	{
		return true;
	}
	return false;
}
