#pragma once

#include "systembase.h"
#include "drive.h"
#include "lifter.h"
#include "gripper.h"
#include "logging.h"
#include "auto.h"
#include "vision.h"
#include "rcpuller.h"
#include "Joystick.h"

/**
 * @brief Directs robot functions in Teleop and Disabled modes
 *
 * It should communicate with the driver station (and its dashboard)
 * and use that information to command the subsystems under its control.
 *
 * The Autonomous class controls the robot in during the autonomous period.
 *
 * @see Drive
 * @see Lifter
 * @see Gripper
 * @see Logging
 */
class Controls : public SystemBase {

public:
	/**
	 * @brief Constructor for the Controls class.
	 *
	 * This sets the initial setup of the class.
	 *
	 */
	Controls(Drive&, Gripper&, Lifter&, Logging&, Vision&, AutoRunner&, TinCanPuller&);
	~Controls();

	void initialize() override;
	void debug(Logger&) override;
	void process() override;

	typedef enum  {
			kArcade, kPorcupine, kAutomaticRoutine
	} DriveMode;

	typedef enum {
			kUp, kNeutral, kDown
	} AutoSwitchMode;
	/**
	 *
	 * @brief Provide the chosen autonomous routine.
	 *
	 * This method is called first of all methods at the start of %Autonomous.
	 *
	 * @return The selected autonomous mode number.
	 */
	typedef enum{
		kAutoInactive,		//state it is in when waiting to be activated
		kWaitForGripperOpen, 	//state in which it is opening the grippers
		kWaitForLifterDown, 	//state in which we are lowering the Lifter
		kWaitForGripperClosed, 	//state in which we are closing the grippers on the bottom totes
		kWaitForLifterUp,	//state in which we are lifting the totes up to level 3
		kWaitForActionComplete	//waits to see if all other states are done
	} AutoLoadState;

	int getAutoMode();
	/**
	 *
	 */
	bool PongMode();
	/**
	 *
	 */
	bool shouldSignalWithLights();
	/**
	 *
	 */
	bool rightPlayerForward();
	/*
	 *
	 */
	bool rightPlayerBackwards();
	/**
	 *
	 */
	bool leftPlayerForward();
	/**
	 *
	 */
	bool leftPlayerBackwards();
	/**
	 *
	 */
	bool checkForAutoLoadAbort();
private:
	Controls::DriveMode processDrive();
	void processToteManipulation();
	void processBrokenSwitches();
	void updateSmartDashboard();
	void updateLaunchpadOutputs(DriveMode d);
	AutoSwitchMode getAutoSwitchPosition();
	void processAutoLoader();
	AutoLoadState autoLoadState;

	Drive& drive;
	Gripper& gripper;
	Lifter& lifter;
	Logging& logging;
	Vision& vision;
	AutoRunner& autonomous;
	TinCanPuller& puller;
	Joystick joystickDriverLeft;
	Joystick joystickDriverRight;
	Joystick joystickAuxiliary;
	Joystick launchpad;

	bool lastDropWheel;
	bool dropDown;
	bool pongModeOn;
	bool pongButtonPressed;
	bool lastStabPos;
	bool isClosed;
	bool stabClosed;
	bool lastIntakePos;
	bool lightState = true;
	bool lastButtonState;
	bool intakeClosed;

	bool backupPressed;

	double prevGripperPosition;
	double prevLifterPosition;
	int prevAutoChoice;
	AutoSwitchMode prevAutoSwitchPosition = kNeutral;


};
