#include "drive.h"
#include "hardware.h"
#include <util/common.h>

const double SHIFTING_SLOWDOWN_PERCENT = .300; //what percent of power we need to go when shifting
const double SHIFT_TIME_REQUIRED = .15;
const double AUTONOMOUS_MIN_POWER = .15;
const double AUTONOMOUS_RC_MIN_POWER = .4;
const double AUTONOMOUS_TURN_MAX_POWER = .45;
const double ACCELERATION_TIME = .5;
const double DECCELERATION_DIST = 40;
const double GYRO_ADJUSTMENT_P = .01;
const double WHEEL_2_MIDDLE_WIDTH = 13.5;
const double STRAFE_GYRO_ADJUSTMENT_P = 0.01;

Drive::Drive() :
				SystemBase("Drive"),
				Base(HardwareAccess::getInstance()) {
	isCenterWheelDown = false;
	initialize();
}

Drive::~Drive() {

}

void Drive::initialize() {
	if (RobotState.IsEnabled()) {
		Base.endPositionSensorCalibration();
	} else if (RobotState.IsDisabled()) {
		Base.beginPositionSensorCalibration();
	}

	last_Shift_Timer.Reset(); // start/reset Timer
	autoState = Manual;
	lastShiftPos = GearPosition::kLowGear;	//last shift was to low gear
	leftPower = 0;
	rightPower = 0;
	centerPower = 0;	//setting all variables mate
	powerMax = .25;
	distGoal = 0;
	turnGoal = 0;
	isStrafing = false;
	strafeAngle = 0.0;
	lastPosition = Base.getPosition();
	lastTime = GetTime();
	weight_comp = 0;

	shiftGears(GearPosition::kLowGear);	//shifting to low gear
}

void Drive::process() {
	switch (autoState) {//switch for if we are currently in autonomous, and what we are trying to do in autonomous
	case Manual: {
		strafeCompensate();

		break;
	}
	case AutoNoRequest: {//we are done doing our previous request and now we do nothing
		slideDrive(0, 0, 0, MiddleWheelPosition::kWheelUp);
		break;
	}
	case AutoSlide: {	//strafing left/right
		Base.extendDropPiston(true);	//drop the piston
		//centerPower = autoAccel(centerPower, powerMax, AUTONOMOUS_MIN_POWER);//accelerate the center wheel to power
		wantedRPower = (leftPower - rightPower) / 2.0;
		wantedYPower = (leftPower + rightPower) / 2.0;
		wantedXPower = (centerPower);
		strafeCompensate();

		if ((GetTime() - startingTime) > distGoal) {
			autoDoNothing();
		}
		break;
	}
	case AutoStraight: {
		Base.extendDropPiston(false);
		if (fabs(Base.getLeftDriveTravel()) - fabs(distanceStartLeft)
				>= (fabs(distGoal) - DECCELERATION_DIST)
				|| fabs(Base.getRightDriveTravel()) - fabs(distanceStartRight)
						>= fabs(distGoal) - DECCELERATION_DIST) {
			printf("Is decelerating L %.4f R %.4f G %.4f\n",fabs(Base.getLeftDriveTravel()) - fabs(distanceStartLeft), fabs(Base.getRightDriveTravel()) - fabs(distanceStartRight), fabs(distGoal));
			doneAccelerating = true;
			rightPower = autoDeccel(rightPower, powerMax, 0);
			leftPower = autoDeccel(leftPower, powerMax, 0);

		}
		if (fabs(Base.getLeftDriveTravel()) - fabs(distanceStartLeft)
				>= fabs(distGoal)
				|| fabs(Base.getRightDriveTravel()) - fabs(distanceStartRight)
						>= fabs(distGoal)) {
			rightPower = 0;
			leftPower = 0;
			autoState = AutoNoRequest;
			break;
		}

		rightPower = autoAccel(rightPower, powerMax, AUTONOMOUS_MIN_POWER);
		leftPower = autoAccel(leftPower, powerMax, AUTONOMOUS_MIN_POWER);
		rightPower = gyroStraighten(rightPower, powerMax, AUTONOMOUS_MIN_POWER,
				Base.getPosition().rotation);
		leftPower = gyroStraighten(leftPower, powerMax, AUTONOMOUS_MIN_POWER,
				-Base.getPosition().rotation);
		break;
	}
	case AutoRCStraight: {
			Base.extendDropPiston(false);
			if (fabs(Base.getLeftDriveTravel()) - fabs(distanceStartLeft)
					>= (fabs(distGoal) - DECCELERATION_DIST)
					|| fabs(Base.getRightDriveTravel()) - fabs(distanceStartRight)
							>= fabs(distGoal) - DECCELERATION_DIST) {
				printf("Is decelerating L %.4f R %.4f G %.4f\n",fabs(Base.getLeftDriveTravel()) - fabs(distanceStartLeft), fabs(Base.getRightDriveTravel()) - fabs(distanceStartRight), fabs(distGoal));
				doneAccelerating = true;
				rightPower = autoDeccel(rightPower, powerMax, 0);
				leftPower = autoDeccel(leftPower, powerMax, 0);

			}
			if (fabs(Base.getLeftDriveTravel()) - fabs(distanceStartLeft)
					>= fabs(distGoal)
					|| fabs(Base.getRightDriveTravel()) - fabs(distanceStartRight)
							>= fabs(distGoal)) {
				rightPower = 0;
				leftPower = 0;
				autoState = AutoNoRequest;
				break;
			}

			rightPower = autoAccel(rightPower, powerMax, AUTONOMOUS_RC_MIN_POWER);
			leftPower = autoAccel(leftPower, powerMax, AUTONOMOUS_RC_MIN_POWER);
			rightPower = gyroStraighten(rightPower, powerMax, AUTONOMOUS_RC_MIN_POWER,
					Base.getPosition().rotation);
			leftPower = gyroStraighten(leftPower, powerMax, AUTONOMOUS_RC_MIN_POWER,
					-Base.getPosition().rotation);
			break;
	}
	case AutoTurn: {
		double weakSidePower = AUTONOMOUS_TURN_MAX_POWER * turnBotDirection; //= autoAccel(
		//((distGoal - WHEEL_2_MIDDLE_WIDTH)
		// (distGoal + WHEEL_2_MIDDLE_WIDTH))
		//* AUTONOMOUS_TURN_MAX_POWER * turnBotDirection,
		//AUTONOMOUS_TURN_MAX_POWER, AUTONOMOUS_MIN_POWER);
		double strongSidePower = AUTONOMOUS_TURN_MAX_POWER * -turnBotDirection; //= autoAccel(
		//AUTONOMOUS_TURN_MAX_POWER * turnBotDirection,
		//AUTONOMOUS_TURN_MAX_POWER, AUTONOMOUS_MIN_POWER);
		if (turnAngleDirection > 0) {
			rightPower = weakSidePower;
			leftPower = strongSidePower;
		} else if (turnAngleDirection < 0) {
			rightPower = strongSidePower;
			leftPower = weakSidePower;
		}
		if (fabs(Base.getPosition().rotation) >= fabs(turnGoal) - 2.5) {
			autoState = AutoNoRequest;
		}

		break;
	}

	case AutoBroke: {

	}
	}
	if (last_Shift_Timer.Get() <= SHIFT_TIME_REQUIRED) {//are we currently in a state of shifting?(in this state for .175 seconds)
		Base.setLeftDriveMotorPower(leftPower * SHIFTING_SLOWDOWN_PERCENT);
		Base.setRightDriveMotorPower(rightPower * SHIFTING_SLOWDOWN_PERCENT);
		//Yes? drive at 30% speed
	} else {
		Base.setLeftDriveMotorPower(leftPower);
		Base.setRightDriveMotorPower(rightPower);
		//No? drive at the speed passed to us
	}

	if (lastShiftPos == GearPosition::kHighGear) { //Set the shifter position
		Base.extendShifterPistons(true); //High gear
	} else {										//
		Base.extendShifterPistons(false);						//low gear
	}
	Base.setCenterWheelPower(centerPower);		//set the center wheel speed
}

void Drive::debug(Logger& logger) {
	logger.write(this, "LP %.4f RP %.4f CP %.4f", leftPower, rightPower,
			centerPower);
	logger.write(this, "Center down %c shift down %c", Yn(isDropWheelDown()),
			Yn(isShiftedDown()));
	logger.write(this, "Drive auto mode %d", autoState);
	logger.write(this, "Drive auto distance goal %f", distGoal);
}

void Drive::slideDrive(double incomingLeftPower, double incomingRightPower,
		double incomingMiddlePower, MiddleWheelPosition position) {

	rightPower = incomingRightPower;
	leftPower = incomingLeftPower;

	if (position == MiddleWheelPosition::kWheelDown) {
		Base.extendDropPiston(true);
		isCenterWheelDown = true;
		centerPower = incomingMiddlePower;
	} else {
		centerPower = 0;
		Base.extendDropPiston(false);
		isCenterWheelDown = false;
	}

	wantedRPower = (leftPower - rightPower) / 2.0;
	wantedYPower = (leftPower + rightPower) / 2.0;
	wantedXPower = (centerPower);
}

void Drive::autonDriveStraight(double distance, double maxPower) {
	double sgn = sign(distance);
	Base.zeroPosition();
	autoState = AutoStraight;
	distGoal = distance;
	powerMax = maxPower;
	rightPower = AUTONOMOUS_MIN_POWER * sgn;
	leftPower = AUTONOMOUS_MIN_POWER * sgn;
	startingTime = GetTime();
	doneAccelerating = false;
	distanceStartLeft = Base.getLeftDriveTravel();
	distanceStartRight = Base.getRightDriveTravel();
}
void Drive::autonRCDriveBack(double distance, double maxPower) {
	double sgn = sign(distance);
	Base.zeroPosition();
	autoState = AutoRCStraight;
	distGoal = distance;
	powerMax = maxPower;
	rightPower = AUTONOMOUS_RC_MIN_POWER * sgn;
	leftPower = AUTONOMOUS_RC_MIN_POWER * sgn;
	startingTime = GetTime();
	doneAccelerating = false;
	distanceStartLeft = Base.getLeftDriveTravel();
	distanceStartRight = Base.getRightDriveTravel();
}

void Drive::autonDriveTurn(double turnAngle, double radius) {
	Base.zeroPosition();
	doneAccelerating = false;
	turnGoal = fabs(turnAngle);
	distGoal = fabs(radius);
	if (turnGoal == 0) {
		turnAngleDirection = 0;
	} else {
		turnAngleDirection = turnAngle / turnGoal;

	}
	if (distGoal == 0) {
		turnBotDirection = 1;
	} else {
		turnBotDirection = radius / distGoal;
	}
	autoState = AutoTurn;
}

void Drive::autonStrafe(double time, double maxPower) {
	Base.extendDropPiston(true);
	startingTime = GetTime();
	autoState = AutoSlide;
	powerMax = fabs(maxPower);
	distGoal = time;
	centerPower = maxPower;
	leftPower = 0;
	rightPower = 0;
	isStrafing = false;
}

bool Drive::autonDone() {
	if (autoState == AutoNoRequest) {
		return true;
	} else {
		return false;
	}

}

void Drive::autoDoNothing() {
	autoState = AutoNoRequest;
}

void Drive::backToManualControl() {
	autoState = Manual;
}

void Drive::shiftGears(GearPosition pos) {
	if (pos != lastShiftPos) {
		lastShiftPos = pos;
		last_Shift_Timer.Reset();
	}
}

double Drive::autoAccel(double motorPower, double maxPower, double minPower) {
	double sign = motorPower / fabs(motorPower);
	if (!doneAccelerating) {		//if we are still in a mode of accelerating
		motorPower = ((maxPower - minPower) * sign / (ACCELERATION_TIME))
				* (GetTime() - startingTime) + minPower * sign;
	}
	if ((GetTime() - startingTime) > ACCELERATION_TIME) {
		doneAccelerating = true;
	}
	return motorPower;
}

double Drive::gyroStraighten(double currentPower, double maxPower,
		double minPower, double gyroValue) {
	if (doneAccelerating) {
		double gyroAdjustment = gyroValue * GYRO_ADJUSTMENT_P;
		double curSign = sign(currentPower);
		currentPower += gyroAdjustment;
		printf("currentPower: %f gyroAdjustment: %f\n", currentPower,
				gyroAdjustment);
		currentPower = bound(currentPower, minPower * curSign,
				maxPower * curSign);
	}
	return currentPower;
}

void Drive::setWeightComp(double comp) {
	weight_comp = comp;
}

void diffNormalize(double*a, double*b) {
	double va = *a;
	double vb = *b;
	double r = fabs(va - vb);
	if (r > 2) {
		if (va > vb) {
			va = 1;
			vb = -1;
		} else {
			va = -1;
			vb = 1;
		}

	} else if (va > vb) {
		if (va > 1) {
			vb = 1 - r;
			va = 1;
		} else if (vb < -1) {
			vb = -1;
			va = -1 + r;
		}
	} else {
		if (vb > 1) {
			va = 1 - r;
			vb = 1;
		} else if (va < -1) {
			va = -1;
			vb = -1 + r;
		}
	}
	*a = va;
	*b = vb;
}

void Drive::strafeCompensate() {
	centerPower = wantedXPower;
	leftPower = ((wantedYPower + wantedRPower));
	rightPower = ((wantedYPower - wantedRPower));

	if (fabs(wantedRPower) < 0.05 && fabs(wantedXPower) > 0.05
			&& fabs(wantedYPower) < 0.05) {
		if (isStrafing != true) {
			strafeAngle = Base.getPosition().rotation;
			strafeIntegral = 0;
		}
		isStrafing = true;

		double newStrafeTime = GetTime();
		double timeStep = newStrafeTime - lastStrafeTime;
		lastStrafeTime = newStrafeTime;

		double error = Base.getPosition().rotation - strafeAngle;

		strafeIntegral += error * timeStep;

		const double KP = 0.01;
		const double KI = 0.01;
		const double KFF = 0.13;

		strafeIntegral = bound(strafeIntegral, -20, 20);
		double compensation = KFF * wantedXPower + KP * error
				+ KI * strafeIntegral;
		rightPower += compensation;
		leftPower -= compensation;
		diffNormalize(&leftPower, &rightPower);
		printf(
				"wantedX %4.3f Y %4.3f R %4.3f left %4.3f right %4.3f err %4.3f int %4.3f comp %4.3f\n",
				wantedXPower, wantedYPower, wantedRPower, leftPower, rightPower,
				error, strafeIntegral, compensation);
	} else {
		isStrafing = false;
	}

//	double adjustment = scaleLinear(weight_comp, -1, 1, 0.18, -0.50)
//			* wantedXPower;
//	leftPower = ((wantedYPower + wantedRPower)) + adjustment;
//	rightPower = ((wantedYPower - wantedRPower)) - adjustment;
////	printf("L %f R %f\n", leftPower, rightPower);
//	diffNormalize(&leftPower, &rightPower);

//	printf("X %4.3f Y %4.3f R %4.3f a %4.3f -> L %4.3f R %4.3f C %4.3f\n", wantedXPower,
//			wantedYPower, wantedRPower, weight_comp, leftPower, rightPower, centerPower);

//	Position newPosition = Base.getPosition();
//	double newTime = GetTime();
//	double rotationSpeed = (newPosition.rotation - lastPosition.rotation)
//			/ (newTime - lastTime);
//	lastTime = newTime;
//	lastPosition = newPosition;
//
//	printf("wantedR %f gyroR %f\n", wantedRPower, rotationSpeed);
}
double Drive::autoDeccel(double motorPower, double maxPower, double minPower) {
	if (motorPower == 0) {
		return 0;
	}
	double sign = motorPower / fabs(motorPower);
	/*	double distAverage =
	 (Base.getLeftDriveTravel() + Base.getRightDriveTravel()) / 2;
	 double remainingDist = (sign * distGoal) - distAverage;
	 double newPower = ((motorPower / DECCELERATION_DIST) * remainingDist)
	 + 0.05; //TODO make better / clean up. Is hack
	 printf("newPower: %f \n", newPower);
	 if (fabs(newPower) > motorPower) {
	 newPower = 0;
	 }*/
	return sign * 0.2;
}
bool Drive::isDropWheelDown() {
	if (isCenterWheelDown == true) {
		return true;

	} else {
		return false;
	}
}

bool Drive::isShiftedDown() {
	return lastShiftPos == GearPosition::kLowGear;
}
