#pragma once

#include "systembase.h"

/**
 * @brief Represents the drivebase and provides control methods for it
 *
 * This class is responsible for commanding the drive motors, shifter, and
 * center wheel. It includes autonomous drive routines, which will use
 * information from the drivebase sensors.
 *
 * @see HardwareAccess
 * @see Controls
 *
 */
class Drive: public SystemBase {
public:
	// TODO document class interface

	/**
	 * @brief Two positions for the shifter
	 *
	 * high gear is fast,
	 * low gear is slow
	 */
	enum class GearPosition {
		kHighGear, kLowGear
	};
	/**
	 * @brief Two positions of the wheels
	 *
	 * Wheel up means no move
	 * Wheel down means maybe move if wheel is move
	 */
	enum class MiddleWheelPosition {
		kWheelDown, kWheelUp
	};
	bool MiddleWheelUp();

	Drive();
	~Drive();

	void initialize() override;
	void process() override;
	void debug(Logger&) override;

	/**
	 * @brief %Controls the Power and direction of the robot's side and middle wheel
	 *
	 * @param leftPower 	The power for the left side wheels,
	 * 		scales from -1(max backward power) to 1(max forward power).
	 * @param rightPower 	Power for the right side wheels,
	 * 		scales from -1(max backward power) to 1(max forward power).
	 * @param middlePower 	The power for the perpendicular middle wheel,
	 * 		scales from -1(max leftward power) to 1(max rightward power).
	 * @param position 		Takes an enum value.
	 * 		down = kWheelDown, up = kWheelUp
	 *
	 * so as to better keep the robots stack from tipping.
	 * Make sure the middle wheel is not running if its not down
	 *
	 */
	void slideDrive(double leftPower, double rightPower, double middlePower,
			MiddleWheelPosition position);

	/**
	 * @brief Method to control the forward and backward movement in Auto mode
	 *
	 * @param distance	Inputs the distance you would like to go in cm
	 * 		- numbers are backward, + numbers are forward
	 * @param maxPower the maximum power the robot should move
	 * 		only positive values
	 *
	 * Accelerate when slowing down and speeding up
	 * Make use of the accelerometers, gyro, and encoders to make sure you will always be going straight
	 * Only used in Auto
	 */
	void autonDriveStraight(double distance, double maxPower);

	void autonRCDriveBack(double distance, double maxPower);
	/**
	 * @brief Method to control the turning of the bot in Auto mode
	 *
	 * Moves the robot around a semi-circle(of a certain radius and for a certain amount of degrees), achieving what some experts call a "turn"
	 *
	 * @param turnAngle	in degrees, the amount of the circle you want to move on.
	 * 		how many degrees away from your current position will you be facing
	 * 		-360 or 360 brings you back to where you started
	 * 		-numbers turns you left, +numbers turns you right
	 * @param radius	the distance from the robot to the point your turning about in cm,
	 * 		the radius of the circle you are following
	 * 		- numbers will make the robot turn backwards
	 * 		0 means you will do a "point" turn meaning no movement other than the turn happens
	 *
	 */
	void autonDriveTurn(double turnAngle, double radius);

	/**
	 * @brief method of controlling the middle wheel during autonomous
	 *
	 * @param time how far you want to strafe
	 *
	 * @param maxPower power that the wheel should be at at its max
	 * 		positive is right
	 * 		negative is left
	 *
	 */
	void autonStrafe(double time, double maxPower);

	/**
	 * @brief Detects when your current task for the robot during Auto mode is finished
	 *
	 * @return a boolean value
	 * 		True means the previous operation is done
	 * 		False means it is still in progress
	 *
	 * 	True for as long as it does not have a current command in progress
	 */
	bool autonDone();

	/**
	 * @brief sets the robot to do nothing/stop during auto mode
	 */
	void autoDoNothing();

	void backToManualControl();

	/**
	 * @brief %Controls the gear that the robot is on
	 *
	 * @param pos	the gear we would like to currently be on, kHighGear or kLowgear, which are High and Low gear respectively
	 *
	 * must tell the robot to slow down for a small period of time while switching
	 */
	void shiftGears(GearPosition pos);

	void setWeightComp(double comp);


	bool isDropWheelDown();
	bool isShiftedDown();
private:
	HardwareAccess& Base;

	TimerW last_Shift_Timer; //keeps time since the last shift

	GearPosition lastShiftPos; //the position the shifter was set during the last shift

	double rightPower; //power for the right motors when not shifting
	double leftPower; // power for the left motors when not shifting
	double centerPower; //power for the center wheel always

	double distGoal;
	double powerMax;
	double distanceStartLeft;
	double distanceStartRight;
	bool doneAccelerating;
	double startingTime;
	double turnGoal;
	double turnAngleDirection;
	double turnBotDirection;

	double wantedXPower;
	double wantedYPower;
	double wantedRPower;
	Position lastPosition;
	double lastTime;

	double weight_comp;

	bool isStrafing;
	double strafeAngle;
	double strafeIntegral;
	double lastStrafeTime;

	bool isCenterWheelDown;
	enum {
		AutoStraight, AutoTurn, AutoNoRequest, AutoBroke, Manual, AutoSlide, AutoRCStraight
	} autoState;

	double autoAccel(double motorPower, double maxPower, double minPower);

	double gyroStraighten(double currentPower, double maxPower, double minPower,
			double gyroValue);
	void strafeCompensate();

	double autoDeccel(double motorPower, double maxPower, double minPower);

};
