#include "gripper.h"
#include "hardware.h"
#include "util/common.h"

#if 0
#define DEBUG(X...) printf(X)
#else
#define DEBUG(X...)
#endif

// TODO: limit over-clamping of the tote

const double OUTER_LIMIT_DEADBAND = .05;
const double INNER_LIMIT_DEADBAND = .07;
const double CAL_SPEED = 0.4;
const double GRIPPER_ALLIN = 0.99;
const double GRIPPER_ALLOUT = 0.01;
const double GRIPPER_NARROWG = 0.9;
const double GRIPPER_WIDEG = 0.1;
const double AUTO_SPEED = .4;
const double DEADBAND = .03;

const double INWARD_POWER = -0.8;
const double OUTWARD_POWER = 0.9;
const double INNER_CURRENT_LIMIT = 30.0;
const double OUTER_CURRENT_LIMIT = 33.0;
const double WINDOW_LENGTH = 0.15;

const double IN_SLOWDOWN_BAND = .2;
const double OUT_SLOWDOWN_BAND = .8;

const double INTAKE_WHEEL_SPEED = 0.7;

double TIME_LIMIT_WAS_TRIPPED = 0;
Gripper::Gripper() :
				SystemBase("Gripper"),
				Base(HardwareAccess::getInstance()) {
	sensorBroken = false;
	currentPower = 0;
	wasCaledBefore = false;
	intakePos = false;
	lastIntakePos = false;
	stallPowerLevel = 0;
	stallOccurred = false;
	stabilizersClosed = false;
}

Gripper::~Gripper() {

}

void Gripper::initialize() {
	//Don't know where arms are
	currentPower = 0;
	wasCaledBefore = false;
	current = 0;
	appliedPower = 0;
	stallOccurred = false;
	stallPowerLevel = 0;
	lastCurrents.clear();
	stabilizersClosed = false;
}

void Gripper::process() {
	if (!sensorBroken) {
		if (!Base.isGripperZeroed()) {
			//If we have not calibrated, move in correct direction until cal occurs
			wasCaledBefore = false;
			DEBUG("calibrating... %d\n", Base.isGripperHomeFlagHit());
			setCurrentControlledPower(
					Base.isGripperHomeFlagHit() ? CAL_SPEED : -CAL_SPEED);
			return;
		} else {
			if (!wasCaledBefore) {
				printf("gripper after cal: currentPower %f\n", currentPower);
			}
			wasCaledBefore = true;

			if ((Base.getGripperPosition() <= INNER_LIMIT_DEADBAND)
					&& (currentPower < 0)) {
				currentPower = 0;
			}
			if ((Base.getGripperPosition() >= (1 - OUTER_LIMIT_DEADBAND))
					&& (currentPower > 0)) {
				currentPower = 0;
			}
		}
	}
	setCurrentControlledPower(currentPower);
}

void Gripper::debug(Logger& logger) {
	logger.write(this, "Is zeroed: %c", Yn(Base.isGripperZeroed()));
	logger.write(this, "current power %5.2f applied power %5.2f", currentPower,
			appliedPower);
	logger.write(this, "broken switch set: %c", Yn(sensorBroken));
	logger.write(this, "tote present: %c", Yn(Base.isGripperBeamBreakClosed()));
	logger.write(this, "position: %6.4f velocity: %6.4f",
			Base.getGripperPosition(), Base.getGripperVelocity());
}

void Gripper::closeStabilizer(bool close) {
	if (close == false) {
		Base.extendStabilizerPiston(false);
	} else {
		Base.extendStabilizerPiston(true);
	}

}

void Gripper::setPositionSensorsBroken(bool broken) {
	sensorBroken = broken;
}

void Gripper::move(Direction dir) {
	switch (dir) {
	case Direction::kClose:
		currentPower = INWARD_POWER;
		break;
	case Direction::kOpen:
		currentPower = OUTWARD_POWER;
		break;
	case Direction::kStop:
		currentPower = 0;
		break;
	}
}

double Gripper::getPosition() {
	if (Base.isGripperZeroed() && !sensorBroken) {
		return Base.getGripperPosition();
	} else {
		return 0.0;
	}
}

bool Gripper::isTotePresent() {
	return Base.isGripperBeamBreakClosed();
}

void Gripper::setCurrentControlledPower(double power) {
	double currentCurrentLimit =
			(power > 0) ? OUTER_CURRENT_LIMIT : INNER_CURRENT_LIMIT;
	double newMotorCurrent = Base.getGripperMotorCurrent();
	Reading next;
	next.current = newMotorCurrent;
	double time = GetTime();
	next.time = GetTime();
	lastCurrents.push_back(next);
	while ((lastCurrents.size() > 0)
			&& (lastCurrents[0].time < (time - WINDOW_LENGTH))) {
		lastCurrents.pop_front();
	}

	double motorCurrent = 0;
	for (const Reading& r : lastCurrents) {
		motorCurrent += r.current;
	}
	motorCurrent /= lastCurrents.size();

	if (motorCurrent > currentCurrentLimit) {
//		appliedPower = 0;
		stallOccurred = true;
		stallPowerLevel = power;
	}

	DEBUG("ccr %f ccc %f mc %f so %d spl %f p %f t %f now %f last_app %f\n",
			newMotorCurrent, currentCurrentLimit, motorCurrent, stallOccurred,
			stallPowerLevel, power, TIME_LIMIT_WAS_TRIPPED, GetTime(),
			appliedPower);

	if (stallOccurred) {
		if ((stallPowerLevel * power) > 0.0) {
			if (TIME_LIMIT_WAS_TRIPPED == 0.0) {
				// we just hit the limit, so start a timer
				TIME_LIMIT_WAS_TRIPPED = GetTime();
				appliedPower = power;
			} else if ((GetTime() - TIME_LIMIT_WAS_TRIPPED) > 0.18) {
				// enough time has past since we hit the limit, to turn off power
				appliedPower = 0;
			} else {
				appliedPower = power;
			}
		} else {
			TIME_LIMIT_WAS_TRIPPED = 0;
			appliedPower = power;
			stallOccurred = false;
			printf("CLEAR STALL\n");
		}
	} else {
		TIME_LIMIT_WAS_TRIPPED = 0;
		appliedPower = power;
	}

//	printf("motorCurrent %f appliedPower %f\n", motorCurrent, appliedPower);
	Base.setGripperMotorPower(appliedPower);
}

bool Gripper::isStopped() {
	printf("Stall %d power %f\n", stallOccurred, currentPower);
	if (stallOccurred || currentPower == 0) {
		printf("isStopped\n");
		return true;
	} else {
		printf("Isn't stopped\n");
		return false;
	}
}
void Gripper::closeIntake(bool close) {
	if (close == true) {
		Base.extendIntakePiston(true);
	} else {
		Base.extendIntakePiston(false);
	}
}
void Gripper::spinWheels(WheelDirection direction) {
	switch (direction) {
	case WheelDirection::kInward:
		Base.setLeftIntakeMotorPower(-INTAKE_WHEEL_SPEED);
		Base.setRightIntakeMotorPower(INTAKE_WHEEL_SPEED);
		break;
	case WheelDirection::kOutward:
		Base.setLeftIntakeMotorPower(INTAKE_WHEEL_SPEED);
		Base.setRightIntakeMotorPower(-INTAKE_WHEEL_SPEED);
		break;
	case WheelDirection::kLeftSpin:
		Base.setLeftIntakeMotorPower(INTAKE_WHEEL_SPEED);
		Base.setRightIntakeMotorPower(INTAKE_WHEEL_SPEED);
		break;
	case WheelDirection::kRightSpin:
		Base.setLeftIntakeMotorPower(-INTAKE_WHEEL_SPEED);
		Base.setRightIntakeMotorPower(-INTAKE_WHEEL_SPEED);
		break;
	case WheelDirection::kStop:
		Base.setLeftIntakeMotorPower(0);
		Base.setRightIntakeMotorPower(0);
		break;
	}
}
