#pragma once

#include "systembase.h"
#include <deque>

/**
 * @brief %Controls the motion of the gripper (and the stabilizer)
 *
 * This class is responsible for commanding the gripper
 * to open and close the gripper claws. It also controls the stabilizer.
 *
 * It acts by calling functions like `Base.extendStabilizerPiston(...)` to
 * control the robot, and responds to "commands" coming from the Controls,
 * which calls methods like `move(...)` on an instance of this class.
 *
 * @see HardwareAccess
 * @see Controls
 */
class Gripper: public SystemBase {
public:
	enum class Direction {
		kOpen, kClose, kStop
	};
	enum class WheelDirection {
		kInward,
		kOutward,
		kStop,
		kLeftSpin,
		kRightSpin
	};

	Gripper();
	~Gripper();

	/**
	 * @brief Configure the starting gripper state
	 *
	 * In addition to setting the internal state for the gripper, this
	 * method may need to prepare for a homing procedure if the gripper
	 * sensors have not yet been homed.
	 *
	 * @copydoc SystemBase::initialize
	 */
	void initialize() override;
	void process() override;
	void debug(Logger&) override;

	/**
	 * @brief Close the stabilizer arms.
	 *
	 * @param close If this is true, the arms close. If false, they should open.
	 */
	void closeStabilizer(bool close);

	/**
	 * @brief Stop relying on position sensors
	 *
	 * Marks the position sensor feedback as broken or not broken.
	 *
	 * @warning This method may be called at any time with any value (even the same value as before).
	 *
	 * @param broken If this is true, the sensors are broken and *cannot* be trusted. If false, they work.
	 */
	void setPositionSensorsBroken(bool broken);

	/**
	 * @brief Command gripper to try to move with a given power
	 *
	 * If this method is called after goToPosition(), it will stop
	 * the position control started by it and apply power in the specified direction.
	 *
	 * The bounds for each edge of the gripper should be respected.
	 * If the gripper is at the outside edge, don't let it go further out.
	 * If it's on the inside edge, don't allow inward commands. But when the
	 * sensors are marked broken, ignore their readings, and *move anyway*.
	 *
	 * @warning This method may be called at any time with any value (even the same value as before).
	 *
	 * @param power The given power to move with. It is in the range [-1, 1].
	 *      +1 means to spin the motor to move out
	 *      -1 means to spin the motor to move in
	 *      0 means not to apply any power.
	 * @see Gripper#goToPosition()
	 * @see Gripper#setPositionSensorsBroken()
	 */
	void move(Direction dir);

	/**
	 * @brief Get the position of the gripper arms
	 *
	 * If the position sensors are broken, return what they provide,
	 * even though the values from them should not be trusted.
	 *
	 * @return A floating point number in the range [0,1].
	 *   0 means that the arms are fully in.
	 *   1 means that the arms are fully out.
	 *   0.30 means they are exactly 30% of the way from the inside to the outside.
	 */
	double getPosition();

	/**
	 * @brief Determine if a tote is being held/is between the arms.
	 *
	 * Reads robot sensors to determine if a tote is present.
	 *
	 * @return True if there is a tote, otherwise false.
	 */
	bool isTotePresent();

	/**
	 * @brief Determine if the gripper is not moving
	 *
	 * @return True if stopped, otherwise ... false.
	 */
	bool isStopped();
	/**
	 * @brief closes the intake
	 *
	 * @param close If true
	 *
	 */
	void closeIntake(bool close);
	/**
	 * @brief tells the intake wheels to move in a given direction
	 *
	 * @param d power is given to both motors to move them forward or backwards.
	 */
	void spinWheels(WheelDirection d);
private:
	void setCurrentControlledPower(double power);

	HardwareAccess& Base;
	double currentPower;
	bool sensorBroken;
	bool wasCaledBefore;
	bool stallOccurred;
	double stallPowerLevel;
	bool intakePos;
	bool lastIntakePos;

	double current;
	double lastPower;
	double appliedPower;
	typedef struct {
		double current;
		double time;
	} Reading;
	std::deque<Reading> lastCurrents;

	bool stabilizersClosed;

};
