#include <systembase.h>
#include "util/common.h"
#include "hardware.h"
#include "hardware_p.h"
#include "ControllerPower.h"

#define NOSET false

#define SLAVE_MOTORS false

#define CALIBRATE_ONLY_ONE_FIRST_CYCLE_ENABLE true

#define UPDATE_ONLY_ON_CHANGE false

// PARAMETERS

const double DRIVE_WHEEL_DIAM = 10.16;
const double R256TICKS = (double) 256;
const double GEAR_AMP = (36.0 / 12.0) * (54.0 / 30.0);
const double CM_PER_DRIVE_TICK = (DRIVE_WHEEL_DIAM * M_PI)
		/ (R256TICKS * GEAR_AMP);
const double LIFTER_DISTANCE_BETWEEN_FLAGS = 1715.0;

const double GRIPPER_FLAG_ERROR_ALLOWANCE = 30.0; // ticks
#ifdef ROBOT_2
const double GRIPPER_IN_POS = -4350;
const double GRIPPER_OUT_POS = 5550;
#else
const double GRIPPER_IN_POS = -3442;
const double GRIPPER_OUT_POS = 6050;
#endif

const double GRIPPER_IMPULSE_TIME = .200;
const double GRIPPER_IMPULSE_BOOST = .100;

// IOMAP aux parameters
const uint8_t PCM_ID = 15;
const uint32_t DIO_MXP_START = 10;

const uint8_t CAN_CONTROL_RATE = 10;
const uint8_t CAN_READOUT_RATE = 200;
const uint8_t CAN_NEVER_RATE = 255;
const uint8_t CAN_FEEDBACK_RATE = 20;

CachedTalon::CachedTalon(uint32_t id) :
				CanTalonSRX(id, CAN_CONTROL_RATE) {
	lastCurrent = 0.0;
	lastComm = 0.0;
	expComm = 0.0;
}
void CachedTalon::updateCurrent() {
	this->GetCurrent(lastCurrent);
}
double CachedTalon::getLastCurrent() {
	return lastCurrent;
}
void CachedTalon::setUpdateRates(CANTalonRates r) {
	this->SetStatusFrameRate(kStatusFrame_General, r.general);
	this->SetStatusFrameRate(kStatusFrame_AnalogTempVbat, r.analogTempVbat);
	this->SetStatusFrameRate(kStatusFrame_Feedback, r.feedback);
	this->SetStatusFrameRate(kStatusFrame_Encoder, r.encoder);
}
void CachedTalon::SetExpectedCommand(double setpoint) {
	expComm = setpoint;
}
void CachedTalon::ActuallyDoExpectedCommand() {
	if (expComm != lastComm) {
		lastComm = expComm;
#if UPDATE_ONLY_ON_CHANGE
		Set(expComm);
	}
#else
    }
	Set(expComm);
#endif
}

CachedSolenoid::CachedSolenoid(uint8_t moduleNumber, uint32_t channel) :
				Solenoid(moduleNumber, channel) {
	lastState = false;
}
void CachedSolenoid::SetOnlyIfChanged(bool on) {
	if (on != lastState) {
		lastState = on;
#if UPDATE_ONLY_ON_CHANGE
		Set(on);
    }
#else
    }
	Set(on);
#endif
}

HardwareAccess::Sensors::Sensors(HardwareAccess* p) :
				SystemBase("Sensors") {
	this->p = p;
}
void HardwareAccess::Sensors::initialize() {
	this->p->initializeSensors();
}
void HardwareAccess::Sensors::debug(Logger& logger) {
	this->p->debugSensors(logger, this);
}
void HardwareAccess::Sensors::process() {
	this->p->processSensors();
}
HardwareAccess::Motors::Motors(HardwareAccess* p) :
				SystemBase("Motors") {
	this->p = p;
}
void HardwareAccess::Motors::initialize() {
	this->p->initializeMotors();
}
void HardwareAccess::Motors::debug(Logger& logger) {
}
void HardwareAccess::Motors::process() {
	this->p->processMotors();
}

HardwareAccessHidden::HardwareAccessHidden() :
				panel(),

				/// Do not remove following comment
				/// [IOMap]
				solenoidShifter(PCM_ID, 0),
				solenoidDrop(PCM_ID, 1),
				solenoidStabilizer(PCM_ID, 2),
				solenoidBrake(PCM_ID, 3),
				solenoidIntake(PCM_ID, 4),

				motorLeftFront(2),
				motorLeftBack(3),
				motorRightFront(4),
				motorRightBack(5),
				motorCenter(6),
				motorLiftLeft(7),
				motorLiftRight(8),
				motorGripper(9),
				motorIntakeLeft(10),
				motorIntakeRight(11),
				motorRCPuller(12),

				analogInGyroRate(0),

				diDriveEncoderLeftA(0),
				diDriveEncoderLeftB(1),
				diDriveEncoderRightA(2),
				diDriveEncoderRightB(3),
				diLifterEncoderA(4),
				diLifterEncoderB(5),
				diLifterFlagBottom(6),
				diGripperEncoderA(7),
				diGripperEncoderB(8),
				diGripperFlagA(9),

				diGripperBannerDetection(DIO_MXP_START + 0),
				diLifterFlagTop(DIO_MXP_START + 1),
				diGripperFlagB(DIO_MXP_START + 2),

				rawAccelerometer(),
				/// [IOMap]
				/// Do not remove preceding comment

#if ONE_CHANNEL_DRIVE
				encoderDriveLeft(&diDriveEncoderLeftA),
				encoderDriveRight(&diDriveEncoderRightA),
#else
				encoderDriveLeft(&diDriveEncoderLeftA, &diDriveEncoderLeftB),
				encoderDriveRight(&diDriveEncoderRightA, &diDriveEncoderRightB),
#endif
				encoderLifter(&diLifterEncoderA, &diLifterEncoderB),
				encoderGripper(&diGripperEncoderA, &diGripperEncoderB),

				driveGyro(&analogInGyroRate),
				driveAccelerometer(&rawAccelerometer) {
}
HardwareAccessHidden::~HardwareAccessHidden() {

}

HardwareAccess::Sensors* HardwareAccess::getSensorSystem() {
	return &sensorAccess;
}
HardwareAccess::Motors* HardwareAccess::getMotorSystem() {
	return &motorAccess;
}

HardwareAccess::HardwareAccess() :
				base(new HardwareAccessHidden()),
				sensorAccess(this),
				motorAccess(this) {
	const uint8_t READOUT = CAN_READOUT_RATE;
	const uint8_t NEVER = CAN_NEVER_RATE;
	const uint8_t FEEDBACK = CAN_FEEDBACK_RATE;

	CANTalonRates driveMotorRates;
	CANTalonRates liftMotorRates;
	CANTalonRates gripperMotorRates;

	driveMotorRates.general = NEVER;
	driveMotorRates.analogTempVbat = NEVER;
	driveMotorRates.encoder = NEVER;
	driveMotorRates.feedback = READOUT;

	liftMotorRates.general = NEVER;
	liftMotorRates.analogTempVbat = NEVER;
	liftMotorRates.encoder = NEVER;
	liftMotorRates.feedback = READOUT;

	gripperMotorRates.general = NEVER;
	gripperMotorRates.analogTempVbat = NEVER;
	gripperMotorRates.encoder = NEVER;
	gripperMotorRates.feedback = FEEDBACK;

	// set only
	base->motorCenter.setUpdateRates(driveMotorRates);
	base->motorLeftFront.setUpdateRates(driveMotorRates);
	base->motorLeftBack.setUpdateRates(driveMotorRates);
	base->motorRightFront.setUpdateRates(driveMotorRates);
	base->motorRightBack.setUpdateRates(driveMotorRates);
	// casual
	base->motorLiftLeft.setUpdateRates(liftMotorRates);
	base->motorLiftRight.setUpdateRates(liftMotorRates);
	// current controlled
	base->motorGripper.setUpdateRates(gripperMotorRates);
	base->motorIntakeLeft.setUpdateRates(gripperMotorRates);
	base->motorIntakeRight.setUpdateRates(gripperMotorRates);
	base->motorRCPuller.setUpdateRates(gripperMotorRates);

#if SLAVE_MOTORS
	base->motorLeftFront.SetStatusFrameRate(CanTalonSRX::kStatusFrame_General, FEEDBACK);
	base->motorLeftBack.SetModeSelect(CanTalonSRX::kMode_SlaveFollower);
	base->motorLeftBack.SetDemand(base->motorLeftFront.GetDeviceNumber());
#endif
	base->encoderDriveLeft.SetReverseDirection(false);
	base->encoderDriveRight.SetReverseDirection(true);
	base->encoderGripper.SetReverseDirection(false);
	base->encoderLifter.SetReverseDirection(true);
	lifterZeroCount = 0;

	// set lifter/gripper to not be zeroed yet
	hasLifterBeenZeroed = false;
	hasGripperBeenZeroed = false;
	bFlagEdgePos = 0;

	lastLeftPower = 0;
	lastRightPower = 0;

	lastGripperPosition = 0;
}
HardwareAccess::~HardwareAccess() {
	delete base;
}

void HardwareAccess::initializeSensors() {
	bool shouldInitialize = true;
#if CALIBRATE_ONLY_ONE_FIRST_CYCLE_ENABLE
	if (RobotState.IsDisabled() || RobotState.IsOperatorControl()) {
		shouldInitialize = false;
	}
#endif
	printf("\n");
	if (shouldInitialize || !hasLifterBeenZeroed) {
		// initialize lifter zero
		lastLowerLimitState = isLifterLowerLimitHit();
		hasLifterBeenZeroed = false;
		if (isLifterLowerLimitHit()) {
			hasLifterBeenZeroed = true;
			base->encoderLifter.Reset();
		}
		lifterZeroCount = base->encoderLifter.Get();
		printf("Lifter sensors WILL need calibration\n");
	} else {
		printf("Lifter sensors do NOT need calibration\n");
	}

	if (shouldInitialize || !hasGripperBeenZeroed) {
		// initialize gripper zero
		lastGripperPosition = base->encoderGripper.Get();
		lastGripperAState = isGripperAFlagHit();
		lastGripperBState = isGripperBFlagHit();
		hasGripperBeenZeroed = false;
		bFlagEdgePos = 0;
		printf("Gripper sensors WILL need calibration\n");
	} else {
		printf("Gripper sensors do NOT need calibration\n");
	}
	printf("\n");

	// miscellaneous
	lastGripperOnsetTime.Reset();
	base->encoderDriveLeft.Reset();
	base->encoderDriveRight.Reset();

	brakeReleaseTime = 0.0;
	loadEstimate = 0.0;
	lastLifterVelocity = 0.0;

	lastLeftPower = 0.0;
	lastRightPower = 0.0;
}
void HardwareAccess::initializeMotors() {
	base->motorCenter.SetExpectedCommand(0);
	base->motorGripper.SetExpectedCommand(0);
	base->motorIntakeLeft.SetExpectedCommand(0);
	base->motorIntakeRight.SetExpectedCommand(0);
	base->motorLeftBack.SetExpectedCommand(0);
	base->motorLeftFront.SetExpectedCommand(0);
	base->motorLiftLeft.SetExpectedCommand(0);
	base->motorLiftRight.SetExpectedCommand(0);
	base->motorRCPuller.SetExpectedCommand(0);
	base->motorRightBack.SetExpectedCommand(0);
	base->motorRightFront.SetExpectedCommand(0);

	wasBrakeExtended = false;
}

void HardwareAccess::processSensors() {
	if (base->driveGyro.isDisconnected()) {
		std::cerr << "GYRO DONE GONE UNPLUGGED! (Voltage < 0.5)" << std::endl;
	}

#if SLAVE_MOTORS
	base->motorLeftBack.SetDemand(base->motorRightFront.GetDeviceNumber());
	base->motorLeftBack.SetModeSelect(CanTalonSRX::kMode_SlaveFollower);
#endif
#if !NOSET
	base->motorCenter.updateCurrent();
	base->motorGripper.updateCurrent();
	base->motorLeftBack.updateCurrent();
	base->motorLeftFront.updateCurrent();
	base->motorLiftLeft.updateCurrent();
	base->motorLiftRight.updateCurrent();
	base->motorRightBack.updateCurrent();
	base->motorRightFront.updateCurrent();
#endif
	bool value = isLifterLowerLimitHit();
	if (value != lastLowerLimitState) {
		lifterZeroCount = base->encoderLifter.Get();
		hasLifterBeenZeroed = true;
		printf("Lifter cal at %d\n", lifterZeroCount);
	}
	lastLowerLimitState = value;

	bool newGripperAState = isGripperAFlagHit();
	bool newGripperBState = isGripperBFlagHit();
	int newGripperPosition = base->encoderGripper.Get();

	if (newGripperAState != lastGripperAState) {
		printf("Gripper A cal\n");
		base->encoderGripper.Reset();
		hasGripperBeenZeroed = true;
	}
	if (newGripperBState != lastGripperBState) {
		printf("Gripper B cal\n");
		bFlagEdgePos = newGripperPosition;
	}
	lastGripperAState = newGripperAState;
	lastGripperBState = newGripperBState;
	lastGripperPosition = newGripperPosition;

#if ONE_CHANNEL_DRIVE
	base->encoderDriveLeft.SetLastAppliedPower(lastLeftPower);
	base->encoderDriveRight.SetLastAppliedPower(lastRightPower);
#endif

	double newVelocity = 1.0 / base->encoderLifter.GetPeriod()
			/ LIFTER_DISTANCE_BETWEEN_FLAGS;
	double newTime = GetTime();
	double acc = (lastLifterVelocity - newVelocity) / (newTime - lastTime);
	lastTime = newTime;
	lastLifterVelocity = newVelocity;

	double netCurrent = base->motorLiftLeft.getLastCurrent()
			- base->motorLiftRight.getLastCurrent();
	double load;
	if (acc == 0) {
		load = 0;
	} else {
		load = netCurrent / acc;
	}

//	printf("acc %f currr %f load %f brtdelta %f\n", acc, netCurrent, load, brakeReleaseTime - GetTime());
	if (brakeReleaseTime < GetTime() && lastLifterVelocity > 0) {
		loadEstimate = load;
		// all occurs in segments
	}
}

void HardwareAccess::processMotors() {
	double appliedPower;
	if (lastGripperOnsetTime.Get() <= GRIPPER_IMPULSE_TIME) {

		if (lastGripperPower == 0) {
			appliedPower = 0;
		} else {
			double powerSign = sign(lastGripperPower);
			appliedPower = bound(
					lastGripperPower + (GRIPPER_IMPULSE_BOOST * powerSign), -1,
					1);
		}
	} else {
		appliedPower = lastGripperPower;
	}
#if !NOSET
	base->motorGripper.SetExpectedCommand(-appliedPower);
#endif

	base->motorCenter.ActuallyDoExpectedCommand();
	base->motorGripper.ActuallyDoExpectedCommand();
	base->motorIntakeLeft.ActuallyDoExpectedCommand();
	base->motorIntakeRight.ActuallyDoExpectedCommand();
	base->motorLeftBack.ActuallyDoExpectedCommand();
	base->motorLeftFront.ActuallyDoExpectedCommand();
	base->motorLiftLeft.ActuallyDoExpectedCommand();
	base->motorLiftRight.ActuallyDoExpectedCommand();
	base->motorRCPuller.ActuallyDoExpectedCommand();
	base->motorRightBack.ActuallyDoExpectedCommand();
	base->motorRightFront.ActuallyDoExpectedCommand();
}

void HardwareAccess::debugSensors(Logger& logger, SystemBase* sys) {
	logger.write(sys, "Enc: DL %d, DR %d, L %d (%d), G %d",
			base->encoderDriveLeft.Get(), base->encoderDriveRight.Get(),
			getAdjustedLifterCount(), base->encoderLifter.Get(),
			base->encoderGripper.Get());
	logger.write(sys, "Enc: Gripper A %d B %d Lifter A %d B %d",
			base->diGripperEncoderA.Get(), base->diGripperEncoderB.Get(),
			base->diLifterEncoderA.Get(), base->diLifterEncoderB.Get());
	logger.write(sys, "MotC: RF % 4.2f RB % 4.2f LF % 4.2f LB % 4.2f",
			base->motorRightFront.getLastCurrent(),
			base->motorRightBack.getLastCurrent(),
			base->motorLeftFront.getLastCurrent(),
			base->motorLeftBack.getLastCurrent());
	ADW22307::Reading r = base->driveGyro.get();
	logger.write(sys, "Gyr: R % 4.2f > % 5.2f (% 4.2f)", r.rateRawVoltage,
			r.angleEstimate, r.angleEstimateError);
	logger.write(sys, "Lft: last %c bot %c top %c zero? %c",
			Yn(lastLowerLimitState), Yn(isLifterLowerLimitHit()),
			Yn(isLifterUpperLimitHit()), Yn(hasLifterBeenZeroed));
	logger.write(sys, "Grp: last pos %d = %3.2f A %c B %c", lastGripperPosition,
			getGripperPosition(), Yn(lastGripperAState), Yn(lastGripperBState));
	logger.write(sys, "Grp: B loc: %d zero? %c %s", bFlagEdgePos,
			Yn(hasGripperBeenZeroed),
			(fabs(bFlagEdgePos) < GRIPPER_FLAG_ERROR_ALLOWANCE
					|| !hasGripperBeenZeroed) ? "matched" : "FIXME!");
	logger.write(sys, "Lft: pos %5.4f vel %5.4f", getLifterPosition(),
			getLifterVelocity());

	IntegratedAccelerometer::AllResults acc =
			base->driveAccelerometer.GetComponents();
	logger.write(sys, "Acc X P %+.3f V %+.3f A %+.3f", acc.right.position,
			acc.right.velocity, acc.right.acceleration);
	logger.write(sys, "Acc Y P %+.3f V %+.3f A %+.3f", acc.forward.position,
			acc.forward.velocity, acc.forward.acceleration);
	logger.write(sys, "Acc Z P %+.3f V %+.3f A %+.3f", acc.up.position,
			acc.up.velocity, acc.up.acceleration);

	logger.write(sys, "3.3v on %c err %d volt %3.2f cur %3.2f",
			Yn(ControllerPower::GetEnabled3V3()),
			ControllerPower::GetFaultCount3V3(),
			ControllerPower::GetVoltage3V3(), ControllerPower::GetCurrent3V3());
	logger.write(sys, "5v   on %c err %d volt %3.2f cur %3.2f",
			Yn(ControllerPower::GetEnabled5V()),
			ControllerPower::GetFaultCount5V(), ControllerPower::GetVoltage5V(),
			ControllerPower::GetCurrent5V());
	logger.write(sys, "6v   on %c err %d volt %3.2f cur %3.2f",
			Yn(ControllerPower::GetEnabled6V()),
			ControllerPower::GetFaultCount6V(), ControllerPower::GetVoltage6V(),
			ControllerPower::GetCurrent6V());
	logger.write(sys, "Load estimate: %f", loadEstimate);
}

HardwareAccess& HardwareAccess::getInstance() {
	static HardwareAccess* hwa = nullptr;
	if (!hwa) {
		hwa = new HardwareAccess();
	}
	return *hwa;
}

//////////////////////////////////////////////////////////////
//                         Drive                            //
//////////////////////////////////////////////////////////////

void HardwareAccess::setLeftDriveMotorPower(double power) {
	lastLeftPower = power;
#if !NOSET
#if !SLAVE_MOTORS
	base->motorLeftBack.SetExpectedCommand(-power);
#endif
	base->motorLeftFront.SetExpectedCommand(-power);
#endif
}

void HardwareAccess::setRightDriveMotorPower(double power) {
	lastRightPower = power;
#if !NOSET
	base->motorRightBack.SetExpectedCommand(power);
	base->motorRightFront.SetExpectedCommand(power);
#endif
}

void HardwareAccess::setCenterWheelPower(double power) {
#if !NOSET
	base->motorCenter.SetExpectedCommand(-power);
#endif
}

double HardwareAccess::getLeftDriveTravel() {
	return base->encoderDriveLeft.Get() * CM_PER_DRIVE_TICK;
}

double HardwareAccess::getRightDriveTravel() {
	return base->encoderDriveRight.Get() * CM_PER_DRIVE_TICK;
}

void HardwareAccess::extendDropPiston(bool extend) {
	base->solenoidDrop.SetOnlyIfChanged(extend);
}

void HardwareAccess::extendShifterPistons(bool extend) {
	base->solenoidShifter.SetOnlyIfChanged(!extend);
}

double HardwareAccess::getCenterWheelTravel() {
	std::cerr << "DON'T USE THIS METHOD: " << __FUNCTION__ << " at " << __LINE__
			<< " in " << __FILE__ << std::endl;
	return NAN;
}

void HardwareAccess::beginPositionSensorCalibration() {
	base->driveGyro.beginCalibration();
	base->driveAccelerometer.BeginCalibration();
}

void HardwareAccess::endPositionSensorCalibration() {
	base->driveGyro.finishCalibration();
	base->driveAccelerometer.FinishCalibration();
}

void HardwareAccess::zeroPosition() {
	base->driveGyro.reset();
	base->driveAccelerometer.Reset();
}

Position HardwareAccess::getPosition() {
	ADW22307::Reading r = base->driveGyro.get();
	Position p;
	p.xPosition = 0;
	p.yPosition = 0;
	p.rotation = r.angleEstimate;

	return p;
}

Position HardwareAccess::getPositionStdError() {
	return {0,0,0};
}

//////////////////////////////////////////////////////////////
//                       Gripper                            //
//////////////////////////////////////////////////////////////

void HardwareAccess::setGripperMotorPower(double power) {
	if (lastGripperPower == 0 && power != 0) {
		lastGripperOnsetTime.Reset();
	}

	lastGripperPower = power;
}

double HardwareAccess::getGripperMotorCurrent() {
	return base->motorGripper.getLastCurrent();
}

double HardwareAccess::getGripperPosition() {
	return scaleLinear(base->encoderGripper.Get(), GRIPPER_IN_POS,
			GRIPPER_OUT_POS, 0.0, 1.0);
}

double HardwareAccess::getGripperVelocity() {
	return base->encoderGripper.GetRate() / (GRIPPER_OUT_POS - GRIPPER_IN_POS);
}

bool HardwareAccess::isGripperHomeFlagHit() {
	return isGripperAFlagHit();
}

bool HardwareAccess::isGripperBeamBreakClosed() {
	return !base->diGripperBannerDetection.Get();
}

bool HardwareAccess::isGripperZeroed() {
	return hasGripperBeenZeroed;
}

void HardwareAccess::extendStabilizerPiston(bool extend) {
	base->solenoidStabilizer.SetOnlyIfChanged(extend);
}

bool HardwareAccess::isGripperAFlagHit() {
	return base->diGripperFlagA.Get();

}
bool HardwareAccess::isGripperBFlagHit() {
	return base->diGripperFlagB.Get();
}

void HardwareAccess::extendIntakePiston(bool extend) {
	base->solenoidIntake.SetOnlyIfChanged(extend);
}

void HardwareAccess::setLeftIntakeMotorPower(double power) {
	base->motorIntakeLeft.SetExpectedCommand(power);
}

void HardwareAccess::setRightIntakeMotorPower(double power) {
	base->motorIntakeRight.SetExpectedCommand(power);
}

//////////////////////////////////////////////////////////////
//                        Lifter                            //
//////////////////////////////////////////////////////////////

void HardwareAccess::setLifterMotorPower(double power) {
#if !NOSET
	base->motorLiftLeft.SetExpectedCommand(-power);
	base->motorLiftRight.SetExpectedCommand(power);
#endif
}

double HardwareAccess::getLifterPosition() {
	double height = getAdjustedLifterCount();
	return height / LIFTER_DISTANCE_BETWEEN_FLAGS;
}

double HardwareAccess::getLifterVelocity() {
	return base->encoderLifter.GetRate() / LIFTER_DISTANCE_BETWEEN_FLAGS;
}

bool HardwareAccess::isLifterZeroed() {
	return hasLifterBeenZeroed;
}

bool HardwareAccess::isLifterUpperLimitHit() {
	return base->diLifterFlagTop.Get();
}

bool HardwareAccess::isLifterLowerLimitHit() {
	return base->diLifterFlagBottom.Get();
}

void HardwareAccess::extendBrakePiston(bool extend) {
	if (extend && !wasBrakeExtended) {
		brakeReleaseTime = GetTime() + 0.125;
	} else if (!extend && wasBrakeExtended) {
		brakeReleaseTime = GetTime() + 1e8;
	}
	wasBrakeExtended = extend;
	base->solenoidBrake.SetOnlyIfChanged(extend);
}
int HardwareAccess::getAdjustedLifterCount() {
	return base->encoderLifter.Get() - lifterZeroCount;
}

double HardwareAccess::getLifterAdditionalLoadEstimate() {
	return loadEstimate;
}

double HardwareAccess::getLifterAbsCurrent() {
	return fabs(base->motorLiftLeft.getLastCurrent())
			+ fabs(base->motorLiftRight.getLastCurrent());
}

//RC Puller

void HardwareAccess::setRCPullerMotorPower(double power) {
	base->motorRCPuller.SetExpectedCommand(power);
}
