#pragma once

#include "systembase.h"
#include "util/misc.h"

class HardwareAccessHidden;

/**
 * @brief Interface for the other subsystems with the robot hardware.
 *
 * This class is a thin wrapper on top of all the sensors and actuators of
 * the robot. It handles sensor inversion, value scaling,
 * motor synchronization, calibration, and channel locations for all parts.
 *
 * As such, it localizes the more physical aspects of the robot, leaving
 * logical control to the higher-level classes.
 *
 * @warning Every value or action that is set is persistent. In other words,
 * if you set a motor power to 0.5 at time T=0.1, and do nothing, the motor
 * will still output 0.5 times its full power at time T=1.2 and T=205.6. Disabling
 * the robot, of course, will stop the motor from moving. But once you reenable,
 * the motor/piston may continue to be active until you command it to stop.
 *
 * IOMap initialization code:
 * @snippet hardware.cpp IOMap
 *
 * @see Lifter
 * @see Gripper
 * @see Drive
 */
class HardwareAccess {
private:
	class Sensors;
	class Motors;
public:
	/**
	 * @brief Get a reference to the one instance of this class.
	 */
	static HardwareAccess& getInstance();
	Sensors* getSensorSystem();
	Motors* getMotorSystem();

	/**
	 * @name Drive
	 * These functions control the portion of the
	 * robot corresponding to the drivebase.
	 */
	/// @{
	/**
	 * @brief Set the power for the left set of drivetrain motors
	 *
	 * @param power The power to apply to each motor. The range is [-1,1], with
	 *      -1 corresponding to full power in reverse, +1 to full power forward,
	 *      and 0 to no power applied.
	 *
	 * @see #setRightDriveMotorPower()
	 */
	void setLeftDriveMotorPower(double power);

	/**
	 * @brief Set the power for the right set of drivetrain motors
	 *
	 * @param power The power to apply to each motor. The range is [-1,1], with
	 *      -1 corresponding to full power in reverse, +1 to full power forward,
	 *      and 0 to no power applied.
	 *
	 * @see #setLeftDriveMotorPower()
	 */
	void setRightDriveMotorPower(double power);

	/**
	 * @brief Set the power for motor driving the center wheel
	 *
	 * @param power The power to apply to the center wheel. Its range is [-1,1], with +0.5
	 *      corresponding to half power *right*, and -1 matching full power *left*.
	 *
	 * @see #setLeftDriveMotorPower()
	 * @see #setRightDriveMotorPower()
	 */
	void setCenterWheelPower(double power);

	/**
	 * @brief Returns the distance that the *left* set of wheels has actively traveled.
	 *
	 * If you want to move a distance, you need to take the difference between
	 * the value of the starting travel and the value of the ending travel. There is
	 * no zero() or reset() method.
	 *
	 * @warning These measurements correspond to an encoder on a gearbox. There may be slop,
	 * and slipping of the wheels that will not be accounted for.
	 *
	 * @return Distance in centimeters that the encoder on the *left* side of the
	 *         drivebase has measured, relative to an arbitrary point.
	 */
	double getLeftDriveTravel();
	/**
	 * @brief Returns the distance that the *right* set of wheels has actively traveled.
	 *
	 * If you want to move a distance, you need to take the difference between
	 * the value of the starting travel and the value of the ending travel. There is
	 * no zero() or reset() method.
	 *
	 * @warning These measurements correspond to an encoder on a gearbox. There may be mechanical slop
	 * and slipping of the wheels that will not be accounted for.
	 *
	 * @return Distance in centimeters that the encoder on the *right* side of the
	 *         drivebase has measured, relative to an arbitrary point.
	 */
	double getRightDriveTravel();
	/**
	 * @brief Extends the pistons on the center wheel module.
	 *
	 * @param extend If true, the pneumatic cylinder will actuate downwards; if false, it will retract.
	 */
	void extendDropPiston(bool extend);
	/**
	 * @brief Extends the pistons on the shifting transmissions.
	 *
	 * @param extend If true, the pneumatic cylinders will actuate out
	 *     (into the transmission bodies); if false, they will retract.
	 */
	void extendShifterPistons(bool extend);
	/**
	 * @brief Get the distance that the center wheel has actively traveled
	 *
	 * @warning **DO NOT USE THIS METHOD** because there possibly will not be a physical sensor
	 *     on the center wheel. It was only included for completeness, in case that sensor is included.
	 *
	 * @warning These measurements correspond to an encoder on a gearbox. There may be mechanical slop
	 * and slipping of the wheels that will not be accounted for.
	 *
	 * @return Distance in centimeters that the encoder on the center wheel has
	 *        measured, relative to an arbitrary point.
	 */
	double getCenterWheelTravel();

	/**
	 * @brief Start calibrating the position sensors (accelerometer, gyro).
	 *
	 * This should be done only when the robot will be in a fixed reference frame
	 * for an extended period of time, and no position values are needed.
	 *
	 * It should reset the sensors and reduce their error.
	 *
	 * @see #endPositionSensorCalibration()
	 */
	void beginPositionSensorCalibration();

	/**
	 * @brief Stop calibration of the position sensors (accelerometer, gyro).
	 *
	 * This also zeros the position estimate.
	 *
	 * @see #zeroPosition()
	 * @see #beginPositionSensorCalibration()
	 */
	void endPositionSensorCalibration();

	/**
	 * @brief Zero the current robot position
	 *
	 * All further position measurements will be delivered relative to this.
	 *
	 * @see #getPosition()
	 */
	void zeroPosition();
	/**
	 * @brief Get an estimate of the current robot position.
	 *
	 * Relative to the last position of zeroing.
	 *
	 * As time passes, the error on this increases because of gyro drift and
	 * physical limitations.
	 *
	 * @see getPositionStdError();
	 */
	Position getPosition();
	/**
	 * @brief Provide a 1-std deviation estimate of the error from #getPosition()
	 *
	 * @return A Position type, with values that don't correspond to real distances,
	 *   but to the estimated error in our calculation of those distances.
	 *
	 * @see #getPosition()
	 */
	Position getPositionStdError();

	/// @}

	/**
	 * @name Gripper
	 * These functions control the portion of the
	 * robot corresponding to the gripper and stabilizer.
	 *
	 * At core, the gripper is a linear slide. The sensors attached to that slide,
	 * however, cannot give a correct position estimate on the slide until they
	 * have been calibrated (or zeroed). There are (under the hood) two sensors:
	 * a home switch, and an encoder. The home switch tells you when you
	 * are at a specific point of the gripper, but nowhere else. The encoder tells
	 * you how far you have gone, but doesn't know where you started. To know where
	 * on the slide the arms are, the arms must be moved into the home switch (to
	 * establish a reference position); and after that, the encoder is used
	 * to extrapolate where exactly you are.
	 *
	 * You may wonder: how do we move the arms into the limit switch, if they can start
	 * on either side of it? A decent proxy is to slowly move the arms out until they hard
	 * stop and stall at the outer end of travel (the encoder stops registering movement),
	 * or until the sensor is homed (#isGripperZeroed() returns true). If the arms stop moving,
	 * despite applied outward force, then the arms should be moved in until the sensor is homed
	 * (same criterion).
	 */
	/// @{
	/**
	 * @brief Set the power applied to the gripper motor
	 *
	 * @param power The power to apply to the motor. The range is [-1,1]. +1 is full power
	 *     directed to *opening* the gripper; -1 should close it; 0 does nothing.
	 */
	void setGripperMotorPower(double power);
	/**
	 * @brief Return a current output estimate
	 *
	 * For DC motors, current is proportional to output torque.
	 *
	 * @return Value in range [0,+inf), where 0 is no current draw, 3 is
	 *     roughly free speed, and 80 is roughly stall current at full
	 *     applied voltage. The units are Amperes.
	 */
	double getGripperMotorCurrent();
	/**
	 * @brief Get an estimate of the gripper arms' location
	 *
	 * @warning The validity of this estimate depends on two things: that the sensors
	 *   on the gripper aren't broken, and that the gripper has been zeroed.
	 *
	 * @return A value in [0,1], with 0 being arms fully in and 1 being arms both at the
	 *    far outside range of travel.
	 *
	 * @see #isGripperZeroed()
	 * @see #getGripperVelocity()
	 */
	double getGripperPosition();
	/**
	 * @brief Return an estimate of the current velocity of the gripper
	 *
	 * @return The velocity in units of full range of travel per second.
	 *
	 * @see #getGripperPosition();
	 */
	double getGripperVelocity();
	/**
	 * @brief Determine if the arms trip the home flag
	 *
	 * @warning This does not represent the inner limit of travel. Talk to an electrical
	 *    member about this.
	 *
	 * @return True when the arms currently are at the home switch.
	 */
	bool isGripperHomeFlagHit();
	/**
	 * @brief Determine if an object is along the beam-break sensor
	 *
	 * @return True when the sensor records an object or other interruption
	 *     along the line of the sensor. False otherwise.
	 */
	bool isGripperBeamBreakClosed();
	/**
	 * @brief Determine if the gripper arm sensors have been zeroed yet
	 *
	 * The sensors on the gripper assembly must be calibrated before you
	 * can use them to find out where the gripper is. This requires that
	 * the gripper start at or be moved into one of the two limit switches
	 * at least once.
	 *
	 * If the gripper starts at a limit switch when you enable, it begins zeroed.
	 *
	 * @see #isGripperHomeFlagHit()
	 * @see #getGripperPosition()
	 *
	 * @return True if the gripper has been zeroed (meaning that you can
	 *      trust #getGripperPosition()); false otherwise.
	 */
	bool isGripperZeroed();
	/**
	 * @brief Extends the pistons on the stabilizer.
	 *
	 * @param extend If true, the pneumatic cylinders will actuate to
	 *    close the stabilizer; if false, it will retract.
	 */
	void extendStabilizerPiston(bool extend);
	/**
	 * @brief Extends the pistons on the intakes.
	 *
	 * @param extend If true, the pneumatic cylinder will actuate to
	 *    close the intake arms; if false, it will retract, and open them.
	 */
	void extendIntakePiston(bool extend);
	/**
	 * @brief Set the power for the left intake wheel
	 *
	 * 	 * @param power Power to apply, in range [-1,+1], with +1 spinning the wheel CLOCKWISE
	 */
	void setLeftIntakeMotorPower(double power);
	/**
	 * @brief Set the power for the right intake wheel
	 *
	 * @param power Power to apply, in range [-1,+1], with +1 spinning the wheel CLOCKWISE
	 */
	void setRightIntakeMotorPower(double power);
	/// @}

	/**
	 * @name Lifter
	 * These functions control the portion of the
	 * robot corresponding to the lifter.
	 *
	 * At core, the lifter is a linear slide. The sensors attached to that slide,
	 * however, cannot give a correct position estimate on the slide until they
	 * have been calibrated (or zeroed). There are (under the hood) three sensors:
	 * two limit switches, and an encoder. The limit switches tell you wwhen you
	 * are at a specific point, but not in between. The encoder tells you how
	 * far you have gone, but doesn't know where you started. To know where
	 * on the slide the arms are, the arms must be moved into a limit switch (to
	 * establish a reference position); and after that, the encoder is used
	 * to extrapolate where exactly you are.
	 */
	/// @{
	/**
	 * @brief Set the power applied to the lifter motors
	 *
	 * Note that you always have something resisting the action of the lifter motors:
	 * gravity. The more totes are loaded, the stronger the pull downwards. To hold the lifter
	 * in place, apply the brake.
	 *
	 * @param power The power to apply to the motors. The range is [-1,1]. +1 is full power
	 *     directed to lifting the elevator that holds the grippers; 0 applies no power.
	 *
	 * @see #extendBrakePiston()
	 */
	void setLifterMotorPower(double power);
	/**
	 * @brief Get an estimate of the lifter's height
	 *
	 * @warning The validity of this estimate depends on two things: that the sensors
	 *   on the gripper aren't broken, and that the gripper has been zeroed.
	 *
	 * @return A value in [0,1], with 0 at the bottom end of
	 *     travel, and 1 at the top end.
	 *
	 * @warning Subject to discontinuities when zeroing
	 *
	 * @see #isLifterZeroed()
	 */
	double getLifterPosition();
	/**
	 * @brief Get an estimate of the load force on the lifter motors
	 *
	 * @warning Does not update when the lifter is braked
	 *
	 * @return double Load force in Newtons
	 */
	double getLifterAdditionalLoadEstimate();
	/**
	 * @brief Estimate of lifter velocity
	 *
	 * @return A velocity, with units of X/s, where 1X=full travel range of lifter
	 */
	double getLifterVelocity();
	/**
	 * @brief Determine if the lifter elevator sensors have been zeroed yet
	 *
	 * The sensors on the lifter assembly must be calibrated before you
	 * can use them to find out where the elevator is. This requires that
	 * the lifter start at or be moved into one of the two limit switches
	 * at least once.
	 *
	 * If the lifter starts at a limit switch when you enable, it will begin zeroed.
	 *
	 * @see #isLifterUpperLimitHit()
	 * @see #isLifterLowerLimitHit()
	 * @see #getLifterPosition()
	 *
	 * @return True if the lifter has been zeroed (meaning that you can
	 *      trust #getLifterPosition()); false otherwise.
	 */
	bool isLifterZeroed();
	/**
	 * @brief Has the elevator carriage reached the upper limit switch?
	 *
	 * The bottom limit switch marks the top range of travel of the lifter.
	 * If this switch is genuinely tripped (sensors aren't broken), then moving
	 * further up than this will damage ... something.
	 *
	 * @return True when the arms currently are at or above the level of
	 *      the top limit switch. If not, false.
	 *
	 * @see #isLifterZeroed()
	 * @see #isLifterLowerLimitHit()
	 */
	bool isLifterUpperLimitHit();
	/**
	 * @brief Has the elevator carriage reached the lower limit switch?
	 *
	 * The bottom limit switch marks the bottom range of travel of the lifter.
	 * If this switch is genuinely tripped (sensors aren't broken), then moving
	 * further down than this will damage ... something.
	 *
	 * @return True when the arms currently are at or below the
	 *       level of the bottom limit switch. If not, false.
	 *
	 * @see #isLifterZeroed()
	 * @see #isLifterUpperLimitHit()
	 */
	bool isLifterLowerLimitHit();
	/**
	 * @brief Extends the piston on the brake.
	 *
	 * @param extend If true, the pneumatic cylinder will actuate to
	 *    release the brake; if false, it will retract, stopping
	 *    the motion of the lifter carriage.
	 */
	void extendBrakePiston(bool extend);
	/**
	 * @brief Get the absolute value of the current on the lifter motors
	 *
	 * @return A value in amps
	 */
	double getLifterAbsCurrent();
	/// @}

	/**
	 * @brief Set the power for the RC puller motor
	 *
	 * @param power Range [-1, 1], with +1 releasing the hooks forwards
	 */
	void setRCPullerMotorPower(double power);
private:
	class Sensors: public SystemBase {
	public:
		Sensors(HardwareAccess* p);
		void initialize() override;
		void debug(Logger&) override;
		void process() override;
	private:
		HardwareAccess* p;
	};
	class Motors: public SystemBase {
	public:
		Motors(HardwareAccess* p);
		void initialize() override;
		void debug(Logger&) override;
		void process() override;
	private:
		HardwareAccess* p;
	};
	/**
	 * @brief Single constructor for HardwareAccess. Function definition includes IO map
	 */
	~HardwareAccess();
	HardwareAccess();

	void initializeMotors();
	void processMotors();

	void initializeSensors();
	void processSensors();
	void debugSensors(Logger&, SystemBase*);

	bool isGripperAFlagHit();
	bool isGripperBFlagHit();
	int getAdjustedLifterCount();

	HardwareAccessHidden* base;
	Sensors sensorAccess;
	Motors motorAccess;

	bool lastLowerLimitState;
	bool hasLifterBeenZeroed;
	int lifterZeroCount;

	int lastGripperPosition;
	bool lastGripperAState;
	bool lastGripperBState;
	bool hasGripperBeenZeroed;
	int bFlagEdgePos;

	double lastLeftPower;
	double lastRightPower;

	double lastLifterVelocity;
	double lastTime;
	double loadEstimate;
	double brakeReleaseTime;

	double lastGripperPower;
	TimerW lastGripperOnsetTime;

	bool wasBrakeExtended;
};
