#include "lifter.h"
#include "util/common.h"
#include "hardware.h"

TuneableConstant(BOTTOM_POSITION, 0.00);
TuneableConstant(TOTE_ON_LEVEL_TWO_OR_DEFAULT_RC, 0.26);
TuneableConstant(PLACE_RC_ON_TOTE, 0.38);
TuneableConstant(CLEAR_SLOT_POSITION, 0.666);
TuneableConstant(TOP_POSITION, 0.99);
TuneableConstant(CLEAR_TOTE_POSITION, 0.20);
TuneableConstant(TWO_INCH_POSITION, 0.05);
TuneableConstant(AUTO_STACK_BOTTOM, 0.08);
TuneableConstant(AUTO_STACK_TOP, 0.59);
TuneableConstant(FEEDER_GRAB_POSITION, 0.09);
TuneableConstant(STACK_RC_AND_TOTE_ON_TOTE, 0.23);

TuneableConstant(DOWNWARD_CURRENT_CUTOFF, 40.5); // amps

const double POSITION_DEAD_BAND = 0.01;

const double CALIBRATION_POWER = -0.1;

const double BRAKE_RANGE = 0.025;
const double LOW_SLOWDOWN_BAND = .1;
const double HIGH_SLOWDOWN_BAND = .88;

// TODO Tune these 9 parameters ;-)
// and add something to account for gravity/load

// TODO: make more precise, by a) increasing loop rate
//       and b) controlling brake actuation -- we don't want
//       it until we are at setpoint; and c) rewriting profile
//       to mathematically allow non-zero starting and stopping speeds
//       Loop rate is the priority: may need to feed encoder into talon
TuneableConstant(PID_P, 0.5);
TuneableConstant(PID_I, 5.0);
TuneableConstant(PID_D, 0.0);
TuneableConstant(PID_F, 1.0);
TuneableConstant(PID_A, 0.1);
TuneableConstant(PID_I_CONTRIB_CAP, 0.3);

TuneableConstant(BUMP_VEL, 0.05);
TuneableConstant(MAX_V,0.75);
TuneableConstant(MAX_A,1.4);
TuneableConstant(MAX_J,5.5);

Lifter::Lifter() :
				SystemBase("Lifter"),
				Base(HardwareAccess::getInstance()),
				controller(0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
				profile(0.0, 0.0, 0.0, 0.0) {
	brokenSwitch = false;
	initialize();
}
Lifter::~Lifter() {
}

void Lifter::initialize() {
	Base.extendBrakePiston(false);
	Base.setLifterMotorPower(0.0);
	isJoyStickMode = true;
	targetLevel = 0;
	wasCalBefore = false;
	startingLevel = 0;
	overCurrentDown = false;
}

void Lifter::applyPowerWithAutoBrake(double power) {
	applyPowerWithManualBrake(power,
			power < BRAKE_RANGE && power > -BRAKE_RANGE);
}
void Lifter::applyPowerWithManualBrake(double power, bool brake) {
	setPower = power;
	double current = Base.getLifterAbsCurrent();
	if (power < 0 && current > DOWNWARD_CURRENT_CUTOFF.get()) {
		overCurrentDown = true;
	} else if (power >= 0) {
		overCurrentDown = false;
	}

	if (brake || overCurrentDown) {
		Base.setLifterMotorPower(0.0);
		Base.extendBrakePiston(false);
	} else {
		Base.setLifterMotorPower(setPower);
		Base.extendBrakePiston(true);
	}
}
void Lifter::process() {
	if (brokenSwitch == true) {
		applyPowerWithAutoBrake(setPower);
		return;
	}
	if (!Base.isLifterZeroed()) {
		applyPowerWithAutoBrake(CALIBRATION_POWER);
		wasCalBefore = false;
		return;
	}
	if (!wasCalBefore) {
		// set power to zero directly after calibration
		wasCalBefore = true;
		controller.Reset();
		applyPowerWithAutoBrake(0.0);
		return;
	}
	if (isJoyStickMode) {
		if ((Base.isLifterLowerLimitHit() == true) && (setPower < 0)) {
			setPower = 0.0;
		} else if (Base.getLifterPosition() <= LOW_SLOWDOWN_BAND) {
			setPower = bound(setPower, -.1, 1);
		}
		if ((Base.isLifterUpperLimitHit() == true) && (setPower > 0)) {
			setPower = 0.0;
		} else if (Base.getLifterPosition() >= HIGH_SLOWDOWN_BAND) {
			setPower = bound(setPower, -1, 0.25);
		}
		applyPowerWithAutoBrake(setPower);
	} else {
		double currentPos = Base.getLifterPosition();
		// TODO: Improve the S-curve profile generator to allow
		// for negative set points. Also, create a short boost power
		// to break static friction (once only, for a short time)
		if (targetLevel > startingLevel) {
			if (currentPos > targetLevel - POSITION_DEAD_BAND) {
				printf ("Brake1 on\n");
				applyPowerWithManualBrake(0.0, true);
			} else {
				CubicProfile::State target = profile.getState(
						fabs(targetLevel - currentPos));
				printf ("Brake1 off %f \n" , fabs(targetLevel - currentPos));
				applyPowerWithManualBrake(
						controller.getNextOutput(target.vel + 0.05,
								Base.getLifterVelocity(), target.acc), false);
			}
		} else {
			if (currentPos < targetLevel + POSITION_DEAD_BAND) {
				printf ("Brake2 on\n");
				applyPowerWithManualBrake(0.0, true);
			} else {
				CubicProfile::State target = profile.getState(
						fabs(targetLevel - currentPos));
				printf ("Brake2 off %f \n" , fabs(targetLevel - currentPos));
				applyPowerWithManualBrake(
						controller.getNextOutput(-target.vel - 0.05,
								Base.getLifterVelocity(), -target.acc), false);
			}
		}
	}
}

void Lifter::debug(Logger& logger) {
	logger.write(this, "Is zeroed: %c", Yn(Base.isLifterZeroed()));
	logger.write(this, "current power %f", setPower);
	logger.write(this, "broken switch set: %c", Yn(brokenSwitch));
	logger.write(this, "Bottom Flag: %c", Yn(Base.isLifterLowerLimitHit()));
	logger.write(this, "Top Flag: %c", Yn(Base.isLifterUpperLimitHit()));
	logger.write(this, "Lifter Position: %f AbsCurrent: %f",
			Base.getLifterPosition(), Base.getLifterAbsCurrent());
	logger.write(this, "Starting %f Target %f Vel %f", startingLevel,
			targetLevel,
			profile.getState(fabs(Base.getLifterPosition() - startingLevel)).vel);
}

void Lifter::setPositionSensorsBroken(bool broken) {
	brokenSwitch = broken;
}

double getPositionForLabel(Lifter::LifterPosition position) {
	typedef Lifter::LifterPosition LLP;
	if (position == LLP::kPickUpBottomTote) {
		return BOTTOM_POSITION.get();
	} else if (position == LLP::kClearToteWithRC) {
		return PLACE_RC_ON_TOTE.get();
	} else if (position == LLP::kTopOfChannel) {
		return TOP_POSITION.get();
	} else if (position == LLP::kPickUpSecondToteInStackOrDefaultRCGrabPosition) {
		return TOTE_ON_LEVEL_TWO_OR_DEFAULT_RC.get();
	} else if (position == LLP::k2inch) {
		return TWO_INCH_POSITION.get();
	} else if (position == LLP::kClearFeederSlot) {
		return CLEAR_SLOT_POSITION.get();
	} else if (position == LLP::kClearTote) {
		return CLEAR_TOTE_POSITION.get();
	} else if (position == LLP::kAutoStackBottom) {
		return AUTO_STACK_BOTTOM.get();
	}else if (position == LLP::kAutoStackTop) {
		return AUTO_STACK_TOP.get();
	} else if (position == LLP::kGrabFromFeeder) {
		return FEEDER_GRAB_POSITION.get();
	}else if (position == LLP::kPlaceToteWithRCOnToteToCreateTwoStackWithRCForAutoModeNine) {
		return STACK_RC_AND_TOTE_ON_TOTE.get();
	}
	return 0;
}

void Lifter::goToPosition(LifterPosition targetPosition) {
	if (targetPosition == LifterPosition::kStop) {
		isJoyStickMode = true;
		setPower = 0;
		return;
	}

	double newTargetLevel = getPositionForLabel(targetPosition);
	if (targetLevel != newTargetLevel || isJoyStickMode == true) {
		isJoyStickMode = false;
		startingLevel = Base.getLifterPosition();
		targetLevel = newTargetLevel;
		controller.Reset();
		controller = PIDFA(PID_P.get(), PID_I.get(), PID_D.get(), PID_F.get(),
				PID_A.get(), PID_I_CONTRIB_CAP.get());
		profile = CubicProfile(fabs(targetLevel - startingLevel), MAX_V.get(), MAX_A.get(),
				MAX_J.get());
	}
}

void Lifter::move(double power) {
	isJoyStickMode = true;
	setPower = power;
}

double Lifter::getPosition() {
	if (!brokenSwitch && Base.isLifterZeroed() == true) {
		return Base.getLifterPosition();
	} else {
		return 0.0;
	}
}

double Lifter::getVelocity() {
	return Base.getLifterVelocity();
}
bool Lifter::isAtPosition() {
	double currentPosition = Base.getLifterPosition();
	if ((currentPosition <= (targetLevel + POSITION_DEAD_BAND))
			&& (currentPosition >= (targetLevel - POSITION_DEAD_BAND))) {
		return true;
	} else {
		return false;
	}
}
