#pragma once

#include "systembase.h"
#include "controllers/closedloop.h"
#include "controllers/profiles.h"

/**
 * @brief %Controls the elevator on which the gripper moves
 *
 * This class is responsible for commanding the elevator motors
 * and brake (ex: by calling `Base.setLifterMotorPower()`),
 * to move the elevator to the location specified by the Controls (which
 * calls methods like `goToPosition(...)` on an instance of this class).
 *
 * @see HardwareAccess
 * @see Controls
 */
class Lifter: public SystemBase {
public:
	/**
	 * @brief List of locations for the lifter to move to
	 *
	 * kStopHere makes the lifter stop where it is.
	 *
	 * TODO: decide if running integral portions of 1000 is good
	 */
	enum class LifterPosition {
		/// Bottom end of travel
		kPickUpBottomTote,
		/// Top end of travel
		kTopOfChannel,
		/// To grip the third tote from a free-standing stack on the ground
		kClearToteWithRC,
		/// Stop where you are. Not a "real" position
		kStop,
		/// To grip the second tote from a free-standing stack on the ground
		kPickUpSecondToteInStackOrDefaultRCGrabPosition,
		 /// We really don't know what this is...Thats because your all bad -Zack
		k2inch,
		 /// Just high enough to have the little stabbing icicles engage the landfill
		kClearTote,
		 /// IDK, yikes!
		kPutOnTote,
		/// NOTE: Change these two for feeder control
		 /// Just high enough to clear the feeder slot
		kClearFeederSlot,
		///
		kAutoStackBottom,
		//
		kAutoStackTop,

		/// Height at which you'd pull a tote from the feeder station
		kGrabFromFeeder,
		kPlaceToteWithRCOnToteToCreateTwoStackWithRCForAutoModeNine
	};

	//TuneableConstant(OneToteStepIncrement);
	//TuneableConstant(TotePositionOnGround);
	//TuneableConstant(TotePositionOnRamp);
	//TuneableConstant(FUDGE_FACTOR_FOR_FULL_LOAD);

	//OneToteStepIncrement.get() + TotePositionOnGround.get() + FUDGE_FACTOR_FOR_FULL_LOAD.get();

	/**
	 * @brief Constructor for the lifter
	 *
	 * Here all the needed state variables and objects for the
	 * lifter are created.
	 */
	Lifter();
	~Lifter();

	/**
	 * @brief Configure the starting lifter state
	 *
	 * In addition to setting the internal state for the lifter,
	 * this method may need to prepare for a homing procedure if
	 * the lifter sensors have not yet been homed.
	 *
	 * @copydoc SystemBase::initialize()
	 * @see SystemBase#initialize()
	 */
	void initialize() override;

	void process() override;
	void debug(Logger&) override;

	/**
	 * @brief Stop relying on position sensors
	 *
	 * Marks the position sensor feedback as broken or not broken.
	 *
	 * @warning This method may be called at any time with any value (even the same value as before).
	 *
	 * @param broken If this is true, the sensors are broken and *cannot* be trusted. If false, they work.
	 */
	void setPositionSensorsBroken(bool broken);

	/**
	 * @brief Command lifter to go to a position
	 *
	 * When this method is called, begin position control,
	 * trying to automatically move the lifter to a given height.
	 *
	 * If the sensors are marked as broken, act as though the
	 * desired position were LifterPosition::kStopHere.
	 *
	 * @warning This method may be called at any time with any value (even the same value as before).
	 *
	 * @param targetPosition Where to try to go to
	 * @see Lifter#move
	 * @see Lifter#LifterPosition
	 */
	void goToPosition(LifterPosition targetPosition);

	/**
	 * @brief Command lifter to try to move with a given power
	 *
	 * If the input power is 0, then the brake on the lifter will
	 * actuate.
	 *
	 * If this method is called after goToPosition(), it will stop
	 * the position control started by it and apply power.
	 *
	 * The bounds at the top and bottom of the lifter should be respected.
	 * No power up should actually be applied if the top flag is tripped,
	 * and none down if the bottom flag is. But **iff** the sensors are marked as
	 * broken, those readings should be ignored.
	 *
	 * @warning This method may be called at any time with any value (even the same value as before).
	 *
	 * @param power The given power to move with. It is in the range [-1, 1].
	 *      +1 means to move with full available power up.
	 *      -1 means to move with full available power down.
	 *      0 means not to apply any power.
	 * @see Lifter#goToPosition
	 */
	void move(double power);

	/**
	 * @brief Get the position of the lifter
	 *
	 * If the position sensors are broken, return what they provide,
	 * even though the values from them should not be trusted.
	 *
	 * @return A floating point number in the range [0,1].
	 *   0 means that the carriage is all the way down, and the high limit switch is tripped.
	 *   1 means that the carriage is all the way up, and the low limit switch is tripped.
	 *   0.5 means (drumroll) that the carriage is halfway in the middle.
	 */
	double getPosition();

	/**
	 * @brief Get the velocity of the lifter
	 *
	 * If the position sensors are broken, return what they provide,
	 * even though the values from them should not be trusted.
	 *
	 * @return A floating point number, in units of X/s, where X is
	 *     the travel distance between the high and low limits of the lifter
	 */
	double getVelocity();
	/**
	 * @brief Get if at position
	 *
	 * @return true if at position
	 */
	bool isAtPosition();
private:
	void applyPowerWithAutoBrake(double power);
	void applyPowerWithManualBrake(double power, bool brake);

	HardwareAccess& Base;

	PIDFA controller;
	CubicProfile profile;
	double startingLevel;

	bool brokenSwitch;
	bool isJoyStickMode;
	bool wasCalBefore;

	double setPower;
	double targetLevel;

	bool overCurrentDown;
};
