/**
 * @mainpage Main Page
 *
 * \section intro_sec Introduction
 *
 * This is the documentation for the 2015 robot code.
 * You should probably start reading with Minesweeper.
 *
 * The 2015 Programming main wiki page is <a href="http://penfieldrobotics.com/wiki/index.php?title=2015:Programming">here</a>.
 *
 * The 2015 IO Map is <a href="http://penfieldrobotics.com/wiki/index.php?title=2015:Robot_IO_Map">here</a>.
 *
 * The 2015 Robot Design overview is <a href="http://penfieldrobotics.com/wiki/index.php?title=2015:Robot_Design_Details">here</a>.
 *
 * WPILIB documentation is hosted <a href="http://first.wpi.edu/FRC/roborio/release/docs/cpp/index.html">here</a>.
 *
 * \section prog_advice Programming Advice
 *
 * Don't be afraid to ask for help --- or search for it.
 *
 * \subsection precommit Before You Commit
 *
 * Run Ctrl+Shift+F to autoformat the files you have been working on. That way all files look roughly the same,
 * and are easier to read. (You also do less work trying to make things look
 * pretty.)
 *
 * Look over your code again, just to be sure you haven't forgot anything.
 *
 * Have someone or sometwo more experienced than you review your code before
 * you push it to the world.
 *
 * \section avoid List of WPILIB Function Calls to Avoid at All Costs
 *
 * Read the source for _anything_ that depends on CAN -- It's expensive.
 *
 * - Solenoid::Get() is evil
 * - DriverStation::IsEnabled() etc. are evil
 * - Serial::Flush() is evil and you shouldn't use it anyway
 *
 * Never call a WPILIB set function twice with the same arguments.
 */

#define FLOATING_UPDATE 1

#include "main.h"
#include "util/common.h"

void notifyMode(std::string modeName) {
	std::cout << "\n" << "     " << modeName << "\n" << std::endl;
}

Minesweeper::Minesweeper() :
				drive(),
				gripper(),
				lifter(),
				logging(),
				vision(),
				puller(),
				auton(drive, gripper, lifter, puller),
				controls(drive, gripper, lifter, logging, vision, auton, puller),
				lights(drive, gripper, lifter, controls) {

}

void Minesweeper::Autonomous() {
	notifyMode("AUTONOMOUS");
	DoubleConstant::UpdateAll();
	auton.setAutoMode(controls.getAutoMode());
	std::vector<SystemBase*> reverse(autonRunOrder.begin(),
			autonRunOrder.end());
	for (SystemBase* system : reverse) {
		system->initialize();
	}
	RobotState.updateMe();
	while (IsAutonomous() && IsEnabled()) {
		HALNetworkCommunicationObserveUserProgramAutonomous();
		RunPeriodic(autonRunOrder);
	}
}

const char* getDisabledMode() {
	if (RobotState.IsAutonomous()) {
		return "DISABLED AUTONOMOUS";
	} else if (RobotState.IsTest()) {
		return "DISABLED TEST";
	} else if (RobotState.IsOperatorControl()) {
		return "DISABLED TELEOP";
	} else {
		return "DISABLED UNKNOWN";
	}
}

void Minesweeper::Disabled() {
	RobotState.updateMe();
	const char* mode = getDisabledMode();
	notifyMode(mode);
	DoubleConstant::UpdateAll();
	MainInit();
	while (RobotState.IsDisabled()) {
		HALNetworkCommunicationObserveUserProgramDisabled();
		const char* nmode = getDisabledMode();
		if (nmode != mode) {
			mode = nmode;
			notifyMode(mode);
			DoubleConstant::UpdateAll();
		}
		RunPeriodic(mainRunOrder);
	}
}
void Minesweeper::OperatorControl() {
	notifyMode("TELEOP");
	DoubleConstant::UpdateAll();
	RobotState.updateMe();
	MainInit();
	while (RobotState.IsEnabled() && RobotState.IsOperatorControl()) {
		HALNetworkCommunicationObserveUserProgramTeleop();
		RunPeriodic(mainRunOrder);
	}
}

void Minesweeper::MainInit() {
	std::vector<SystemBase*> reverse(mainRunOrder.begin(), mainRunOrder.end());
	for (SystemBase* system : reverse) {
		system->initialize();
	}
}

void Minesweeper::Test() {
	notifyMode("TEST");
	DoubleConstant::UpdateAll();
	RobotState.updateMe();
	while (RobotState.IsTest() && RobotState.IsEnabled()) {
		HALNetworkCommunicationObserveUserProgramTest();
		Wait(0.05);
		RobotState.updateMe();
	}
}

void Minesweeper::RunPeriodic(const std::vector<SystemBase*>& runOrder) {
	Profiler prof;
	double t = GetTime();
	for (SystemBase* system : runOrder) {
		system->process();
		prof.add(system->getName());
	}
#if 0
	prof.finish();
#else
	prof.wipe();
#endif
	double diff = GetTime() - t;
#if FLOATING_UPDATE
	Wait(0.007);
#else
	double makeUpWait = bound(0.050 - diff, 0.005, 0.100);
	Wait(makeUpWait);
#endif
	double diff2 = GetTime() - t;
	logging.periodic("Main loop time (run) %f (net) %f", diff, diff2);
	RobotState.updateMe();
}

// Interface and hooks to system running code
void Minesweeper::Prestart() {
	DoubleConstant::LoadAll();
	notifyMode("PRESTART");
}
void Minesweeper::RobotInit() {
	HALNetworkCommunicationObserveUserProgramStarting();
	notifyMode("ROBOT INIT");
}

int main() {
	// load WPILib's HAL
	if (!HALInitialize()) {
		std::cerr << "FATAL ERROR: HAL could not be initialized" << std::endl;
		return -1;
	}
	// Disable buffering on output
	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);
	// Usage reporting for WPILib's internals.
	HALReport(HALUsageReporting::kResourceType_Language,
			HALUsageReporting::kLanguage_CPlusPlus);
	// Create the robot object, load DriverStation,
	// run Prestart() and begin competition cycle
	Minesweeper *robot = new Minesweeper();
	RobotBase::robotSetup(robot);
	return 0;
}
