#pragma once

class Logger;

/**
 * @brief Base class for all system classes on the robot code.
 *
 * Used to organize logging and ensure that the same
 * core methods are present on every class.
 */
class SystemBase {
protected:
	/**
	 * @brief Constructor for SystemBase.
	 *
	 * It records the name of the system subclassing
	 * SystemBase for logging purposes.
	 *
	 * @param name Name of the system
	 */
	explicit SystemBase(const char* name);
public:
	virtual ~SystemBase();

	/**
	 * @brief Called when Teleop, Auton, or Disabled begin
	 *
	 * This method will be called exactly once when a new mode is entered.
	 * It will be called before process() is called.
	 *
	 * In here, the state of the subsystem should be reset.
	 *
	 * (If you want to know what mode the robot is in, call
	 *  <code>RobotState.IsEnabled()</code>)
	 *
	 * @see SystemBase#process()
	 */
	virtual void initialize() = 0;

	/**
	 * @brief Will be called periodically.
	 *
	 * This method will be called by the main robot code roughly
	 * (but not exactly) every 20 milliseconds.
	 *
	 * All control of other mechanisms and systems should
	 * occur in this method. Other methods typically exist
	 * to change or reveal internal state.
	 *
	 * @see SystemBase#initialize()
	 */
	virtual void process() = 0;

	/**
	 * @brief Called to record numeric or state information.
	 *
	 * Uses the Logger class to record information. That information
	 * will be optionally sent to the dashboard on the controls, so
	 * that we can know what sensor values the robot is receiving and
	 * what its internal state is.
	 */
	virtual void debug(Logger&);

	/**
	 * @brief Get the name of the subclass
	 */

	const char* getName() const;
private:
	const char* m_name;
	SystemBase(const SystemBase&) = delete;
	void operator=(const SystemBase&) = delete;
};

// provide base includes for every subclass of SystemBase
#include "util/constants.h"
#include "util/logger.h"
#include "util/misc.h"
#include "util/udplog.h"

class HardwareAccess;
