#include "camera.h"
#include "udplog.h"

#include <sstream>
#include <nivision.h>
#include "Timer.h"

#define MULTI_STREAM_CAMERAS true

#define DEBUG_CAMERAS true
#define DEBUG_MINIMAL true

#if DEBUG_CAMERAS
UDPLog debug(4001);
#endif

Camera* Camera::activeCamera = nullptr;

// IMAQdx can support as many camera acquisitions as the usb driver can.
// On the Roborio, this number is positive, integral, and less than two under default driver settings.
// We therefore can't simultaneously acquire from the cameras;
// For more info, see:
// * http://digital.ni.com/manuals.nsf/websearch/43FDE97F57A473BD8625761C006707B4
// * https://decibel.ni.com/content/docs/DOC-11641
// * http://digital.ni.com/public.nsf/allkb/D13527D8471D5FC5862576A4006BD081?OpenDocument

IMAQdxError onError(IMAQdxError e, std::string call, int line,
		std::string function) {
	std::stringstream s;
	s << call << "at " << function << ":" << line << " ";
	if (e != IMAQdxErrorSuccess) {
		char buf[1024];
		IMAQdxGetErrorString(e, buf, 1024);
		s << "gives error: " << (long) e << " is: " << std::string(buf);
	} else {
#if DEBUG_MINIMAL
		return e;
#endif
		s << "succeeded";
	}
	s << std::endl;
#if DEBUG_CAMERAS
	debug.log(s.str().c_str());
#endif
	return e;
}
#define DXERR( M ) imaqError = onError( (M) , #M, 	__LINE__, __FUNCTION__);

/**
 * From WPILIB/USBCamera.cpp
 *
 * Helper function to determine the size of a jpeg. The general structure of
 * how to parse a jpeg for length can be found in this stackoverflow article:
 * http://stackoverflow.com/a/1602428. Be sure to also read the comments for
 * the SOS flag explanation.
 */
unsigned int GetJpegSize(void* buffer, unsigned int buffSize) {
	uint8_t* data = (uint8_t*) buffer;
	if (!(data[0] == 0xff && data[1] == 0xd8))
		return 0;
	unsigned int pos = 2;
	while (pos < buffSize) {
		// All control markers start with 0xff, so if this isn't present,
		// the JPEG is not valid
		if (!(data[pos] == 0xff))
			return 0;
		unsigned char t = data[pos + 1];
		// These are RST markers. We just skip them and move onto the next marker
		if (t == 0x01 || (t >= 0xd0 && t <= 0xd7)) {
			pos += 2;
		} else if (t == 0xd9) {
			// End of Image, add 2 for this and 0-indexed
			return pos + 2;
		} else if (!(t != 0xd8)) {
			// Another start of image, invalid image
			return 0;
		} else if (t == 0xda) {
			// SOS marker. The next two bytes are a 16-bit big-endian int that is
			// the length of the SOS header, skip that
			unsigned int len = (((unsigned int) (data[pos + 2] & 0xff)) << 8
					| ((unsigned int) data[pos + 3] & 0xff));
			pos += len + 2;
			// The next marker is the first marker that is 0xff followed by a non-RST
			// element. 0xff followed by 0x00 is an escaped 0xff. 0xd0-0xd7 are RST
			// markers
			while (data[pos] != 0xff || data[pos + 1] == 0x00
					|| (data[pos + 1] >= 0xd0 && data[pos + 1] <= 0xd7)) {
				pos += 1;
				if (pos >= buffSize)
					return 0;
			}
		} else {
			// This is one of several possible markers. The next two bytes are a 16-bit
			// big-endian int with the length of the marker header, skip that then
			// continue searching
			unsigned int len = (((unsigned int) (data[pos + 2] & 0xff)) << 8
					| ((unsigned int) data[pos + 3] & 0xff));
			pos += len + 2;
		}
	}

	return 0;
}

Image* createFillerImage() {
	Image* image = imaqCreateImage(IMAQ_IMAGE_RGB, 0);
	imaqSetImageSize(image, 320, 240);
	PixelValue pixel;
	pixel.rgb.alpha = 255;
	pixel.rgb.R = 255;
	pixel.rgb.G = 230;
	pixel.rgb.B = 40;
	imaqFillImage(image, pixel, NULL);
	return image;
}
Camera::Camera(const char* namex, Size s, FPS f, Type t) :
				name(namex) {
	image = createFillerImage();
	isReady = false;
	lastAttemptTime = 0;
	videoMode = ConstructMode(s, f, t);
	Start();
}
Camera::~Camera() {
	Stop();
	imaqDispose(image);
}
Image* Camera::GetNewImage() {
	if (PreCapture()) {
		uInt32 frameNumber = (uInt32) -1;
		DXERR(IMAQdxGrab(s_id, image, true, &frameNumber));
		PostCapture(imaqError, frameNumber);
	}
	return image;
}
/// @return Actual image size in buffer
uint32_t Camera::GetNewImageData(uint8_t* buffer, uint32_t size) {
	uint32_t imgSize = 0;
	if (PreCapture()) {
		uInt32 frameNumber = (uInt32) -1;
		DXERR(
				IMAQdxGetImageData(s_id, buffer, (uInt32 )size,
						IMAQdxBufferNumberModeNext, 0, &frameNumber));
		if (!imaqError) {
			imgSize = GetJpegSize(buffer, size);
		}
		PostCapture(imaqError, frameNumber);
	}
	return imgSize;
}
bool Camera::PreCapture() {
	if (!isReady) {
		if (GetTime() < lastAttemptTime + 0.500) {
			return false;
		}
		Start();
		if (!isReady) {
			return false;
		}
	}

#if MULTI_STREAM_CAMERAS

#else
	if (this != activeCamera) {
		if (activeCamera != nullptr) {
			DXERR(IMAQdxStopAcquisition(activeCamera->s_id));
			DXERR(IMAQdxUnconfigureAcquisition(activeCamera->s_id));
		}
		DXERR(IMAQdxConfigureGrab(this->s_id));
		DXERR(IMAQdxStartAcquisition(this->s_id));
		activeCamera = this;
	}
#endif
	return true;
}
#if DEBUG_CAMERAS
void Camera::PostCapture(IMAQdxError e, uInt32 frameNo) {
	debug.log(
			std::string("Camera " + std::to_string(this->s_id) + " Frame "
					+ std::to_string(frameNo) + "\n").c_str());
#else
	void Camera::PostCapture(IMAQdxError e, uInt32) {
#endif
	if (e != IMAQdxErrorSuccess) {
		activeCamera = nullptr;
		Stop();
		Fail();
	}
}

void Camera::Fail() {
	isReady = false;
	lastAttemptTime = GetTime();
}
std::string Camera::ConstructMode(Size s, FPS f, Type t) {
	if (s == Size::x1280y800) {
		// mysterious special case
		return "1280 x 800 YUY 2 10.00 fps";
	}

	// select size first, then type, then list all fps types
	int x = 640, y = 360;
	switch (s) {
	case Size::x640y480:
		x = 640, y = 480;
		break;
	case Size::x1280y720:
		x = 1280, y = 720;
		break;
	case Size::x1280y800:
		x = 1280, y = 800;
		break;
	case Size::x960y544:
		x = 960, y = 544;
		break;
	case Size::x800y448:
		x = 800, y = 448;
		break;
	case Size::x640y360:
		x = 640, y = 360;
		break;
	case Size::x424y240:
		x = 424, y = 240;
		break;
	case Size::x352y288:
		x = 352, y = 288;
		break;
	case Size::x320y240:
		x = 320, y = 240;
		break;
	case Size::x800y600:
		x = 800, y = 600;
		break;
	case Size::x176y144:
		x = 176, y = 144;
		break;
	case Size::x160y120:
		x = 160, y = 120;
		break;
	}
	int fps = 750;
	switch (f) {
	case FPS::fps30_00:
		fps = 3000;
		break;
	case FPS::fps20_00:
		fps = 2000;
		break;
	case FPS::fps15_00:
		fps = 1500;
		break;
	case FPS::fps10_00:
		fps = 1000;
		break;
	case FPS::fps7_50:
		fps = 750;
		break;
	}
	const int max = 921600000; // centipixels/second for max rate YUY mode (1280*720*10)
	if (t == Type::YUY2) {
		int max_allowed_fps = (max / (x * y));
		if (max_allowed_fps < fps) {
			if (max_allowed_fps >= 2000) {
				fps = 2000;
			} else if (max_allowed_fps >= 1500) {
				fps = 1500;
			} else if (max_allowed_fps >= 1000) {
				fps = 1000;
			} else {
				fps = 750;
			}
		}
	}

	char buf[30];
	sprintf(buf, "%d x %d %s %3.2f fps", x, y,
			t == Type::JPEG ? "JPEG" : "YUY 2", fps / 100.0);
	std::string result(buf);
	return result;

}

void Camera::ConfigCameraMode() {
	uInt32 count = 0;
	uInt32 currentMode = 0;
	DXERR(IMAQdxEnumerateVideoModes(s_id, NULL, &count, &currentMode));
	IMAQdxVideoMode modes[count];
	DXERR(IMAQdxEnumerateVideoModes(s_id, modes, &count, &currentMode));

	IMAQdxVideoMode selectedMode = modes[currentMode];
	for (uInt32 i = 0; i < count; i++) {
		IMAQdxVideoMode mode = modes[i];
		std::string option(mode.Name);
		if (videoMode == option) {
			selectedMode = mode;
			break;
		}
	}
	printf("Camera mode (session %d) is >%s<\n", (int) s_id, selectedMode.Name);
	DXERR(
			IMAQdxSetAttribute(s_id, IMAQdxAttributeVideoMode, IMAQdxValueTypeU32, selectedMode.Value));
}

void Camera::Start() {
	DXERR(IMAQdxOpenCamera(name, IMAQdxCameraControlModeController, &s_id));
	if (imaqError != IMAQdxErrorSuccess) {
		Fail();
		return;
	}

	ConfigCameraMode();

#if MULTI_STREAM_CAMERAS
	DXERR(IMAQdxConfigureAcquisition(this->s_id, 1, 3));
	DXERR(IMAQdxStartAcquisition(this->s_id));
#endif

	isReady = true;
	lastAttemptTime = 0;
	return;
}
void Camera::Stop() {
	DXERR(IMAQdxCloseCamera(s_id));
}

