#include "constants.h"
#include <set>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include "common.h"
#include <networktables/NetworkTable.h>

std::set<DoubleConstant*> constants;
NetworkTable* table = nullptr;
const char* PATH = "/home/lvuser/constants.txt";
const char* TABLE = "Constants";

DoubleConstant::DoubleConstant(double value, const char* dname) :
				codeValue(value),
				name(dname) {
	constants.insert(this);
	currentValue = codeValue;
}

DoubleConstant::~DoubleConstant() {
	constants.erase(this);
}
double DoubleConstant::get() const {
	return currentValue;
}

void DoubleConstant::WriteAll() {
	std::ofstream f(PATH, std::ios::out);
	if (!f.good()) {
		printf("Constants file could not be written");
		return;
	}

	printf("How many doubles? %d\n", constants.size());
	for (DoubleConstant* x : constants) {
		f << "D " << x->currentValue << " " << x->name << std::endl;
		printf("%3.6f <<- %s\n", x->currentValue, x->name);
	}
	printf("Writing constants done\n");
}

bool DoubleConstant::ReadAll() {
	if (!table) {
		table = NetworkTable::GetTable(TABLE);
	}

	std::cout << "Reading constants..." << std::endl;

	bool perfect = true;
	std::map<std::string, double> kvp;

	std::ifstream in(PATH, std::ios::in);
	if (!in.good()) {
		return false;
	}
	std::string line;
	while (std::getline(in, line)) {
		std::istringstream i(line);
		std::string name;
		std::string key;
		double yy;
		i >> key >> yy >> name;
		if (i.eof() && !i.fail() && !i.bad()) {
			if (key == "D") {
				kvp[name] = yy;
			} else {
				printf("Unidentified key: |%s|\n", key.c_str());
				perfect = false;
			}
		} else {
			printf("Malformatted line: |%s|\n", line.c_str());
			perfect = false;
		}
	}

	// update values, put into /Constants/
	for (DoubleConstant* x : constants) {
		if (kvp.count(x->name) > 0) {
			double newVal = kvp[x->name];
			if (newVal != x->currentValue) {
				std::cout << "Value " << x->name << " changed from "
						<< x->currentValue << " to " << newVal;
			}
			x->currentValue = newVal;
		} else {
			perfect = false;
			printf("%s not yet present\n", x->name);
		}
		try {
			table->PutNumber(x->name, x->currentValue);
		} catch (const std::exception &) {
			printf("Nettables failed on pushing %s\n", x->name);
		}
	}
	std::cout << "... Done." << std::endl;
	return perfect;
}
bool DoubleConstant::CheckAll() {
	if (table == NULL) {
		table = NetworkTable::GetTable(TABLE);
	}
	// read values from SmartDashboard/Constants/
	// based on the keys
	std::stringstream msg;
	msg << std::setprecision(4) << std::fixed;
	msg << "Updating all constants ...\n";
	bool changed = false;
	for (DoubleConstant *x : constants) {
		double newValue = x->currentValue;
		try {
			newValue = table->GetNumber(x->name);
		} catch (const std::exception &) {
			printf("Nettables errored on %s\n", x->name);
			continue;
		}

		if (newValue != x->currentValue) {
			changed = true;
			msg << "constant " << x->name << " value changed from "
					<< x->currentValue << " to " << newValue << " (code has "
					<< x->codeValue << ")\n";
		} else if (x->currentValue != x->codeValue) {
			msg << "constant " << x->name << " value (" << x->currentValue
					<< ") does not match code (" << x->codeValue << ")\n";
		}
		x->currentValue = newValue;
	}

	msg << "... Done\n";
	std::cout << msg.str();
	std::flush(std::cout);
	return changed;
}
void DoubleConstant::LoadAll() {
	double time = GetTime();
	if (!table) {
		table= NetworkTable::GetTable(TABLE);
	}

	if (!ReadAll()) {
		std::cout << "Constants file not present: writing\n" << std::endl;
		WriteAll();
		if (ReadAll()) {
			std::cout << "Second read succeeded!" << std::endl;
		} else {
			std::cout << "Second read failed. :-(" << std::endl;
		}
	}
	printf("Constants loaded (%.4f s)\n", GetTime() - time);
}

void DoubleConstant::UpdateAll() {
	double time = GetTime();
	if (CheckAll()) {
		WriteAll();
	}
	printf("Constants reloaded (%.4f s)\n", GetTime() - time);
}
