#include "imageserver.h"
#include "util/common.h"
#include <iostream>
#include "udplog.h"
#include <netinet/in.h>
#include <unistd.h>
#include <memory.h>

#define DEBUG_SERVER true

#if DEBUG_SERVER
UDPLog debugServer(4002);
#define DBG(X) debugServer.log(std::string("line " + std::to_string(__LINE__) + ", port " + std::to_string(m_port) + ": " X + "\n").c_str())
#else
#define DBG(X)
#endif

void ImageServer::FreeImageData(MemoryBuffer* imageData) {
	if (imageData->data == nullptr) {
		return;
	}
	(*imageData->deleter)(imageData->data);
	imageData->data = nullptr;
}
void ImageServer::IMAQDelete(uint8_t* data) {
	imaqDispose(data);
}

ImageServer::ImageServer(uint16_t port) :
				m_serverThread(
						(std::string("ImageServer Sending (port ")
								+ std::to_string((int) port) + ")").c_str(),
						(FUNCPTR) &ImageServer::RunServe),
				m_connectionThread(
						(std::string("ImageServer Connections (port ")
								+ std::to_string((int) port) + ")").c_str(),
						(FUNCPTR) &ImageServer::RunManageConnections) {
	m_port = port;
	m_serverThread.Start((uint32_t) this);
	m_connectionThread.Start((uint32_t) this);
}
ImageServer::~ImageServer() {

}

void ImageServer::SetImage(Image* image) {
	unsigned int dataSize = 0;
	uint8_t* data = (uint8_t*) imaqFlatten(image, IMAQ_FLATTEN_IMAGE,
			IMAQ_COMPRESSION_JPEG, 800, &dataSize);
	SetImageRawData(data, dataSize, dataSize, &ImageServer::IMAQDelete);
}
void ImageServer::SetImageRawData(uint8_t* data, uint32_t bufferSize,
		uint32_t imageSize, DeleterFunc deleter) {
	if (data == nullptr) {
		std::cerr << "Input data should NOT be null" << std::endl;
		return;
	}
	if (imageSize > bufferSize) {
		std::cerr << "Image is too large" << std::endl;
		return;
	}

	// find JPEG start sequence
	unsigned int start = 0;
	while (start < bufferSize - 1) {
		if (data[start] == 0xFF && data[start + 1] == 0xD8)
			break;
		else
			start++;
	}
	bufferSize -= start;

	if (bufferSize < 2) {
		DBG("DATASIZE < 2");
	}

	{
		std::unique_lock<std::recursive_mutex> lock(m_imageMutex);
		FreeImageData(&m_imageData);
		m_imageData.data = data;
		m_imageData.imageSize = imageSize;
		m_imageData.start = start;
		m_imageData.deleter = deleter;
		m_newImageVariable.notify_all();
	}
}

void ImageServer::ManageConnections() {
	constexpr int maxConnections = 10;
	int sock = socket(AF_INET, SOCK_STREAM, 0);

	if (sock == -1) {
		DBG("SOCK = -1");
	}

	int reuseAddr = 1;
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuseAddr,
			sizeof(reuseAddr)) == -1) {
		DBG("SOCK can't setsockopt to reuse: "  + std::string(strerror(errno)));
	}

	int optval = 1;
	if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *) (&optval),
			sizeof(optval))) {
		DBG("SOCK can't setsockopt to enable flushing: "  + std::string(strerror(errno)));
	}

	struct timeval timeout;
	timeout.tv_sec = 1;
	timeout.tv_usec = 0;
	if (setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, &timeout,
			sizeof(struct timeval))) {
		DBG("SOCK can't setsockopt to set a low receive timeout: " + std::string(strerror(errno)));
	}
	if (setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, &timeout,
			sizeof(struct timeval))) {
		DBG("SOCK can't setsockopt to set a low sending timeout: "  + std::string(strerror(errno)));
	}

	sockaddr_in address;
	memset(&address, 0, sizeof(address));
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = htonl(INADDR_ANY);
	address.sin_port = htons(m_port);

	if (bind(sock, (struct sockaddr *) &address, sizeof(address)) == -1) {
		DBG("SOCK can't bind: " + std::string(strerror(errno)));
	}

	if (listen(sock, maxConnections) == -1) {
		DBG("SOCK can't listen: "+ std::string(strerror(errno)));
	}

	while (true) {
		sockaddr_in clientAddress;
		socklen_t clientAddressLen = sizeof(clientAddress);
		int conn = accept(sock, (struct sockaddr*) &clientAddress,
				&clientAddressLen);
		if (conn == -1) {
			DBG("Accept failed: " + std::string(strerror(errno)));
			continue;
		}

		{
			std::unique_lock<std::recursive_mutex> lock(m_connectionMutex);
			m_connections.insert(conn);
			DBG("Added new connection");
		}
	}
	close(sock);
}

void ImageServer::Serve() {
	DBG("Serving");
	constexpr uint8_t kMagicNumber[] = { 0x01, 0x00, 0x00, 0x00 };
	while (true) {
		while (true) {
			MemoryBuffer imageData;
			{
				std::unique_lock<std::recursive_mutex> lock(m_imageMutex);
				m_newImageVariable.wait(lock);
				imageData = m_imageData;
				m_imageData.data = nullptr;
			}

			unsigned int size = imageData.imageSize;
			unsigned int netSize = htonl(size);
			unsigned int start = imageData.start;
			uint8_t *data = imageData.data;

			if (data == nullptr)
				continue;

			DBG("Got an image");
			std::set<int> connections;
			{
				std::unique_lock<std::recursive_mutex> lock(m_connectionMutex);
				connections = m_connections;
			}
			for (int conn : connections) {
				DBG("Writing to a connection");
				if (write(conn, kMagicNumber, sizeof(kMagicNumber)) == -1) {
					DBG("Sending magic number failed: " + std::string(strerror(errno)));
					FlushConnection(conn);
					continue;
				}
				if (write(conn, &netSize, sizeof(netSize)) == -1) {
					DBG("Sending image size failed: " + std::string(strerror(errno)));
					FlushConnection(conn);
					continue;
				}
				if (write(conn, &data[start], sizeof(uint8_t) * size) == -1) {
					DBG("Sending image data failed: " + std::string(strerror(errno)));
					FlushConnection(conn);
					continue;
				}
			}

			FreeImageData(&imageData);
			Wait(0.005);
		}
	}
}

void ImageServer::FlushConnection(int conn) {
	DBG("Flushing a connection");
	close(conn);
	{
		std::unique_lock<std::recursive_mutex> lock(m_connectionMutex);
		m_connections.erase(conn);
	}
}

void ImageServer::RunServe(uint32_t d0) {
	((ImageServer*) d0)->Serve();
}
void ImageServer::RunManageConnections(uint32_t d0) {
	((ImageServer*) d0)->ManageConnections();
}
