#pragma once

#include <stdint.h>

class SystemBase;

/**
 * @brief Helper class to log data from subsystems.
 *
 * Accumulates a list of info.
 */
class Logger {
public:
	Logger();
	~Logger();
	/**
	 * @brief Add more information to the logger
	 *
	 * Example use:
	 *
	 * @code
	 * void MySystem::debug(Logger& logger) {
	 *	    logger.write(this, "Cost of chicken: %d & eggs: %4.2f", 15, getEggFactor());
	 *	    logger.write(this, "Current intake state: %s", isBroken ? "Broken" : "Good");
	 * }
	 * @endcode
	 *
	 * The line length limit is about 35 characters
	 * (including the name of your system + 2 more).
	 *
	 * @param this_ Use `this` in code.
	 * @param formatString A `printf`-compatible format string. Does not need a newline.
	 * @param ... Arguments to match the format string (like `printf`).
	 *
	 *
	 *
	 */
	void write(SystemBase* this_, const char* formatString, ...);
	/**
	 * @brief Get the so-far accumulated lines of text of the Logger
	 *
	 * @return A vector of strings (without newline).
	 */
	const char* getText();
private:
	Logger(const Logger&) = delete;
	void operator=(const Logger&) = delete;
	char* buffer;
	uint32_t head;
};
