#include <util/sensors.h>
#include "AnalogInput.h"
#include "Counter.h"
#include "DigitalInput.h"
#include "interfaces/Accelerometer.h"

#define ENABLE_SEMAPHORE true

const double GYRO_SAMPLE_PERIOD = 0.5; // seconds
const double GYRO_SAVE_PERIOD = 6.0; // seconds
const int GYRO_BUFFER_SIZE = GYRO_SAVE_PERIOD / GYRO_SAMPLE_PERIOD; // #
const double ACCELEROMETER_PERIOD = 0.1;
const double ACCELEROMETER_SAVE_PERIOD = 6.0;
const double ACCELEROMETER_BUFFER_SIZE = ACCELEROMETER_SAVE_PERIOD
		/ ACCELEROMETER_PERIOD;
const double ONE_EARTH_GRAVITY = 9.801;

ADW22307::ADW22307(AnalogInput* output) :
				rateInput(output),
				semaphore(),
				notifier(callCalibrate, this),
				buffer() {
	sensitivity = 0.007;

	rateInput->SetAverageBits(0);
	rateInput->SetOversampleBits(10);

	// microcal
	float avgRd = ((float) rateInput->GetAverageValue() + 0.5)
			/ ((float) (1 << rateInput->GetAverageBits()));
	centerIntegral = (int32_t) avgRd;
	centerFractional = (double) avgRd - (double) centerFractional;

	rateInput->SetAccumulatorDeadband(0);
	rateInput->SetAccumulatorCenter(centerIntegral);
	rateInput->InitAccumulator();
	runningTotal.count = 0;
	runningTotal.rvalue = 0;

	isCalibrating = false;
	notifier.StartPeriodic(GYRO_SAMPLE_PERIOD);
}

ADW22307::~ADW22307() {
	notifier.Stop();
}

void ADW22307::beginCalibration() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	if (isCalibrating) {
		return;
	}

	rateInput->GetAccumulatorOutput(&runningTotal.rvalue, &runningTotal.count);
	rateInput->ResetAccumulator();

	buffer.clear();

	isCalibrating = true;
}

ADW22307::Sample& ADW22307::Sample::operator+=(const ADW22307::Sample& x) {
	rvalue += x.rvalue;
	count += x.count;
	return *this;
}

void ADW22307::finishCalibration() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	if (!isCalibrating) {
		return;
	}

	if (buffer.size() == 0) {
		calibrate();
	}

	Sample total = { 0, 0 };
	for (Sample s : buffer) {
		total += s;
	}

	float avgRd = ((float) total.rvalue / (float) total.count + 0.5);
	int32_t jump = (int32_t) avgRd;
	centerIntegral += avgRd;
	centerFractional = (double) avgRd - (double) jump;

	rateInput->SetAccumulatorCenter(centerIntegral);
	rateInput->ResetAccumulator();
	runningTotal.count = 0;
	runningTotal.rvalue = 0;

	isCalibrating = false;
}

bool ADW22307::isDisconnected() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	return rateInput->GetVoltage() < 0.5;
}

ADW22307::Reading ADW22307::get() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif

	int64_t rawValue;
	uint32_t count;
	rateInput->GetAccumulatorOutput(&rawValue, &count);
	rawValue += runningTotal.rvalue;
	count += runningTotal.count;

	int64_t value = rawValue - (int64_t) ((float) count * centerFractional);

	double correctedValue = ((double) value) * 0.000000001
			* ((double) rateInput->GetLSBWeight())
			* ((double) (1 << rateInput->GetAverageBits()))
			/ (rateInput->GetSampleRate());
	double scaledValue = correctedValue / sensitivity;

	Reading r { rateInput->GetVoltage(), scaledValue, 2.0 };

	return r;
}

void ADW22307::reset() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif

	rateInput->ResetAccumulator();
}

void ADW22307::calibrate() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	if (!isCalibrating) {
		return;
	}

	Sample s;
	{
		rateInput->GetAccumulatorOutput(&s.rvalue, &s.count);
		rateInput->ResetAccumulator();
		runningTotal += s;
	}
	buffer.push_back(s);
	while (buffer.size() > GYRO_BUFFER_SIZE) {
		buffer.pop_front();
	}
}

void ADW22307::callCalibrate(void* p) {
	((ADW22307*) p)->calibrate();
}

BidirectionalCounter::BidirectionalCounter(DigitalInput* input) {
	counter = new Counter(input);
	lastCount = 0;
	upward = true;
}
BidirectionalCounter::~BidirectionalCounter() {
	delete counter;
}
int BidirectionalCounter::Get() {
	int value;
	if (upward) {
		value = lastCount + counter->Get();
	} else {
		value = lastCount - counter->Get();
	}

	return value;
}
void BidirectionalCounter::SetReverseDirection(bool) {
}
void BidirectionalCounter::Reset() {
	lastCount = 0;
	counter->Reset();
}
void BidirectionalCounter::SetLastAppliedPower(double pow) {
	if (upward) {
		lastCount += counter->Get();
		counter->Reset();
	} else {
		lastCount -= counter->Get();
		counter->Reset();
	}
	if (pow > 0) {
		upward = true;
	}
	if (pow < 0) {
		upward = false;
	}
}

IntegratedAccelerometer::IntegratedAccelerometer(Accelerometer* a) :
				acc(a),
				notifier(callUpdate, this),
				semaphore() {
	FinishCalibration();
	notifier.StartPeriodic(ACCELEROMETER_PERIOD);
}
IntegratedAccelerometer::~IntegratedAccelerometer() {
	notifier.Stop();
}
IntegratedAccelerometer::Result IntegratedAccelerometer::GetForwardComponent() {
	return GetComponents().forward;
}
IntegratedAccelerometer::Result IntegratedAccelerometer::GetRightComponent() {
	return GetComponents().right;
}
IntegratedAccelerometer::Result IntegratedAccelerometer::GetUpComponent() {
	return GetComponents().up;
}

IntegratedAccelerometer::AllResults IntegratedAccelerometer::GetComponents() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	AllResults r;
	r.forward = currentY;
	r.right = currentX;
	r.up = currentZ;
	return r;
}
void IntegratedAccelerometer::Reset() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	currentX.position = 0.0;
	currentY.position = 0.0;
	currentZ.position = 0.0;
	currentX.velocity = 0.0;
	currentY.velocity = 0.0;
	currentZ.velocity = 0.0;
}
void IntegratedAccelerometer::ResetPosition() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	currentX.position = 0.0;
	currentY.position = 0.0;
	currentZ.position = 0.0;
}
void IntegratedAccelerometer::ResetVelocity() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	currentX.velocity = 0.0;
	currentY.velocity = 0.0;
	currentZ.velocity = 0.0;
}

void IntegratedAccelerometer::BeginCalibration() {
	// TODO calibrate to discover angle & scale to world plane
	// (and to take multiple sample)
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	samples.clear();
	isCalibrating = true;
}
void IntegratedAccelerometer::FinishCalibration() {
#if ENABLE_SEMAPHORE
	Synchronized sync(semaphore);
#endif
	isCalibrating = false;
	samples.push_back( { acc->GetX(), acc->GetY(), acc->GetZ() });
	Point r = { 0.0, 0.0, 0.0 };
	for (Point q : samples) {
		r.X += q.X;
		r.Y += q.Y;
		r.Z += q.Z;
	}
	double scale = 1 / samples.size();
	r.X *= scale;
	r.Y *= scale;
	r.Z *= scale;
	samples.clear();

	restingX = r.X;
	restingY = r.Y;
	restingZ = r.Z;
	printf("Acc cal: rest X %f Y %f Z %f\n", restingX, restingY, restingZ);
}
void IntegratedAccelerometer::callUpdate(void* p) {
	((IntegratedAccelerometer*) p)->update();
}
void interpolateAcc(IntegratedAccelerometer::Result* r, double gs) {
	r->acceleration = gs / ONE_EARTH_GRAVITY;
	r->velocity += ACCELEROMETER_PERIOD * r->acceleration;
	r->position += ACCELEROMETER_PERIOD * r->velocity;
}

void IntegratedAccelerometer::update() {
	Point p;
	{
#if ENABLE_SEMAPHORE
		Synchronized sync(semaphore);
#endif
		p.X = acc->GetX();
		p.Y = acc->GetY();
		p.Z = acc->GetZ();
	}
	if (isCalibrating) {
		samples.push_back( { p.X, p.Y, p.Z });
		while (samples.size() > ACCELEROMETER_BUFFER_SIZE) {
			samples.pop_front();
		}
	}
	interpolateAcc(&currentX, p.X - restingX);
	interpolateAcc(&currentY, p.Y - restingY);
	interpolateAcc(&currentZ, p.Z - restingZ);
}
