#pragma once

#include "Notifier.h"
#include <vector>
#include <utility>
#include <deque>

class AnalogInput;

/**
 * @brief Class to use the ADW22307 Gyro
 */
class ADW22307 {
public:
	/// @cond DoNotInclude
	typedef struct {
		double rateRawVoltage;
		double angleEstimate;
		double angleEstimateError;
	} Reading;
	/// @endcond

	/**
	 * @brief Constructor
	 *
	 * @param input Analog channel from which to find gyro readings
	 */
	ADW22307(AnalogInput* input);
	virtual ~ADW22307();

	/**
	 * @brief Begin calibration procedure
	 *
	 * After calibration is begin, you must #finishCalibration() for meaningful readings
	 *
	 * The recommended calibration time is 6+ seconds. The last few should be perfectly still.
	 */
	void beginCalibration();
	/**
	 * @brief End calibration procedure
	 *
	 * @warning #beginCalibration() must be called first
	 *
	 * This also calls #reset().
	 */
	void finishCalibration();

	/**
	 * @brief Get the estimated change in angle since last #reset()
	 *
	 * @return A struct comprising the change in angle (and other info)
	 */
	Reading get();
	/**
	 * @brief Reset the gyro's current angle estimate to be zero.
	 */
	void reset();

	/**
	 * @brief Tells if the gyro is disconnected
	 *
	 * Only check for this when the robot is at rest
	 */
	bool isDisconnected();
private:
	class Sample {
	public:
		int64_t rvalue;
		uint32_t count;
		Sample& operator+=(const Sample& x);
	};

	ADW22307(const ADW22307&) = delete;
	void operator=(const ADW22307&) = delete;

	void calibrate();
	static void callCalibrate(void*);

	AnalogInput* rateInput;
	ReentrantSemaphore semaphore;
	Notifier notifier;
	std::deque<Sample> buffer;

	int32_t centerIntegral;
	float centerFractional;
	Sample runningTotal;

	bool isCalibrating;
	double sensitivity;
};


class Counter;
class DigitalInput;
class BidirectionalCounter {
public:
	BidirectionalCounter(DigitalInput* channel);
	~BidirectionalCounter();
	int Get();
	void Reset();
	void SetLastAppliedPower(double pow);
	void SetReverseDirection(bool reverse);
private:
	Counter* counter;
	int lastCount;
	bool upward;
};

/// Does not compensate for rotation
/// TODO: make it compensate for angling and gravity
class Accelerometer;
class IntegratedAccelerometer {
public:
	typedef struct {
		double position;
		double velocity;
		double acceleration;
	} Result;
	typedef struct {
		Result forward;
		Result right;
		Result up;
	} AllResults;
	IntegratedAccelerometer(Accelerometer*);
	~IntegratedAccelerometer();
	Result GetForwardComponent();
	Result GetRightComponent();
	Result GetUpComponent();
	AllResults GetComponents();

	void Reset();
	void ResetPosition();
	void ResetVelocity();
	void BeginCalibration();
	void FinishCalibration();
private:
	static void callUpdate(void*);
	void update();
	Accelerometer* acc;
	Notifier notifier;
	ReentrantSemaphore semaphore;
	Result currentX;
	Result currentY;
	Result currentZ;
	double restingX;
	double restingY;
	double restingZ;
	typedef struct {
		double X;
		double Y;
		double Z;
	} Point;
	std::deque<Point> samples;
	bool isCalibrating;
};

